Require Import Hahn Omega Operations LibBase Language LangSemantics ExpandFcall.
Require Import ClassicalDescription ClassicalChoice.

Set Implicit Arguments.

Structure libimpl_base (lib : libspec) := 
  { li_assumes : libspec -> Prop ; 
    li_code : impl_environment ;
    li_locf : libspec_logaddr lib -> forall lib', libspec_logaddr lib' -> Prop }.

Definition compat_loc (libs : libspec -> Prop) (g : event -> Prop) 
           A (reqf : event -> A -> Prop) 
           (locf : A -> forall lib, libspec_logaddr lib -> Prop) :=
  forall a (Ga : g a) loca (ALLa : reqf a loca)
         b (Gb : g b) locb (ALLb : reqf b locb) (NEQ: loca <> locb) 
         libi (LIBI: libs libi),
    set_disjoint (locf loca libi) (locf locb libi).

Definition compat_alloc (lib: libspec) (g: event -> Prop) (fs : event -> event) := 
  forall loc a (Ga : g a) (ALLa : libspec_allocs lib (fs a) loc) 
             b (Gb : g b) (ALLb : libspec_allocs lib (fs b) loc),
    fs a = fs b.

Definition locally_sound_impl (lib : libspec) (impl : libimpl_base lib) :=
  forall (g : event -> Prop) po rf so hb (fs : event -> event)
         (GDOM : forall a (Ga: g a), lib_events lib (fs a))
         (PO : po ⊆ g × g)
         (RF : rf ⊆ g × g) 
         (RFDOM : rf ⊆ ⋃l ∈ impl.(li_assumes), lib_events l × lib_events l) 
         (SODOM : so ⊆ ⋃l ∈ impl.(li_assumes), lib_events l × lib_events l) 
         (HBDOM : hb ⊆ set_collect fs g × set_collect fs g)
         (poWF: forall x y (PO: po x y) (NEQ: fs x <> fs y) z (Gz: g z)
                          (EQ1: fs z = fs x) w (Gw: g w) (EQ2: fs w = fs y), po z w)
         (IRR: acyclic (po ⨾ rf^*))
         (FS: fsupp (po ∪ rf)⁺)
         (HBACYC : acyclic ((po ∪ restr_rel g (map_rel fs hb)) ⨾ rf^*))
         (ESEM: forall a (Ga: g a) body (ENV: impl.(li_code) (ev_op (fs a)) = Some body), 
            expr_sem (body (ev_args (fs a))) (Some (RTbase (ev_res (fs a)))) 
                     (g ∩₁ (fun x => fs x = fs a))
                     (restr_rel (fun x => fs x = fs a) po))
         (LCONS: forall lib' (L: impl.(li_assumes) lib'),
            libspec_consistent lib'
               {| tup_ev := g ∩₁ lib_events lib' ;
                  tup_po := restr_rel (lib_events lib') po ;
                  tup_rf := restr_rel (lib_events lib') rf ;
                  tup_so := restr_rel (lib_events lib') so ;
                  tup_hb := restr_rel (lib_events lib') (po ∪ so ∪ restr_rel g (map_rel fs hb))⁺ |})
         (ENCAPS: lib_encapsulated lib (set_collect fs g) hb)
         (C_ALLOC: compat_alloc lib g fs)
         (C_LOC: compat_loc impl.(li_assumes) g 
                      (fun x => libspec_allocs lib (fs x) ∪₁ 
                                libspec_requires lib (fs x)) impl.(li_locf)),
      exists rf' so',
        ⟪RF': rf' ⊆ set_collect fs g × set_collect fs g⟫ /\
        ⟪SO': so' ⊆ set_collect fs g × set_collect fs g⟫ /\
        ⟪RF'': rf'⁺ ⊆ collect_rel fs (po ∪ rf)⁺⟫ /\
        ⟪SO'': so'⁺ ⊆ collect_rel fs (po ∪ so ∪ restr_rel g (map_rel fs hb))⁺⟫ /\
        ⟪FSRF': fsupp rf'⁺⟫ /\
        ⟪CONS': libspec_consistent lib
           {|
           tup_ev := set_collect fs g;
           tup_po := collect_rel fs po \ ⦗fun _ => True⦘;
           tup_rf := rf';
           tup_so := so';
           tup_hb := (collect_rel fs po \ ⦗fun _ => True⦘ ∪ so' ∪ hb)⁺ |}⟫.


Structure libimpl_verified (lib : libspec) := 
  { li_base :> libimpl_base lib ;
    li_implements_lib: 
      forall op, In op (libspec_ops lib) <-> li_base.(li_code) op <> None ;
    li_non_empty :
      forall op body (ENV: li_base.(li_code) op = Some body) args r gi poi 
             (SEM: expr_sem (body args) (Some r) gi poi), ~ gi ≡₁ ∅ ;
    li_in_assumes :
      forall op body (ENV: li_base.(li_code) op = Some body) args r gi poi 
             (SEM: expr_sem (body args) (Some r) gi poi), 
        gi ⊆₁ ⋃₁ lib0 ∈ li_base.(li_assumes), lib_events lib0 ;
    li_allocs_nemp : 
      forall a body (ENV: li_base.(li_code) (ev_op a) = Some body) g po 
             (SEM: expr_sem (body (ev_args a)) (Some (RTbase (ev_res a))) g po) 
             loc (Aa: libspec_allocs lib a loc),
        (exists libi loci, li_base.(li_locf) loc libi loci /\ li_base.(li_assumes) libi) ;
    li_allocs_uniq : 
      forall a body (ENV: li_base.(li_code) (ev_op a) = Some body) g po 
             (SEM: expr_sem (body (ev_args a)) (Some (RTbase (ev_res a))) g po) 
             loc (Aa: libspec_allocs lib a loc)
             loc' (Aa': libspec_allocs lib a loc') (NEQ: loc <> loc') libi (LIBI: li_base.(li_assumes) libi),
        set_disjoint (li_base.(li_locf) loc libi) (li_base.(li_locf) loc' libi) ;
    li_requires :
      forall a body (ENV: li_base.(li_code) (ev_op a) = Some body) gi poi
             (SEM: expr_sem (body (ev_args a)) (Some (RTbase (ev_res a))) gi poi)
             e (Ge: gi e) le (LIBe: li_base.(li_assumes) le) (Le: lib_events le e), 
        libspec_requires le e ⊆₁ 
        (⋃₁loc ∈ libspec_requires lib a, li_base.(li_locf) loc le) ∪₁
        ⋃₁e' ∈ gi ∩₁ lib_events le ∩₁ dom_rel (poi ;; <|eq e|>), libspec_allocs le e' ;
    li_allocs: 
      forall a body (ENV: li_base.(li_code) (ev_op a) = Some body) gi poi 
             (SEM: expr_sem (body (ev_args a)) (Some (RTbase (ev_res a))) gi poi) loc 
             (ALLOC: libspec_allocs lib a loc) le, 
        li_base.(li_locf) loc le ⊆₁ ⋃₁e' ∈ gi ∩₁ lib_events le, libspec_allocs le e' ;
    li_consistent : locally_sound_impl li_base
  }.

Definition good_behaviours libs A As (e: expr A As) (r : return_type A As) := 
  exists g, 
    ⟪SEM: expr_sem e (Some r) (exec_ev g) (exec_po g)⟫ /\ 
    ⟪CONS: exec_consistent g libs⟫.

Definition has_bad_behaviour libs A As (e: expr A As) := 
  exists r' g, 
    ⟪SEM: expr_sem e (Some r') (exec_ev g) (exec_po g)⟫ /\ 
    ⟪BWF: exec_basic_wf g libs⟫ /\
    ⟪NENCAP: ~ exec_encapsulated g libs⟫.

(*    ⟪PREF_CONS: 
       forall es (PREFIX: dom_rel ((exec_po g ∪ exec_rf g)⁺ ;; <| es |>) ⊆₁ es),
         exec_consistent 
           {| exec_ev := exec_ev g ∩₁ es ;
              exec_po := restr_rel es (exec_po g) ;
              exec_rf := restr_rel es (exec_rf g) ;
              exec_so := restr_rel es (exec_so g) |} libs⟫ /\ 
(exists a la b lb loc, 
      ⟪Ga : exec_ev g a⟫ /\ ⟪La: lib_events la a⟫ /\ ⟪Ua: op_requires (ev_op a) (ev_args a) loc⟫ /\
      ⟪Gb : exec_ev g b⟫ /\ ⟪Lb: lib_events lb b⟫ /\ ⟪Ub: op_requires (ev_op b) (ev_args b) loc⟫ /\ 
      ⟪LNEQ: la <> lb⟫ /\
      ⟪NBa: forall c (Gc: exec_ev g c) (Lc: lib_events lb c) (CLAIMSc: op_claims (ev_op c) (ev_args c) loc)
                   (HBac : (exec_po g ∪ exec_rf g)⁺ a c) (HBcb : (exec_po g ∪ exec_rf g)^* c b), False ⟫ /\
      ⟪NBb: forall c (Gc: exec_ev g c) (Lc: lib_events la c) (CLAIMSc: op_claims (ev_op c) (ev_args c) loc)
                   (HBac : (exec_po g ∪ exec_rf g)⁺ b c) (HBcb : (exec_po g ∪ exec_rf g)^* c a), False ⟫) \/
     exists badlib,
      ⟪Lbad: libs badlib⟫ /\
      ⟪NFW: ~ badlib.(libspec_wf)
         {| tup_ev := exec_ev g ∩₁ lib_events badlib ;
            tup_po := restr_rel (lib_events badlib) (exec_po g) ;
            tup_rf := restr_rel (lib_events badlib) (exec_rf g) ;
            tup_so := restr_rel (lib_events badlib) (exec_so g) ;
            tup_hb := restr_rel (lib_events badlib) (happens_before g) |} ⟫). *)

Definition behaviours libs A As (e: expr A As) r := 
  good_behaviours libs e r \/ has_bad_behaviour libs e.


Lemma path_not_lib lib (impl : impl_environment) 
      (OK : forall op : operation, In op (libspec_ops lib) <-> impl op <> None)
      libs (g : event -> Prop) po rf pos fs
      (poS : pos ≡ collect_rel fs po \ ⦗fun _ => True⦘)
      (poWF : forall x y,  po x y -> fs x <> fs y -> forall z, g z -> fs z = fs x -> 
              forall w, g w -> fs w = fs y -> po z w)
      (INJ : forall x, g x -> forall y, g y -> fs x = fs y -> impl (ev_op (fs x)) = None -> x = y)
      (PO : po ⊆ g × g)
      (RF : rf ⊆ (⋃l ∈ libs, (lib_events l ∩₁ g) × (lib_events l ∩₁ g)))
      x (Gx: g x) y (Gy: g y)
      (M: (pos ∪ ⦗set_compl (lib_events lib)⦘ ⨾ collect_rel fs rf ⨾ ⦗set_compl (lib_events lib)⦘)⁺ (fs x) (fs y)) :
  (po ∪ rf)⁺ x y.
Proof.
  intros; remember (fs x) as x'; revert x Heqx' Gx.
  remember (fs y) as y'; revert y Heqy' Gy.
  induction M; ins; desf; eauto using t_trans.
    hahn_rewrite poS in H; unfolder; unfolder in H; desf; clarify_not. 
    eapply t_step; left; eapply poWF; try eassumption; congruence. 
    assert (MRF:=H0); eapply RF in H0; unfolder in H0; desc. 
    eapply INJ in H2; desf; eauto.
    2: unfold lib_events in *; apply NNPP; intro FOO; apply OK in FOO; congruence.
    eapply INJ in H3; desf; eauto.
    2: unfold lib_events in *; apply NNPP; intro FOO; apply OK in FOO; congruence.
    eauto using t_step. 
  assert (exists y', fs y' = y /\ g y'); desc; eauto using t_trans.
  clear - M1 poS PO RF; apply ct_end in M1; unfolder in *; desf. 
    by eapply poS, proj1 in M0; desf; eapply PO in M0; desf; eauto.
  eapply RF in M2; desf; eauto.
Qed.

Lemma path_not_lib2 lib (impl : impl_environment) 
      (OK : forall op : operation, In op (libspec_ops lib) <-> impl op <> None)
      libs (g : event -> Prop) (po rf: relation event) fs
      (poWF : forall x y,  po x y -> fs x <> fs y -> forall z, g z -> fs z = fs x -> 
              forall w, g w -> fs w = fs y -> po z w)
      (INJ : forall x, g x -> forall y, g y -> fs x = fs y -> impl (ev_op (fs x)) = None -> x = y)
      (PO : po ⊆ g × g)
      (RF : rf ⊆ (⋃l ∈ libs, (lib_events l ∩₁ g) × (lib_events l ∩₁ g)))
      x (Gx: g x) y (Gy: g y)
      (M: (collect_rel fs po \ ⦗fun _ => True⦘ 
           ∪ restr_rel (set_compl (lib_events lib)) (collect_rel fs rf))⁺ (fs x) (fs y)) :
  (po ∪ rf)⁺ x y.
Proof.
  eapply path_not_lib; eauto.
  by hahn_rewrite <- restr_relE.
Qed.


Lemma path_rfs_other lib (impl : impl_environment) 
      (OK : forall op : operation, In op (libspec_ops lib) <-> impl op <> None)
      libs (g : event -> Prop) rf fs
      (INJ : forall x, g x -> forall y, g y -> fs x = fs y -> impl (ev_op (fs x)) = None -> x = y)
      (RF : rf ⊆ (⋃l ∈ libs, (lib_events l ∩₁ g) × (lib_events l ∩₁ g)))
      x (Gx: g x) y (Gy: g y)
      (P: (restr_rel (set_compl (lib_events lib)) (collect_rel fs rf))＊(fs x) (fs y)) :
  rf＊ x y \/ fs x = fs y. 
Proof.
  intros; remember (fs x) as x'; revert x Heqx' Gx.
  remember (fs y) as y'; revert y Heqy' Gy.
  induction P using clos_refl_trans_ind_left; ins; desf; eauto.
    unfolder in H; desf.
    specialize (RF _ _ H); unfolder in RF; desf.
    apply INJ in H3; ins; desf.
    2: by apply NNPP; rewrite <- OK, H3; ins. 
    specialize_full IHP; eauto.
    desf; eauto using clos_refl_trans.
    apply INJ in IHP; desf; eauto using clos_refl_trans.
    by apply NNPP; rewrite <- OK, IHP; ins. 
Qed.


Lemma path_not_lib' lib (impl : impl_environment) 
      (OK : forall op : operation, In op (libspec_ops lib) <-> impl op <> None)
      libs (g : event -> Prop) po rf pos fs
      (poS : pos ≡ collect_rel fs po \ ⦗fun _ => True⦘)
      (poWF : forall x y,  po x y -> fs x <> fs y -> forall z, g z -> fs z = fs x -> 
              forall w, g w -> fs w = fs y -> po z w)
      (INJ : forall x, g x -> forall y, g y -> fs x = fs y -> impl (ev_op (fs x)) = None -> x = y)
      (PO : po ⊆ g × g)
      (RF : rf ⊆ (⋃l ∈ libs, (lib_events l ∩₁ g) × (lib_events l ∩₁ g)))
      x (Gx: g x) y (Gy: g y)
      (M: (pos ⨾ (restr_rel (set_compl (lib_events lib)) (collect_rel fs rf))＊)⁺ (fs x) (fs y)) :
  (po ⨾ rf＊)⁺ x y.
Proof.
  intros; remember (fs x) as x'; revert x Heqx' Gx.
  remember (fs y) as y'; revert y Heqy' Gy.
  induction M; ins; desf; cycle 1. 
    enough (exists y', fs y' = y /\ g y') by (desc; eauto using t_trans).
    clear - M2 poS PO; hahn_rewrite poS in M2; hahn_rewrite PO in M2.
    by apply ct_begin in M2; unfolder in *; desf; eauto. 
  hahn_rewrite poS in H; unfolder; unfolder in H; desf; clarify_not.
  specialize (PO _ _ H); unfolder in PO; desc.
  eapply t_step; eapply path_rfs_other in H0; eauto.
  desf. 
    forward eapply poWF with (w := y') as POxy; eauto; congruence.
  forward eapply poWF with (w := y0) as POxy; eauto using clos_refl_trans; congruence.
Qed.

Lemma exec_basic_wf_min 
  lib libs (COMPAT : libs_compatible (libs ∪₁ eq lib)) (impl : impl_environment) 
      (OK : forall op : operation, In op (libspec_ops lib) <-> impl op <> None) deps
      (D: deps ⊆₁ libs)
      (locf : libspec_logaddr lib -> forall lib', libspec_logaddr lib' -> Prop)
      (__ALLOCSNE : 
      forall a body (ENV: impl (ev_op a) = Some body) g po 
             (SEM: expr_sem (body (ev_args a)) (Some (RTbase (ev_res a))) g po) 
             loc (Aa: libspec_allocs lib a loc),
        exists libi loci, locf loc libi loci /\ deps libi)
      (__ALLOCS : forall (a : event) (body : op_impl_type (ev_op a)),
             impl (ev_op a) = Some body ->
             forall (gi : event -> Prop) (poi : relation event),
             expr_sem (body (ev_args a)) (Some (RTbase (ev_res a))) gi poi ->
             forall loc : libspec_logaddr lib,
             libspec_allocs lib a loc ->
             forall le : libspec, locf loc le ⊆₁ (⋃₁e' ∈ gi ∩₁ lib_events le, libspec_allocs le e'))
      (g : event -> Prop) fs
      gs (gS : gs ≡₁ set_collect fs g)
      po pos (poS : pos ≡ collect_rel fs po \ ⦗fun _ => True⦘)
      (poWF : forall x y,  po x y -> fs x <> fs y -> forall z, g z -> fs z = fs x -> 
              forall w, g w -> fs w = fs y -> po z w)
      (opEQ : forall x, g x -> impl (ev_op (fs x)) = None -> fs x = x)
      (INJ : forall x, g x -> forall y, g y -> fs x = fs y -> impl (ev_op (fs x)) = None -> x = y)
      rf so 
      (ESEM0 : forall a : event,
          gs a ->
          forall body : op_impl_type (ev_op a),
            impl (ev_op a) = Some body ->
            expr_sem (body (ev_args a)) (Some (RTbase (ev_res a))) (g ∩₁ (fun x => fs x = a))
                     (restr_rel (fun x => fs x = a) po))
      (BWF : exec_basic_wf {| exec_ev := g; exec_po := po; exec_rf := rf; exec_so := so |} libs) :
  exec_basic_wf
    {|
    exec_ev := gs;
    exec_po := pos;
    exec_rf := ⦗set_compl (lib_events lib)⦘ ⨾ collect_rel fs rf ⨾ ⦗set_compl (lib_events lib)⦘;
    exec_so := ⦗set_compl (lib_events lib)⦘ ⨾ collect_rel fs so ⨾ ⦗set_compl (lib_events lib)⦘ |}
    (libs ∪₁ eq lib).
Proof.
  red in BWF; red; desc; splits; ins.
  - rewrite gS; unfolder; ins; desf; eauto.
    specialize (EV _ H); red in EV; desf.
    tertium_non_datur (impl (ev_op (fs y)) = None) as [X|X]; [by rewrite opEQ; eauto|]. 
    apply OK in X; eauto.
  - by rewrite poS, PO, gS, collect_rel_cross, inclusion_minus_rel.
  - rewrite RF; unfolder; unfold lib_events in *; ins; desf.
    tertium_non_datur (impl (ev_op (fs x')) = None) as [X|X]; [|by apply OK in X; ins].
    apply opEQ in X; ins; rewrite X in *.    
    tertium_non_datur (impl (ev_op (fs y')) = None) as [Y|Y]; [|by apply OK in Y; ins].
    apply opEQ in Y; ins; rewrite Y in *.
    exists a; splits; eauto; apply gS; red; eauto.
  - rewrite SO; unfolder; unfold lib_events in *; ins; desf.
    tertium_non_datur (impl (ev_op (fs x')) = None) as [X|X]; [|by apply OK in X; ins].
    apply opEQ in X; ins; rewrite X in *.    
    tertium_non_datur (impl (ev_op (fs y')) = None) as [Y|Y]; [|by apply OK in Y; ins].
    apply opEQ in Y; ins; rewrite Y in *.
    exists a; splits; eauto; apply gS; red; eauto.
  - intros x X.
    assert (exists y, fs y = x /\ g y); desf.
      hahn_rewrite poS in X; hahn_rewrite PO in X; hahn_rewrite RF in X.
      by clear - X; apply ct_begin in X; unfolder in *; desf; eauto. 
    by hahn_rewrite <- restr_relE in X; eapply path_not_lib' in X; eauto.
  - assert (FS': fsupp (pos ∪ ⦗set_compl (lib_events lib)⦘ ⨾ collect_rel fs rf ⨾ ⦗set_compl (lib_events lib)⦘)⁺).
    { red; ins.
      tertium_non_datur (exists y', fs y' = y /\ g y'); desf; cycle 1.
        exists nil; ins; apply H; hahn_rewrite poS in REL; apply ct_end in REL.
        by clear - PO RF REL; unfolder in *; desf; [specialize_full PO|specialize_full RF]; desf; eauto.
      specialize (FS y'); desc; exists (map fs findom); ins.
      enough (exists x', fs x' = x /\ g x') by (desf; eauto using in_map, path_not_lib). 
      hahn_rewrite poS in REL; apply ct_begin in REL.
      by clear - PO RF REL; unfolder in *; desf; [specialize_full PO|specialize_full RF]; desf; eauto. }
    done.
  - tertium_non_datur (lib0 = lib) as [|NEQ]; desf.
    {
      destruct (impl (ev_op a)) eqn: Ia; [|by apply OK in LIBa].
      destruct (impl (ev_op b)) eqn: Ib; [|by apply OK in LIBb].
      forward eapply ESEM0 with (a:=a) as S; eauto; eapply __ALLOCSNE in S; eauto; desc.
      forward eapply ESEM0 with (a:=a) as Sa; eauto; eapply __ALLOCS in Sa; eauto.
      forward eapply ESEM0 with (a:=b) as Sb; eauto; eapply __ALLOCS in Sb; eauto.
      specialize (Sa _ S); specialize (Sb _ S); unfolder in Sa; unfolder in Sb; desf.
      f_equal; eapply ALLOC; eauto.
    }
    red in ISLIB; desf. 
    assert (ALL' : 
        forall a (Ga : gs a) (L: lib_events lib0 a) (A: libspec_allocs lib0 a loc), 
          exists e',  g e' /\ fs e' = a /\ lib_events lib0 e' /\ libspec_allocs lib0 e' loc).
    { clear -gS ISLIB COMPAT ESEM0 opEQ OK NEQ; ins.
      assert (impl (ev_op a) = None).
        apply NNPP; intro X; apply OK in X.
        by destruct NEQ; eapply COMPAT; vauto.
      apply gS in Ga; destruct Ga as [x G]; desf; rewrite opEQ; ins.
      exists x; splits; ins; eauto; rewrite <- (opEQ x); ins. }
    apply ALL' in ALLa; ins; unfolder in ALLa; desf.
    apply ALL' in ALLb; ins; unfolder in ALLb; desf.
    f_equal; eapply ALLOC; ins; eauto.
Qed. 


Lemma impl_bad lib libs (COMPAT : libs_compatible (libs ∪₁ eq lib)) 
      (impl : libimpl_verified lib) (Ldeps: li_assumes impl ⊆₁ libs)
      A As (e: expr A As) (BAD: has_bad_behaviour libs (expand_fcall (li_code impl) e)) :
  has_bad_behaviour (libs ∪₁ eq lib) e.
Proof.
  unfold has_bad_behaviour in *; desc. 
  destruct impl as [[deps impl locf] OK NEMP INASS __ALLOCE __ALLOCU __REQ __ALLOCS REL']; ins.
  clear REL'.
  destruct g as [g po rf so]; ins.
  eapply expand_related in SEM; desc; ins; eauto.
  red in REL; desc.
  forward eapply exec_basic_wf_min as BWFs; eauto.
  red in BWF; desc.
  exists r'; eexists; splits; try edone; ins. 
  intro ENCAP; destruct NENCAP; red; ins.
  - destruct (impl (ev_op (fs e0))) as [body|] eqn:X; cycle 1. 
      eapply ENCAP in Ue; ins; vauto; [desf|by apply gS; exists e0; rewrite opEQ].
      apply gS in Ga; destruct Ga as [a' Ga]; desf.
      rewrite <- (opEQ e0) in HBae; ins; eapply path_not_lib in HBae; eauto; ins.
      assert (FOO: impl (ev_op (fs a')) = None).
        apply NNPP; intro Y; eapply OK in Y. 
        assert (lib = lib0) by (eapply COMPAT; vauto). 
        by desf; rewrite <- opEQ in Le; ins; eapply OK in Le.
      by exists (fs a'); splits; eauto; rewrite opEQ; ins. 
   assert (M:=ESEM0); specialize_full M; eauto; [by apply gS; vauto|].
   forward eapply __REQ with (e:=e0) as Y; try exact M; vauto.
     eapply INASS in M; unfolder in M; specialize_full M; eauto; desf; eauto.
     enough (y = lib0) by congruence.
     by eapply Ldeps in M; eapply COMPAT; vauto.
   unfolder in Y; eapply Y in Ue; clear Y; desf; cycle 1.
     by eexists; unfolder; splits; eauto using t_step.
   assert (C:=Ue); eapply ENCAP in C; ins; try apply OK; try apply gS; try congruence; desf; vauto.
   assert (exists body', impl (ev_op a) = Some body'); desc.
     by apply OK in La; destruct (impl (ev_op a)); vauto.
   eapply __ALLOCS in Aa; eauto.
   specialize (Aa _ Ue0); unfolder in Aa; desf.
   eexists; splits; eauto.
   eapply path_not_lib; try exact HBae; eauto.
Qed.

Lemma restrK A (d : A -> Prop) r : restr_rel d (restr_rel d r) ≡ restr_rel d r.
Proof.
  unfolder; intuition. 
Qed.


Lemma map_collect_rel A B (f: A -> B) r : 
  map_rel f (collect_rel f r) ≡ map_rel f ⦗fun _ => True⦘ ⨾ r ⨾ map_rel f ⦗fun _ => True⦘.
Proof.
  unfolder; split; ins; desf; eauto 8. 
Qed.

Lemma restr_seq_cond A (d : A -> Prop) r r' (DOM: domb r d \/ doma r' d) : 
  restr_rel d (r ⨾ r') ≡ restr_rel d r ⨾ restr_rel d r'.
Proof.
  split; unfolder in *; ins; desf; eauto 10. 
Qed.

Lemma restr_ct_cond A (d : A -> Prop) r (DOM: doma r d \/ domb r d) : 
  restr_rel d r⁺ ≡ (restr_rel d r)⁺.
Proof.
  split; auto using restr_ct. 
  desf; unfolder in *; ins; desf; 
    [apply clos_trans_tn1 in H | apply clos_trans_t1n in H ]; 
  induction H; ins; desf; 
  eauto 8 using clos_t1n_trans, clos_tn1_trans, clos_trans.
Qed.

Lemma restr_map_rel_seq A B (f: A -> B) d r r' 
      (DOM: domb r (set_collect f d) \/ doma r' (set_collect f d)) : 
  restr_rel d (map_rel f (r ⨾ r')) ≡ restr_rel d (map_rel f r) ⨾ restr_rel d (map_rel f r').
Proof.
  split; unfolder in *; ins; desf. 
  all: specialize_full DOM; eauto; desf; eauto 8.
Qed.


Lemma restr_map_rel_ct A B (f: A -> B) d r  
      (DOM: doma r (set_collect f d) \/ domb r (set_collect f d)) :
  restr_rel d (map_rel f r^+) ⊆ (restr_rel d (map_rel f r))^+.
Proof.
  unfolder in *; ins; desf; remember (f x) as x'; remember (f y) as y'; 
  generalize dependent x; generalize dependent y; rename H into T.
  1: apply clos_trans_tn1 in T. 
  2: apply clos_trans_t1n in T. 
  all: induction T; ins; desf; auto using t_step.
  all: specialize_full DOM; eauto; desf; 
    eauto 6 using clos_t1n_trans, clos_tn1_trans, clos_trans.
Qed.

Lemma restr_map_rel_rt A B (f: A -> B) d r  
      (DOM: doma r (set_collect f d) \/ domb r (set_collect f d)) :
  restr_rel d (map_rel f r^*) ⊆ (fun x y => f x = f y /\ d x /\ d y) +++ (restr_rel d (map_rel f r))^+.
Proof.
  rewrite !rtE, map_rel_union, restr_union, restr_map_rel_ct; ins.
  unfolder; ins; desf; eauto.
Qed.

Lemma doma_collect A B (f: A -> B) r d : doma r d -> doma (collect_rel f r) (set_collect f d).
Proof.
  unfolder; ins; desf; eauto.
Qed.

Lemma domb_collect A B (f: A -> B) r d : domb r d -> domb (collect_rel f r) (set_collect f d).
Proof.
  unfolder; ins; desf; eauto.
Qed.



Lemma impl_good lib libs (COMPAT : libs_compatible (libs ∪₁ eq lib))
      (impl : libimpl_verified lib) (Ldeps: li_assumes impl ⊆₁ libs)
      A As (e: expr A As) (NBAD: ~ has_bad_behaviour (libs ∪₁ eq lib) e) : 
  good_behaviours libs (expand_fcall (li_code impl) e) ⊆₁ 
  good_behaviours (libs ∪₁ eq lib) e.
Proof.
  unfold good_behaviours, exec_consistent, happens_before.
  intros ret H; desf; ins.
  destruct impl as [[deps impl locf] OK NEMP INASS __ALLOCE __ALLOCU 
                    __REQ __ALLOCS REL']; ins.
  destruct g as [g po rf so]; ins.
  (*assert (PO:=H); apply esem_dom in PO.*)
  eapply expand_related in SEM; desc; ins; eauto.
  red in REL; desc.

  forward eapply exec_basic_wf_min as BWFs_MIN; eauto.
  red in BWF; desf; ins.

  assert (ENCAP : exec_encapsulated
    {| exec_ev := gs;
       exec_po := pos;
       exec_rf := ⦗set_compl (lib_events lib)⦘ ⨾ collect_rel fs rf ⨾ 
                  ⦗set_compl (lib_events lib)⦘;
       exec_so := ⦗set_compl (lib_events lib)⦘ ⨾ collect_rel fs so ⨾ 
                  ⦗set_compl (lib_events lib)⦘ |}
    (libs ∪₁ eq lib)).
  {
    by apply NNPP; intro Z; apply NBAD; do 2 eexists; splits; eauto; ins; eauto.
  }

  assert (COMPATALLOC: compat_loc deps (fun x => g x /\ lib_events lib (fs x))
                                  (fun x => libspec_allocs lib (fs x)) locf).
  { red; unfolder; ins; desf.
    assert (N := IN); assert (N' := IN'). 
    destruct (impl (ev_op (fs a))) as [body|] eqn:Xa; [|by eapply OK in Xa]. 
    forward eapply ESEM0 with (a:= fs a) as Sa'; try apply gS; try red; eauto.
    eapply __ALLOCS with (a:=fs a) in IN; eauto; unfolder in IN; desf. 
    destruct (impl (ev_op (fs b))) as [body'|] eqn:Xb; [|by eapply OK in Xb].
    forward eapply ESEM0 with (a:= fs b) as Sb'; try apply gS; try red; eauto.
    eapply __ALLOCS with (a:=fs b) in IN'; eauto; unfolder in IN'; desf. 
    assert (y0 = y) by (eapply ALLOC with (lib:=libi); eauto); desf. 
    eapply __ALLOCU in NEQ; eauto; congruence.
  }
  assert (COMPATLOC: compat_loc deps (fun x => g x /\ lib_events lib (fs x))
                                (fun x => libspec_allocs lib (fs x) ∪₁ 
                                          libspec_requires lib (fs x)) locf).
  { red; unfolder; ins; desf.
    all: try (apply ENCAP in ALLa; ins; try apply gS; try red; eauto; desc; 
              clear a Ga Ga0 HBae; forward eapply (proj1 gS _) as (a & Ga); eauto; desf).
    all: try (apply ENCAP in ALLb; ins; try apply gS; try red; eauto; desc; 
              clear b Gb Gb0 HBae; forward eapply (proj1 gS _) as (b & Gb); eauto; desf).
    all: by eapply COMPATALLOC with (a:=a) (b:=b); eauto. 
  }
  clear COMPATALLOC.


  assert (LIBCONS_INT :
    forall (lib' : { lib' | deps lib'}),
    exists so_restr, 
        libspec_consistent (proj1_sig lib')
         {| tup_ev := g ∩₁ (fun x => lib_events lib (fs x)) ∩₁ lib_events (proj1_sig lib');
            tup_po := restr_rel (lib_events (proj1_sig lib'))
                        (restr_rel (fun x => lib_events lib (fs x)) po);
            tup_rf := restr_rel (lib_events (proj1_sig lib'))
                        (restr_rel (fun x => lib_events lib (fs x)) rf);
            tup_so := so_restr;
            tup_hb := restr_rel (lib_events (proj1_sig lib'))
                        (restr_rel (fun x => lib_events lib (fs x)) po ∪ 
                         restr_rel (fun x => lib_events lib (fs x)) so
                         ∪ restr_rel (g ∩₁ (fun x => lib_events lib (fs x)))
                             (map_rel fs (restr_rel (lib_events lib)
                                (collect_rel fs po \ ⦗fun _ : event => True⦘
                                 ∪ restr_rel (set_compl (lib_events lib)) (collect_rel fs so))^+)))⁺ |}  /\  so_restr <<= 
restr_rel (lib_events (proj1_sig lib')) 
(restr_rel (fun x : event => lib_events lib (fs x)) so)).
  {

  { intros [lib' L]; ins.
    assert (opEQ' : forall x, g x -> lib' <> lib -> lib_events lib' (fs x) -> fs x = x).
      ins; apply opEQ; ins.
      apply NNPP; intro X; apply OK in X; eauto.
      by eapply H0, COMPAT; vauto; unfolder; eauto.

    forward apply (LIBOK lib') as M; eauto. 
(*    assert (KK := libspec_cons_sodom _ _ M); ins. *)
    eapply libspec_locs_ok with
      (usedLocs := fun loc => ~ exists e, 
                       g e /\ lib_events lib' e /\ lib_events lib (fs e) /\ 
                       libspec_requires lib' e loc) in M; ins; desc.
      eapply libspec_monotonicity in CONSR; simpl in *.
      eexists; split.
      eapply libspec_equiv; try exact CONSR; ins; try rewrite restrC; ins.
        by unfolder; split; ins; desf; splits; ins.
        by rewrite SOR, restrC.
        rewrite SOR, !union_restr, restrC, restr_ct. 
        by apply restr_rel_mori; ins; apply clos_trans_mori; unfolder; ins; desf; eauto.

        rewrite restrC; apply restr_rel_mori; ins.
        rewrite <- restr_restr, map_rel_restr.
        rewrite restrC, restrK, restrC, ?union_restr, restr_ct. 
        arewrite (restr_rel g (map_rel fs (collect_rel fs po \ ⦗fun _ : event => True⦘
                            ∪ restr_rel (set_compl (lib_events lib)) (collect_rel fs so))⁺)
                  <<= (po ∪ so)⁺); cycle 1.
          by rewrite ct_of_union_ct_r, unionK.
        unfold map_rel; red; ins; desf.
        assert (SO': so <<= g × g).
          by rewrite SO; unfolder; clear; ins; desf.
        red in H; desf.
        hahn_rewrite <- poS in H. 
        hahn_rewrite restr_relE in H. 
        eapply path_not_lib; try exact H; eauto.
    * red; splits; ins; rewrite ?PO, ?RF, ?SO; clear; unfolder; ins; desf.
      splits; eauto; clear - H; induction H; desf; eauto.
    * unfolder in INe; desc; unfolder; split; ins; desf; clarify_not. 
         by destruct IN'; eauto.
      clarify_not; apply NNPP; intro NLIB.
      eapply H; clear H; eauto; intro X; desf.
      enough (N: exists a0, libspec_allocs lib' a0 n /\ lib_events lib' a0 /\ 
                            lib_events lib (fs a0) /\ g a0).
      {
        clear e1 X X0 X1 X2.
        specialize (LIBOK lib'); specialize_full LIBOK; eauto. 
        eapply ENCAP in REQ_NEMP; ins; unfolder; eauto; desc; cycle 1. 
          apply gS; exists e0; split; ins; apply opEQ; ins.
          by unfold lib_events in NLIB; rewrite OK in NLIB; clarify_not.
        tertium_non_datur (lib' = lib) as [|LIBNEQ]; desf.
          enough (fs e0 = e0) by congruence.
          by unfold lib_events in NLIB; rewrite OK in NLIB; clarify_not; eauto.
        apply gS in Ga; destruct Ga as (a' & Ga & ?); subst; rename a' into a.
        tertium_non_datur (a0 = fs a) as [|NEQ]; subst.
          eapply LIBNEQ, COMPAT with (op := fs a); unfolder; eauto. 
          by rewrite opEQ' with (x:=a) in N1; ins.
        eapply (libspec_cons_allocs _ _ LIBOK) in NEQ; unfolder; ins; splits; ins; eauto.
        by rewrite opEQ'.
      }
      clear e0 INe INe0 REQ_NEMP NLIB.
      destruct (impl (ev_op (fs e1))) as [body|] eqn:Xe'; [|by exfalso; eapply OK; eauto].
      forward eapply ESEM0 as Se'; try exact Xe'; [by apply gS; exists e1; ins|]. 
      forward eapply __REQ with (le:=lib') as R; eauto; vauto. 
        eapply INASS in Se'; ins; unfolder in Se'; destruct (Se' e1) as [lib1 ?]; clear Se'; desf. 
      assert (lib1 = lib'); desf.
        by eapply COMPAT; vauto; left; eapply Ldeps; eauto.
      unfolder in R; specialize_full R; try edone; desf. 
      2: eexists; splits; eauto; congruence.
      eapply ENCAP in R; ins; desc; vauto.
      2: by apply gS; red; eauto. 
      destruct (impl (ev_op a)) as [body'|] eqn:Xy; [|by exfalso; eapply OK; eauto].
      eapply __ALLOCS in R0; eauto; unfolder in R0; desf.
      eexists; splits; eauto; congruence.
    * unfolder; unfolder in INe; ins; desc.
      destruct (impl (ev_op (fs e0))) as [body|] eqn:Xe'; [|by exfalso; eapply OK; eauto].
      forward eapply ESEM0 as Se'; try exact Xe'; [by apply gS; exists e0; ins|]. 
      forward eapply __REQ with (le:=lib') as R; eauto; vauto. 
      eapply INASS in Se'; ins; unfolder in Se'; destruct (Se' e0) as [lib1 ?]; clear Se'; desf. 
      assert (lib1 = lib'); desf.
        by eapply COMPAT; vauto; left; eapply Ldeps; eauto.
      unfolder in R; specialize_full R; try edone; desf. 
      2: eexists; splits; eauto; congruence.
      eapply ENCAP in R; ins; desc; vauto.
      2: by apply gS; red; eauto. 
      destruct (impl (ev_op a)) as [body'|] eqn:Xy; [|by exfalso; eapply OK; eauto].
      eapply __ALLOCS in R0; eauto; unfolder in R0; desf.
      eexists; splits; eauto; congruence.
    }
  }

  assert (LIBCONS' :
    exists so_restr, so_restr <<= restr_rel (fun x : event => lib_events lib (fs x)) so /\ 
      forall lib' (L: deps lib'),
        libspec_consistent lib'
         {| tup_ev := g ∩₁ (fun x => lib_events lib (fs x)) ∩₁ lib_events lib';
            tup_po := restr_rel (lib_events lib')
                        (restr_rel (fun x => lib_events lib (fs x)) po);
            tup_rf := restr_rel (lib_events lib')
                        (restr_rel (fun x => lib_events lib (fs x)) rf);
            tup_so := restr_rel (lib_events lib') so_restr;
            tup_hb := restr_rel (lib_events lib')
                        (restr_rel (fun x => lib_events lib (fs x)) po ∪ so_restr
                         ∪ restr_rel (g ∩₁ (fun x => lib_events lib (fs x)))
                             (map_rel fs (restr_rel (lib_events lib)
                                (collect_rel fs po \ ⦗fun _ : event => True⦘
                                 ∪ restr_rel (set_compl (lib_events lib)) (collect_rel fs so))^+)))⁺ |}).
  { ins.
    eapply choice in LIBCONS_INT; desc.
    exists (restr_rel (fun x => lib_events lib (fs x)) (⋃lib', f lib')); ins.
    split; ins.
      rewrite inclusion_restr.
      by unfolder; ins; desc; apply (LIBCONS_INT a); ins.
    assert (M:=LIBCONS_INT (exist _ lib' L)); ins; desc.
    assert (fEQ:  f (exist deps lib' L)
                    ≡ restr_rel (lib_events lib')
                    (restr_rel (fun x => lib_events lib (fs x)) (⋃lib'0, f lib'0))).
    { split.
        unfolder; unfolder in M0; ins. 
        specialize (M0 _ _ H); desc; splits; ins; eexists; split; eauto.
      unfolder; ins; desc. 
      enough (proj1_sig a = lib'). 
        destruct a; ins; desf.
        by enough (d = L); desf; apply proof_irrelevance. 
      apply LIBCONS_INT in H4; unfolder in H4; ins; desc.
      destruct a as [lib'' D]; ins.
      eapply COMPAT; eauto; unfolder; eauto.
    }
    eapply libspec_monotonicity in M; [eapply libspec_equiv; try exact M| |]; ins. 
    * rewrite fEQ, union_restr, restr_ct; apply restr_rel_mori; ins.
      by apply clos_trans_mori; unionL; ins; auto with hahn.
    * arewrite ((⋃lib'0, f lib'0) <<= so); ins.
      by unfolder; ins; desc; apply (LIBCONS_INT a); ins.
  }
  clear LIBCONS_INT; desc.

  specialize (REL' (g ∩₁ (fun x => lib_events lib (fs x)))
                   (restr_rel (fun x => lib_events lib (fs x)) po)
                   (restr_rel (fun x => lib_events lib (fs x)) rf)
                   so_restr
                   (restr_rel (lib_events lib)
                     (collect_rel fs po \ ⦗fun _ : event => True⦘
                      ∪ restr_rel (set_compl (lib_events lib)) (collect_rel fs so))^+)
                   fs).
  specialize_full REL'; eauto; desc.
    by unfolder; ins; desf.
    by rewrite PO; clear; unfolder; ins; desf.
    by rewrite RF; clear; unfolder; ins; desf.
  { rewrite RF; unfolder; ins; desf.
    destruct (impl (ev_op (fs x))) as [body|] eqn:Xx; [|by exfalso; eapply OK, Xx].
    forward apply ESEM0 as S; eauto; try apply gS; try red; eauto. 
    eapply INASS in S; eauto. 
    forward eapply (S x) as S'; ins; unfolder in S'; desc.
    enough (y0 = a) by (desf; eauto).
    eapply COMPAT with (op := x); unfolder; auto.
  }
  { rewrite LIBCONS', SO; unfolder; ins; desf.
    eexists; splits; eauto.
    destruct (impl (ev_op (fs x))) as [body|] eqn:Xx; [|by exfalso; eapply OK, Xx].
    forward apply ESEM0 as S; eauto; try apply gS; try red; eauto. 
    eapply INASS in S; eauto. 
    forward eapply (S x) as S'; ins; unfolder in S'; desc.
    enough (y0 = a) by (desf; eauto).
    eapply COMPAT with (op := x); unfolder; auto.
  }
  {
    arewrite (so ⊆ g × g) by (rewrite SO; unfolder; ins; desf).
    rewrite PO, inclusion_restr with (dom := set_compl _). 
    rewrite union_absorb_l, collect_rel_cross; eauto with hahn. 
    rewrite ct_of_cross; unfolder; ins; desf; splits; eauto.
  }
    by unfolder; ins; desf; splits; eauto.
    by rewrite !inclusion_restr. 
    by rewrite !inclusion_restr. 
  {
    clear REL'.
    eapply irreflexive_inclusion, IRR; apply inclusion_t_t2.
    rewrite <- restr_restr, !inclusion_restr with (dom := fun x => lib_events lib (fs x)).
    rewrite ct_begin with (r := _ +++ _); relsf.
    unionL; eauto with hahn.
    rewrite restr_union.
    arewrite 
      (fun x y => restr_rel (lib_events lib) (restr_rel (set_compl (lib_events lib)) x ⨾ y) ⊆ ∅₂).
      unfolder; ins; desf.
    rels.
    rewrite rt_unionE, <- ?seqA, <- ct_begin.
    rewrite ct_end with (r := _ ;; (restr_rel _ _) ^* ).
    rewrite rtE with (r := restr_rel _ _) at 2; relsf.
    rewrite !restr_union.
    arewrite 
      (restr_rel (lib_events lib)
             (((collect_rel fs po \ ⦗fun _ : event => True⦘)
               ⨾ (restr_rel (set_compl (lib_events lib)) (collect_rel fs so))＊)＊
              ⨾ (collect_rel fs po \ ⦗fun _ : event => True⦘)
                ⨾ (restr_rel (set_compl (lib_events lib)) (collect_rel fs so))⁺) ⊆ ∅₂).
      by rewrite ct_end; unfolder; ins; desf.
    rels.

    arewrite ((restr_rel (set_compl (lib_events lib)) (collect_rel fs so))＊
              ⊆ (restr_rel (set_compl (lib_events lib)) (collect_rel fs so⁺))^?).
    {  rewrite rtE, crE; unionL; auto with hahn.
      unfolder; ins; desf.
      right.
      rename H into T; induction T; desf; splits; eauto 6 using clos_trans.
      enough (x' = y'0) by (desf; eauto 7 using clos_trans).
      unfold lib_events in *; rewrite OK in *; clarify_not.
      rewrite <- IHT4 in IHT7.
      rewrite <- (opEQ y'0), <- (opEQ x'); ins. 
      by hahn_rewrite SO in IHT2; unfolder in IHT2; clear - IHT2; induction IHT2; desf; eauto.
      by hahn_rewrite SO in IHT1; unfolder in IHT1; clear - IHT1; induction IHT1; desf; eauto.
    }
    arewrite (so⁺ ⊆ (po ∪ rf)⁺).
    {
      arewrite (so ⊆ ⋃l ∈ libs, restr_rel (lib_events l) so).
        by clear -SO; unfolder in *; ins; desf; edestruct SO; eauto; desf; eauto.
  
      arewrite ((⋃x ∈ libs, restr_rel (lib_events x) so) ⊆ (po ∪ rf)⁺); rels.
        intros x y [l [X Y]].
        specialize (LIBOK _ X); apply libspec_cons_sodom in LIBOK; ins.
        apply (inclusion_trans LIBOK); ins.
        by rewrite !inclusion_restr.
    }
    rewrite !inclusion_restr with (dom := lib_events lib).
    rewrite !inclusion_restr with (dom := set_compl (lib_events lib)).

    assert (L : (fun x y => fs x = fs y /\ g x /\ g y)
               ⨾ restr_rel g (map_rel fs (collect_rel fs po \ ⦗fun _ : event => True⦘)) ⊆
               restr_rel g (map_rel fs (collect_rel fs po \ ⦗fun _ : event => True⦘))).
    { unfolder; ins; desf; clarify_not; splits; ins. 
        repeat eexists; eapply poWF; eauto; congruence.
      intro; desf; congruence. }
    assert (R : restr_rel g (map_rel fs (collect_rel fs po \ ⦗fun _ : event => True⦘)) ⨾ 
                (fun x y => fs x = fs y /\ g x /\ g y) ⊆
               restr_rel g (map_rel fs (collect_rel fs po \ ⦗fun _ : event => True⦘))).
    { unfolder; ins; desf; clarify_not; splits; ins. 
        repeat eexists; eapply poWF; eauto; congruence.
      intro; desf; congruence. }
    assert (M : restr_rel g (map_rel fs (collect_rel fs po \ ⦗fun _ => True⦘)) ⊆ po).
      unfolder; ins; desf; clarify_not; eapply poWF; eauto; congruence.
  
    rewrite restr_map_rel_seq; eauto with hahn.
    2: by rewrite PO, inclusion_minus_rel; rels.
    rewrite restr_map_rel_rt; eauto with hahn.
    2: by rewrite PO, inclusion_minus_rel; rels.
    relsf.
    rewrite L.
    unionL; eauto with hahn. 
    rewrite restr_map_rel_seq; eauto with hahn.
    2: by rewrite PO, inclusion_minus_rel; rels.
  
  
    assert (DOM: po ∪ rf ⊆ g × g) by (rewrite PO, RF; unfolder; ins; desf).
  
    rewrite crE, map_rel_union, restr_union.
    rewrite map_collect_rel. 
    rewrite !restr_seq_cond; try (by rewrite DOM; eauto with hahn).
    arewrite (restr_rel g (map_rel fs ⦗fun _ => True⦘) ≡ (fun x y => fs x = fs y /\ g x /\ g y)) 
      by (unfolder; splits; ins; desf). 
    relsf; sin_rewrite !R.
    arewrite
      (restr_rel g (map_rel fs (collect_rel fs po \ ⦗fun _ => True⦘))
       ∪ restr_rel g (map_rel fs (collect_rel fs po \ ⦗fun _ => True⦘))
         ⨾ restr_rel g (po ∪ rf)⁺ ⨾ (fun x y : event => fs x = fs y /\ g x /\ g y) ⊆ 
       restr_rel g (map_rel fs (collect_rel fs po \ ⦗fun _ => True⦘)) ⨾
         (po ∪ rf)^* ⨾ (fun x y => fs x = fs y /\ g x /\ g y)).
      rewrite rtE; relsf; apply union_mori.
      1-2: unfolder; ins; desf; eexists; splits; eauto.
    pattern_lhs 
      (restr_rel g (map_rel fs (collect_rel fs po \ ⦗fun _ : event => True⦘))
                 ⨾ (po ∪ rf)＊ ⨾ (fun x y : event => fs x = fs y /\ g x /\ g y))⁺. 
    apply ct_ind_right; eauto with hahn. 
    ins; rewrite ?seqA; sin_rewrite L; rewrite M, rt_unionE, ?seqA; rels.
      rewrite <- seqA, inclusion_t_rt at 1; rels.
    ins; rewrite ?seqA; sin_rewrite L. 
    rewrite <- inclusion_id_rt in H at 1.
    rewrite seq_id_r in H; sin_rewrite H. 
    rewrite M, rt_unionE, ?seqA; rels.   
    rewrite ct_end at 1; rewrite ?seqA; rels.
    arewrite (fun x => x ⨾ po ⨾ rf＊ <<= x ⨾ (po ⨾ rf＊)＊); rels.
  }
  { ins; red in Ga; desc.
    rewrite restr_restr, set_interA.
    rewrite set_inter_absorb_l with (s := fun x => fs x = fs a).
    eapply ESEM0; ins; apply gS; red; eauto.
    unfolder; ins; congruence.
  }
  {
    red; ins; desf.
    eapply ENCAP in Ue; ins; vauto; cycle 1.
      by apply gS; revert Ge; clear; unfolder; ins; desf; eauto.
    desf; eexists; splits; eauto.
      by apply gS in Ga; revert Ga; unfolder; ins; desf; eauto.
    hahn_rewrite <- restr_relE in HBae.
    by hahn_rewrite <- poS; ins.
  }
  { red; unfolder; ins; desf; eauto.
    eapply BWFs_MIN; eauto; ins; vauto.
    all: apply gS; red; eauto.
  }
  clear COMPATLOC.
  rewrite LIBCONS' in *; clear so_restr LIBCONS' LIBCONS'0.

  ins; desf.
  set (R := restr_rel (g ∩₁ (fun x : event => lib_events lib (fs x))) _)
    in SO''.
  assert (LEM: R <<= (po ∪ so)⁺). 
  { subst R; red; ins; desf.
    unfolder in H; desf; eapply path_not_lib2 with (fs := fs) (lib:=lib); eauto. 
  }

  assert (DOM: forall x y, (po ∪ rf)⁺ x y -> g x /\ g y).
  {
    clear -PO RF; unfolder in *; induction 1; desf; eauto.
    firstorder.
  }

  assert (NN': forall x' y' (M: (rf'⁺ ⨾ (pos ⨾ (restr_rel (set_compl (lib_events lib)) (collect_rel fs rf))＊)⁺)⁺
                                  x' y'),
          exists x, g x /\ fs x = x' /\ 
             << IMP: forall y (G: g y) (EQ: fs y = y'), (rf＊ ;; (po ;; rf＊)⁺) x y >>).
  {
    ins; induction M; ins; desf.
      red in H; desc.
      apply RF'' in H; unfolder in H; desf; clarify_not.
      apply clos_trans_mon with (r':= po ∪ rf) in H.
      2: by unfolder; ins; desf; eauto using t_step.
      exists x'; splits; ins; desf.
        by apply DOM in H; desf.
      red; eapply path_not_lib' in H0; eauto using t_trans.
        apply clos_trans_in_rt, rt_unionE in H; red in H; desf.
        by eexists; split; eauto; apply rt_ct; eexists; eauto.
      by apply DOM in H; desf. 
    eexists x1; splits; ins; desf.
    specialize_full IMP0; try apply eq_refl; eauto; red in IMP0; desc.
    eexists; split; eauto.
    specialize_full IMP; try apply eq_refl; eauto; red in IMP; desc. 
    eapply ct_end in IMP1; unfolder in IMP1; desf.
    eapply rt_ct; eexists; split; eauto.
    eapply t_trans; eauto; apply t_step; unfolder; eauto using clos_refl_trans. 
  }

  assert (LIBOK_INT : 
            forall lib' : {lib' | (libs ∪₁ eq lib) lib' }, 
              exists so_restr,
              libspec_consistent (proj1_sig lib')
                 {| tup_ev := gs ∩₁ lib_events (proj1_sig lib');
                    tup_po := restr_rel (lib_events (proj1_sig lib')) pos;
                    tup_rf := restr_rel (lib_events (proj1_sig lib'))
                       (restr_rel (set_compl (lib_events lib)) (collect_rel fs rf) ∪ rf');
                    tup_so := so_restr ; 
                    tup_hb := restr_rel (lib_events (proj1_sig lib'))
                       (pos ∪ (restr_rel (set_compl (lib_events lib)) (collect_rel fs so) ∪ so'))⁺ |} 
              /\ so_restr <<= restr_rel (lib_events (proj1_sig lib'))
                           (restr_rel (set_compl (lib_events lib)) (collect_rel fs so) ∪ so')).
  {
    intros [lib' ISLIB]; ins; tertium_non_datur (lib' = lib) as [|LIBNEQ]; desf.
      eexists so'; split; ins; cycle 1.
        by unfolder; split; ins; desf; splits; ins; auto; eapply SO' in H; unfolder in H; desf.
        eapply libspec_monotonicity in CONS'; [eapply libspec_equiv; eauto| |]; ins;
        rewrite ?poS, ?restr_union, ?restr_restr, ?union_false_l; ins.
        by rewrite gS; unfolder; splits; ins; desf; eauto.
        by unfolder; split; ins; desf; splits; ins; eauto 8.
        arewrite (set_compl (lib_events lib) ∩₁ lib_events lib ≡₁ ∅) by (clear; firstorder). 
        rewrite restr_relE; rels.
        by unfolder; split; ins; desf; splits; ins; eapply RF' in H; unfolder in H; desf. 
      {
        etransitivity; [|apply restr_ct].
        rewrite !restr_union.
        apply clos_trans_mori; ins; unionL. 
          by unionR left; unfolder; ins; desf; splits; eauto.
        unionR right -> right; unfolder; ins; desf; eauto.
        by unfolder in SO'; clear - H SO'; specialize_full SO'; 
           eauto using t_step; desf. }
      { ins; rewrite <- unionA, <- ct_of_union_ct_r. 
        etransitivity; cycle 1.
          rewrite unionA.
          apply clos_trans_mori, inclusion_union_r2.
        red; ins; red in H; desc.
        assert (Hy: exists y', g y' /\ fs y' = y); desf.
        { hahn_rewrite PO in H; hahn_rewrite SO' in H; hahn_rewrite SO in H.
          eapply ct_end in H; unfolder in H; desf; eauto.
          eapply ct_end in H2; unfolder in H2; desf; eauto. }
        assert (Hx: exists x', g x' /\ fs x' = x); desf.
        { hahn_rewrite PO in H; hahn_rewrite SO' in H; hahn_rewrite SO in H.
          eapply ct_begin in H; unfolder in H; desf; eauto.
          eapply ct_begin in H; unfolder in H; desf; eauto. }
        assert (DOM': forall x y, (po ∪ so)⁺ x y -> g x /\ g y).
        { clear -PO SO; unfolder in *; induction 1; desf; eauto; firstorder. }
        apply path_ut2 in H; ins. 
        destruct H as [B|(z & B & w & C & D)].
          by apply t_step; right; ins. 
        assert (Hw: exists w', g w' /\ fs w' = w /\ lib_events lib (fs w') /\
            (so'
               ∪ restr_rel (lib_events lib)
               (collect_rel fs po \ ⦗fun _ : event => True⦘
                ∪ restr_rel (set_compl (lib_events lib)) (collect_rel fs so))⁺)⁺ 
              (fs w') (fs y')).
        { clear B C; destruct D as (m & B & C).
          specialize (SO'' _ _ B); unfolder in SO''; desf; clarify_not.
          exists x'0; splits; ins.
            by eapply clos_trans_mon,  (proj1 (ct_of_ct _)), DOM' in SO''; 
               unfolder; ins; desf; eauto using t_step.
            hahn_rewrite SO' in B; apply ct_begin in B; unfolder in B; desf; congruence.
          apply rtE in C; destruct C as [(C & _)|C]. 
            by rewrite C in *; eapply clos_trans_mon; eauto; unfolder; ins; desf; eauto.
          eapply t_trans.
            by eapply clos_trans_mon; eauto; unfolder; ins; desf; eauto.
          eapply t_step; right; split; splits; ins.
          hahn_rewrite SO' in B; apply ct_end in B; unfolder in B; desf; congruence. }
        assert (Hz: exists z', g z' /\ fs z' = z); desf.
        { hahn_rewrite PO in B; hahn_rewrite SO in B. 
          eapply rt_end in B; unfolder in B; desf; eauto. }
        assert (LIBz: lib_events lib (fs z')).
        { apply rt_begin in C; destruct C as [(-> & _)|(? & C & _)]; ins.
          hahn_rewrite SO' in C; hahn_rewrite ct_begin in C.
          clear -C; unfolder in C; ins; desf; congruence. }
        desf; clear D H1 Hy; apply rt_t_trans with (fs w'); eauto.
        clear Hw2.
        apply rt_trans with (fs z').
          by apply rtE in B; destruct B as [(-> & _)|B]; vauto.
        clear B. 
        apply rtE in C; destruct C as [(-> & _)|C]; vauto.
        eapply clos_trans_in_rt.
        revert C Hw1 LIBz.
        generalize (collect_rel fs po \ ⦗fun _ : event => True⦘
        ∪ restr_rel (set_compl (lib_events lib)) (collect_rel fs so)) as r.
        induction 1; ins.
          red in H; desf; ins; desf.
          eapply t_trans, t_step. 
            by eapply clos_trans_mon; eauto; vauto.
          right; split; splits; ins. 
          by hahn_rewrite SO' in H; apply ct_end in H; unfolder in H; desf.
        enough (lib_events lib y) by eauto using t_trans.
        hahn_rewrite SO' in C2. 
        apply ct_begin in C2; destruct C2 as [? [C2 _]]; unfolder in C2; desf. 
        apply ct_begin in C2; destruct C2 as [? [C2 _]]; unfolder in C2; desf.
      }

    red in ISLIB; desf; clear CONS'; specialize (LIBOK _ ISLIB).
    assert (LIBDISJ': forall x, lib_events lib' x -> lib_events lib x -> False).
      by ins; apply LIBNEQ; symmetry; eapply COMPAT; vauto.
    
    assert (opEQ' : forall x, g x -> lib_events lib' (fs x) -> fs x = x).
      ins; apply opEQ; ins.
      by apply NNPP; intro X; apply OK in X; eauto.
   
    assert (exists es usedLocs, 
               << MEQ : forall x, es x <-> lib_events lib' (fs x) >> /\
               << FL : g ∩₁ lib_events lib' ∩₁ es ≡₁ gs ∩₁ lib_events lib' >> /\
               << REQ : forall e (INe: (g ∩₁ lib_events lib') e) (REQ_NEMP : ~ libspec_requires lib' e ≡₁ ∅), 
                   es e <-> set_disjoint (libspec_requires lib' e) usedLocs >> /\
               << WFLOC : forall e (INe : (g ∩₁ lib_events lib') e) (INe : es e)
                               loc (REQ: libspec_requires lib' e loc),
                   exists a,
                     ⟪Aa: libspec_allocs lib' a loc⟫ /\ ⟪Ga: gs a⟫ /\ ⟪Ga: lib_events lib' a⟫ /\ ⟪RESTRa: es a⟫ >>). 
    {
      exists (fun e => lib_events lib' (fs e)).
      exists (fun loc => exists e, 
                  g e /\ lib_events lib' e /\ lib_events lib (fs e) /\ 
                  libspec_requires lib' e loc).
      splits; ins.
      * by rewrite gS; unfolder; split; ins; desf; splits; eauto; rewrite (opEQ' y).
      * unfolder in INe; desc; unfolder; split; ins; desf; cycle 1.
          tertium_non_datur (impl (ev_op (fs e0)) = None) as [X|X]; ins.
            by rewrite opEQ; ins. 
          by apply OK in X; ins; exfalso; eapply H; eauto.
        enough (M: exists a0, libspec_allocs lib' a0 x /\ lib_events lib' a0 /\ lib_events lib (fs a0) /\ g a0).
        { eapply ENCAP in IN; ins; vauto; desc; [|by apply gS; red; eauto]. 
          apply gS in Ga; destruct Ga as (a' & Ga & ?); subst; rename a' into a.
          tertium_non_datur (a0 = fs a) as [|NEQ]; subst.
            destruct LIBNEQ; eapply COMPAT with (op := fs a); vauto.
            by rewrite opEQ' with (x:=a) in M1; ins.
          eapply (libspec_cons_allocs _ _ LIBOK) in NEQ; unfolder; ins; splits; ins; eauto.
          by rewrite opEQ'. }
        clear e0 INe INe0 REQ_NEMP H IN.
        destruct (impl (ev_op (fs e1))) as [body|] eqn:Xe'; [|by exfalso; eapply OK; eauto].
        forward eapply ESEM0 as Se'; try exact Xe'; [by apply gS; exists e1; ins|]. 
        forward eapply __REQ with (le:=lib') as M; eauto; vauto. 
          eapply INASS in Se'; ins; unfolder in Se'; destruct (Se' e1) as [lib1 ?]; clear Se'; desf. 
          enough (lib1 = lib') by desf.
          by eapply COMPAT; vauto; left; eapply Ldeps; eauto.
        unfolder in M; specialize_full M; try edone; desf. 
        2: eexists; splits; eauto; congruence.
        eapply ENCAP in M; ins; desc; vauto.
        2: by apply gS; red; eauto. 
        destruct (impl (ev_op a)) as [body'|] eqn:Xy; [|by exfalso; eapply OK; eauto].
        eapply __ALLOCS in M0; eauto; unfolder in M0; desf.
        eexists; splits; eauto; congruence.
     * unfolder in INe; desc.
       eapply ENCAP in REQ; ins; vauto; desc.
       2: by apply gS; eexists; splits; eauto.
       exists a; splits; ins.
       apply gS in Ga; destruct Ga as [a' Ga]; desf.
       by rewrite opEQ' with (x := a'); ins.
    }
    desc.
    eapply libspec_locs_ok with (es:=es) (usedLocs := usedLocs) in LIBOK; 
      ins; desf; eauto; cycle 1.
      { red; ins; splits.
        by rewrite PO; unfolder; ins; desf.
        by rewrite RF; unfolder; ins; desf.
        by rewrite SO; unfolder; ins; desf.
        by rewrite PO, SO; unfolder; ins; splits; desf;
           clear -H; induction H; desf; eauto.
      }
      by specialize_full WFLOC; eauto; desc; eexists; splits; eauto; eapply FL; split; ins.
   
    eapply libspec_monotonicity in CONSR; simpl in *.
    eexists; split. 
      eapply libspec_equiv; try apply CONSR; ins; ins.
   
    * arewrite (po <--> restr_rel g po).
        by clear - PO; unfolder in *; firstorder.
      rewrite !restr_restr, <- set_interA, FL.
      arewrite (pos <--> <| gs |> ;; pos ;; <| gs |>).
      { rewrite gS, <- restr_relE, poS.
        unfolder; split; ins; desf; splits; ins; clarify_not; eauto. 
        all: apply PO in H; red in H; desf; eauto. }
      rewrite !restr_relE, gS, <- !restr_relE, restr_restr, poS. 
      unfolder; unfolder in IRR; intuition; desf; eauto using rt_refl, t_step.
      by rewrite !opEQ' in H0; ins; eauto.
      by unfolder in PO; specialize_full PO; eauto; desf; rewrite !opEQ'.
   
    * rewrite gS in FL.
      arewrite (rf <--> restr_rel g rf).
        by clear - RF; unfolder in *; firstorder.
      rewrite !restr_restr, <- set_interA, FL.
      rewrite restr_union, restr_restr.
      arewrite_false (restr_rel (lib_events lib') rf'); rels.
        by rewrite RF'; unfolder; ins; desf; eauto.
      unfolder; split; ins; desf; splits; ins; eauto.
      by rewrite !opEQ' in H; ins; eauto 8.
      by rewrite !opEQ'.

    * rewrite gS in FL.
      rewrite SOR. 
      arewrite (so <--> restr_rel g so).
        by clear - SO; unfolder in *; firstorder.
      rewrite !restr_restr, <- set_interA, FL.
      rewrite restr_union, restr_restr.
      rewrite <- inclusion_union_r1.
      unfolder; split; ins; desf; splits; ins; eauto.
      by rewrite !opEQ' in H; ins; eauto 8.

    * rewrite poS, SOR.
      rewrite !union_restr, restr_restr.
      unfolder; ins; desf; eauto.
      assert (X:=clos_trans_restrD H); desc; splits; ins. 
      clear X X0 X1 X2; induction H; eauto using clos_trans.
      rewrite !MEQ in *; desf; apply t_step; desf. 
        specialize (PO _ _ H); red in PO; desc.
        left; splits; try red; ins; desf; eauto 8.
        by eapply IRR, t_step; red; eauto using rt_refl.
      specialize (SO _ _ H); unfolder in SO; desc.
      by right; left; splits; eauto 8.
   
    * assert (DOM': forall x y, (po ∪ so)⁺ x y -> g x /\ g y).
      {
        clear -PO SO; unfolder in *; induction 1; desf; eauto.
        firstorder.
      }
    clear NN'.
    assert (NN': forall x' y' 
                     (M: (so'⁺ ⨾ (pos ∪ restr_rel (set_compl (lib_events lib)) (collect_rel fs so))⁺)⁺ x' y'),
            exists x, g x /\ fs x = x' /\ 
               << IMP: forall y (G: g y) (EQ: fs y = y'), (po ∪ so)⁺ x y >>).
    {
      ins; induction M; ins; desf.
        red in H; desc.
        apply SO'' in H; unfolder in H; desf; clarify_not.
        eapply clos_trans_mon with (r':= (po ∪ so)^+), (proj1 (ct_of_ct _)) in H.
        2: by unfolder; ins; desf; eauto using t_step.
        exists x'; splits; ins; desf.
          by apply DOM' in H; desf. 
        hahn_rewrite poS in H0; eapply path_not_lib2 in H0; eauto using t_trans.
        by apply DOM' in H; desf. 
      eexists x1; splits; ins; eauto using t_trans.
    }
   
    rewrite <- unionA.
      red; ins; desf; unfolder; red in H; desc.
      assert (Hy: exists y', g y' /\ fs y' = y); desf.
      { hahn_rewrite poS in H; hahn_rewrite PO in H; hahn_rewrite SO' in H; hahn_rewrite SO in H.
        eapply ct_end in H; unfolder in H; desf; eauto; exfalso; eauto. }
      assert (Hx: exists x', g x' /\ fs x' = x); desf.
      { hahn_rewrite poS in H; hahn_rewrite PO in H; hahn_rewrite SO' in H; hahn_rewrite SO in H.
        eapply ct_begin in H; unfolder in H; desf; eauto; exfalso; eauto. }
     splits; eauto.
     all: try by rewrite opEQ'; ins; apply MEQ.
     rewrite !opEQ'; ins.
     eapply ct_of_union_ct_r in H; ins.
     eapply path_ut2 in H; ins.
     destruct H as [B|(z & B & w & C & D)].
       by hahn_rewrite poS in B; eapply path_not_lib2; try exact B; eauto. 
    assert (Hz: exists z', g z' /\ fs z' = z); desf.
      { hahn_rewrite poS in B; hahn_rewrite PO in B; hahn_rewrite SO in B. 
        eapply rt_end in B; unfolder in B; desf; eauto. }
    assert (Hw: exists w', g w' /\ fs w' = w /\ (po ∪ so)⁺ w' y').
      { clear B C; destruct D as (m & B & C).
        assert (B':=B). 
        eapply SO'' in B; unfolder in B; desf.
        eapply clos_trans_mon with (r':= (po ∪ so)⁺), (proj1 (ct_of_ct _)) in B.
        2: by unfolder; ins; desf; eauto using t_step.
        exists x'0; splits; ins.
          by eapply DOM' in B; desf.
        apply rtE in C; destruct C as [(C & _)|C]. 
          destruct LIBNEQ; eapply COMPAT; vauto; try congruence.
          apply ct_end in B'; destruct B' as (? & _ & B'); 
          apply SO' in B'; unfolder in B'; desf; congruence.
        hahn_rewrite poS in C; eapply path_not_lib2 with (g:=g) in C; eauto using t_trans.
        by eapply DOM' in B; desf. }
    desf; clear D H1 Hy; apply rt_t_trans with w'; eauto.
    clear Hw1.
    apply rtE in C; destruct C as [(C & _)|C].
      rewrite C in B.
      apply rtE in B; destruct B as [(B & _)|B].
        enough (x' = w') by (desf; eauto using rt_refl).
        rewrite <- (opEQ' w'), <- (opEQ' x'); ins; congruence.
      by hahn_rewrite restr_relE in B; eapply path_not_lib in B; eauto using clos_trans_in_rt.
    eapply NN' in C; desf.
    rewrite <- C0 in B.
    apply rtE in B; destruct B as [(B & _)|B].
      enough (x' = x) by (desf; eauto using clos_trans_in_rt).
      rewrite <- (opEQ' x), <- (opEQ' x'); ins; congruence.
    by hahn_rewrite restr_relE in B;eapply path_not_lib in B; eauto using clos_trans_in_rt, t_trans.
  }

  assert (LIBOK' :
    exists so_restr, so_restr ⊆ restr_rel (set_compl (lib_events lib)) (collect_rel fs so) ∪ so' /\ 
      forall lib' (ISLIB : (libs ∪₁ eq lib) lib'), 
        libspec_consistent lib'
          {| tup_ev := gs ∩₁ lib_events lib';
             tup_po := restr_rel (lib_events lib') pos;
             tup_rf := restr_rel (lib_events lib')
                                 (restr_rel (set_compl (lib_events lib)) (collect_rel fs rf) ∪ rf');
             tup_so := restr_rel (lib_events lib') so_restr ; 
             tup_hb := restr_rel (lib_events lib')
                       (pos ∪ (restr_rel (set_compl (lib_events lib)) (collect_rel fs so) ∪ so'))⁺ |}).
  { ins; eapply ClassicalChoice.choice in LIBOK_INT; desc.
    exists (⋃lib', f lib'); ins.
    split; ins.
      by unfolder; ins; desc; apply (LIBOK_INT a); ins.
    assert (M:=LIBOK_INT (exist _ lib' ISLIB)); ins; desc.
    assert (fEQ:  f (exist _ lib' ISLIB)
                    ≡ restr_rel (lib_events lib') (⋃lib'0, f lib'0)).
    { split.
        unfolder; unfolder in M0; ins. 
        specialize (M0 _ _ H); desc; splits; ins; desf; eauto. 
      unfolder; ins; desc. 
      enough (proj1_sig a = lib'). 
        destruct a; ins; subst.
        by enough (o = ISLIB); desf; apply proof_irrelevance. 
      apply LIBOK_INT in H2; unfolder in H2; ins; desc.
      destruct a as [lib'' D]; ins.
      eapply COMPAT; eauto; unfolder; eauto.
    }
    eapply libspec_monotonicity in M; [eapply libspec_equiv; try exact M| |]; ins. 
    rewrite M0, union_restr, restr_ct; apply restr_rel_mori; ins.
  }
  clear LIBOK_INT; desc.

  assert (BWFs : exec_basic_wf
    {| exec_ev := gs;
       exec_po := pos;
       exec_rf := restr_rel (set_compl (lib_events lib)) (collect_rel fs rf) ∪ rf';
       exec_so := so_restr |}
    (libs ∪₁ eq lib)).
  { red; red in BWFs_MIN; desc; splits; ins. 
  - rewrite RF, RF'; unionL; 
    unfold lib_events in *; unfolder; ins; desf. 
    rewrite OK in *; ins; clarify_not.
    by exists a; splits; eauto; try (apply gS; eexists; eauto); rewrite opEQ; ins. 
    by exists lib; splits; eauto; try (apply gS; eexists; eauto). 
  - rewrite LIBOK', SO, SO'; unionL; 
    unfold lib_events in *; unfolder; ins; desf. 
    rewrite OK in *; ins; clarify_not.
    by exists a; splits; eauto; try (apply gS; eexists; eauto); rewrite opEQ; ins. 
    by exists lib; splits; eauto; try (apply gS; eexists; eauto). 
  - apply acyclic_seq_union_incompat.
    1-2: rewrite RF'; unfolder; ins; desf; congruence.
    by rewrite restr_relE.
    intros x H; apply NN' in H; desf.
    apply IMP in H; unfolder in H; desf.
    apply ct_end in H0; unfolder in H0; desf.
    eapply IRR, ct_end; eexists; split; eauto.
    unfolder; eauto using clos_refl_trans.
  - rewrite <- unionA, <- ct_of_union_ct_r, path_ut2; ins.
    rewrite restr_relE.
    apply fsupp_unionI; ins; repeat apply fsupp_seqI; auto with hahn.
    rewrite <- restr_relE.
    rewrite rtE; apply fsupp_unionI; ins; auto with hahn.
    red; ins.
    tertium_non_datur (exists y', fs y' = y /\ g y'); desf; cycle 1.
      exists nil; ins; apply H; hahn_rewrite poS in REL; apply ct_end in REL.
      destruct REL as (_ & _ & ? & _ & REL); apply ct_end in REL.
      by clear - PO RF REL; unfolder in *; desf; [specialize_full PO|specialize_full RF]; desf; eauto.
    specialize (FS y'); desc; exists (map fs findom); ins.
    eapply clos_trans_mori, NN' in REL; desf.
      apply in_map, FS.
      apply IMP in H0; ins; unfolder in H0; desf.
      unfolder; apply rt_ct; eexists; split; eauto using clos_refl_trans_mon.
      eapply ct_of_ct, clos_trans_mon; eauto; ins; desf.
      eapply ct_begin; eexists; eauto using clos_refl_trans_mon.
    rewrite ct_unionE; relsf. 
    rewrite rtE at 1; relsf.
    rewrite ?seqA.
    arewrite_false !(rf'⁺ ⨾ (restr_rel (set_compl (lib_events lib)) (collect_rel fs rf))⁺); rels.
    rewrite RF', ct_end, ct_begin; clear; unfolder; ins; desf.
  }
  clear BWFs_MIN NN'.

  eexists; splits; eauto; ins; eauto. 
  eapply libspec_monotonicity in LIBOK'0; eauto; ins.
    by rewrite union_restr, restr_ct.
  by rewrite LIBOK'.
Qed.


Theorem impl_ok lib libs (COMPAT : libs_compatible (libs ∪₁ eq lib)) (impl : libimpl_verified lib) 
        (Ldeps: li_assumes impl ⊆₁ libs) A As (e: expr A As) :
  behaviours libs (expand_fcall (li_code impl) e) ⊆₁ behaviours (libs ∪₁ eq lib) e.
Proof.
  tertium_non_datur (has_bad_behaviour (libs ∪₁ eq lib) e); unfolder in *; vauto.
    unfold behaviours; ins; desf; eauto 8.
    left; eapply impl_good; eauto.
  destruct H; eapply impl_bad; eauto. 
Qed.
