Require Import Hahn Omega Operations Language LangSemantics ExpandFcall ImplBase. 
Require Import LibBase LibLock LibC11 impl_resources.

Set Implicit Arguments.

Definition initlock_body : op_impl_type op_initlock := fun argvector =>
  let addr := fst argvector in 
  CALL C11_store (addr , 0%Z , WMrlx).

Definition newlock_body : op_impl_type op_newlock := fun argvector =>
  LET block <- CALL alloc_block ( ) in 
  let addr := (block,0) in
  CALL C11_store (addr , 0%Z , WMrlx) ;;;
  Evalue (addr).

Definition lock_body : op_impl_type op_lock := fun argvector =>
  let addr := fst argvector in 
  REPEAT LET i <- CALL C11_cas (addr , 0%Z , 1%Z , UMacq , RMrlx) in 
         Evalue (if Z.eq_dec i 0 then true else false)
  END.

Definition unlock_body : op_impl_type op_unlock := fun argvector =>
  let addr := fst argvector in 
  CALL C11_store (addr , 0%Z , WMrel).

Lemma unique_loc: forall e v v', RC11.has_loc v e -> RC11.has_loc v' e -> v = v'.
Proof.
  by unfold RC11.has_loc; destruct e; ins; desf; ins; rewrite <- !eq_rect_eq in *.
Qed.

Lemma rmw_unique_rval: forall e v v', RC11.is_rmw e -> RC11.has_rval v e -> RC11.has_rval v' e -> v = v'.
Proof.
  by unfold RC11.is_rmw, RC11.has_rval; destruct e; ins; desf; ins; rewrite <- !eq_rect_eq in *.
Qed.


(*  try solve [let X := fresh in clear; intro X; apply proj1 in X; eapply X; unfolder; eauto]. *)

Definition set_non_empty A (r: A -> Prop) := exists x, r x.

Hint Unfold set_non_empty : unfolderDb. 

Lemma set_non_empty_eq A (a: A) : set_non_empty (eq a).
Proof. vauto. Qed.

Lemma set_non_empty_union_l A (s s': A -> Prop) : set_non_empty s -> set_non_empty (s ∪₁ s').
Proof. unfolder; ins; desf; eauto. Qed.

Lemma set_non_empty_union_r A (s s': A -> Prop) : set_non_empty s' -> set_non_empty (s ∪₁ s').
Proof. unfolder; ins; desf; eauto. Qed.

Hint Resolve set_non_empty_eq set_non_empty_union_l set_non_empty_union_r : hahn. 

Lemma set_non_emptyE A (s : A -> Prop) : set_non_empty s <-> ~ s ≡₁ ∅.
Proof. unfolder; split; try red; ins; clarify_not; desf; eauto. Qed.

Ltac solve_non_empty := 
  ins; desf; ins; unfold idom_seq, idom_loop in *; desf;
  repeat first [match goal with H : _ ≡ _ |- _ => clear H |
                           H : _ ≡₁ _ |- _ => rewrite H; clear H end
               | rewrite set_le, set_bunion_union_l, set_bunion_eq];
  apply set_non_emptyE; try solve [clear; auto 10 with hahn].

Program Definition lock_impl : libimpl_base LockLib.library := 
  {| li_assumes := eq RC11.library ;
     li_code := IMPL newlock_body , lock_body , unlock_body ; 
     li_locf addr lib' addr' := exists (LIB: lib' = RC11.library), addr = (eq_rect addr' LIB) 
  |}.

Tactic Notation "set_rewrite" constr(E) "in" hyp(H) :=
  match type of H with
  | _ ?x =>
     forward apply (set_equiv_inr x E); [done|clear H; intro H; unfolder in H; desf]
  end.

Lemma lock_impl_is_locally_sound :
  locally_sound_impl lock_impl.
Proof.
  red; ins. 
  specialize_full LCONS; try edone; simpls.
  remember (fun a b => << aLIB: lib_events LockLib.library a >> /\
                       << bLIB: lib_events LockLib.library b >> /\
                       exists ai bi, 
                         << Gai: g ai >> /\ << FA: fs ai = a >> /\ << STOREai: RC11.is_store ai >> /\
                         << Gbi: g bi >> /\ << FB: fs bi = b >> /\ << RMWbi: RC11.is_rmw bi >> /\  
                         << ACQbi: RC11.is_acquire bi >> /\
                         << RFi: rf ai bi >>) as rf' eqn:RF'.
  assert (RF'base: forall a b, rf' a b -> In (ev_op a) (op_newlock :: op_unlock :: nil) /\
                                          ev_op b = op_lock).
  { split; desf.
    * clear LCONS bLIB Gbi ACQbi RMWbi RFi bi.
      specialize (ESEM _ Gai).
      case_eq (fs ai); intros * AI; rewrite AI in *; ins; desf.
      red in aLIB; ins.
      unfold env_compose, env_of_impl in ESEM; desf; eauto.
      specialize_full ESEM; eauto.
      rewrite <- eq_rect_eq in *.
      simpls.
      unfold idom_loop, idom_seq in ESEM; ins; desf; ins; desf.
      clear_gpos; rewrite_gpos; event_ofs.
      set_rewrite gEQ in Gai.
      + specialize_full NLAST; eauto; desf; ins; desf.
        clear_gpos; rewrite_gpos; event_ofs; desf.
        set_rewrite gEQ1 in Gai0.
        by clear - STOREai CASFAIL; contra.
      + by clear - STOREai CASOK; contra. 
    * clear LCONS aLIB Gai STOREai RFi ai.
      specialize (ESEM _ Gbi).
      case_eq (fs bi); intros * BI; rewrite BI in *; ins; desf.
      red in bLIB; ins.
      unfold env_compose, env_of_impl in ESEM; desf; eauto.
      all: specialize_full ESEM; eauto; rewrite <- eq_rect_eq in *; simpls.
      all: unfold idom_loop, idom_seq in ESEM; ins; desf; ins; desf.
      all: clear_gpos; rewrite_gpos; event_ofs.
      all: set_rewrite gEQ in Gbi. 
      all: generalize dependent bi; clear; ins; contra.
  }
  assert (RFt: rf'⁺ ≡ rf').
  { clear - RF'base; apply ct_no_step.
    split; red; ins.
    red in H; desf.
    apply RF'base in H; apply RF'base in H0; desf.
    all: by rewrite H2 in H0.
  } 
  assert (RFinvfunc: functional rf'⁻¹).
  { clear RFt; red; intros.
    red in H, H0.
    specialize (RF'base _ _ H); desf.
    clear RF'base.
    f_equal.
    assert(bi0 = bi).
    { specialize (ESEM _ Gbi0).
      case_eq (fs bi0); intros * FBI; rewrite FBI in *; simpls.
      subst ev_op.
      unfold env_compose, env_of_impl in ESEM; desf.
      specialize_full ESEM; eauto.
      rewrite <- eq_rect_eq in *; simpls; desf.
      unfold idom_loop, idom_seq in ESEM; ins; desf; ins; desf.
      clear_gpos; rewrite_gpos; event_ofs.
      set_rewrite gEQ in Gbi0.
      * specialize (NLAST _ Gbi0); unfold idom_seq in NLAST; desf; ins; desf.
        clear_gpos; rewrite_gpos; event_ofs; apply gEQ1 in Gbi1; desf.
        clear - RMWbi0 CASFAIL; contra.
      * set_rewrite gEQ in Gbi.
        specialize (NLAST _ Gbi); unfold idom_seq in NLAST; desf; ins; desf.
        clear_gpos; rewrite_gpos; event_ofs; apply gEQ1 in Gbi0; desf.
        clear - RMWbi CASFAIL; contra.
    }
    red in LCONS; simpls; desf.
    red in RMWbi; desc.
    eapply FUNC.
    all: apply show_restr_rel; try eassumption; try by red; ins; eauto.
  }

  assert (RFfunc: functional rf').
  { intros x y z RF'xy RF'xz.
    generalize (RF'base _ _ RF'xy), (RF'base _ _ RF'xz); clear RF'base; intros XY XZ; desf.
    assert (ai0 = ai); subst.
    { clear XZ; ins; desf.
      all: specialize (ESEM _ Gai0).
      all: case_eq (fs ai0); intros * FAI; rewrite FAI in *; simpls. 
      all: subst ev_op.
      all: unfold env_compose, env_of_impl in ESEM; desf.
      all: specialize_full ESEM; eauto.
      all: rewrite <- eq_rect_eq in *; simpls; desf.
      * unfold idom_seq in *; desf. 
        clear_gpos; rewrite_gpos; event_ofs; desf.
        set_rewrite gEQ in Gai.
          by clear - STOREai ALLOC; contra.
        set_rewrite gEQ in Gai0.
        by clear - STOREai0 ALLOC; contra. 
      * assert (a = ai) by (by apply gEQ); desf.
        by symmetry; apply gEQ.
    } 
    f_equal. 
    clear -LCONS Gai STOREai RFi RMWbi RMWbi Gbi RFi0 RMWbi0 RMWbi0 Gbi0 aLIB bLIB bLIB0.
    red in LCONS; ins; desf.
    assert (lib_events RC11.library ai).
      by red; ins; red in STOREai; desf; eauto. 
    assert (lib_events RC11.library bi).
      by red; ins; red in RMWbi; desf; eauto. 
    assert (lib_events RC11.library bi0).
      by red; ins; red in RMWbi0; desf; eauto. 
    assert (RC11.is_write bi) by vauto.
    assert (RC11.is_write bi0) by vauto.
    unfolder in RFLOC.
    forward apply RFLOC with (y:=bi) as [l X]; desf; splits; eauto.
    forward apply RFLOC with (y:=bi0) as [l' X']; desf; splits; eauto.
    assert (l' = l); desf.
      clear - X0 X'0; red in X0, X'0; destruct ai; ins; desf; ins; 
        by rewrite <- eq_rect_eq.
    assert (mo ai bi). {
      tertium_non_datur (bi = ai) as [|NEQ]; desf.
        by edestruct ACYC; unfolder; eauto 8 using t_step.
        eapply MOT in NEQ; unfolder; splits; eauto; desf; vauto.
        by edestruct ACYC; unfolder; eauto 20 using t_trans, t_step.
    } 
    assert (mo ai bi0). {
      tertium_non_datur (bi0 = ai) as [|NEQ]; desf.
        by edestruct ACYC; unfolder; eauto 8 using t_step.
        eapply MOT in NEQ; unfolder; splits; eauto; desf; vauto.
        by edestruct ACYC; unfolder; eauto 20 using t_trans, t_step.
    } 
    apply NNPP; intro NEQ; assert (FOO:=NEQ); eapply MOT in NEQ; unfolder; splits; eauto.
    desf;
     apply (ACYC bi), t_trans with (y := bi0); apply t_step; unfolder; eauto.
    all: right; split; try red; ins; desf; eauto 8.
  }

  assert (RF'dom : rf' ⊆ (⋃l, (LockLib.is_newlock_ok l ∪₁ LockLib.is_unlock_ok l) × LockLib.is_lock_ok l)).
  { clear RFt RFfunc RFinvfunc.
    red; red; intros.
    specialize (RF'base _ _ H); desc.
    subst rf'; desf.
    assert (M: exists l, RC11.has_loc l ai /\ RC11.has_loc l bi); desc.
    { red in LCONS; ins; desc.
      eapply show_restr_rel, RFLOC in RFi; ins; unfolder in RFi; desc; eauto.
      by clear - STOREai; red in STOREai; red; ins; eauto.
      by clear - RMWbi; red in RMWbi; red; ins; desf; eauto. }
    exists l; unfolder; splits; ins.
    * specialize (ESEM _ Gai); generalize dependent ai; clear; ins.
      unfold env_compose, env_of_impl in *; ins; desf; specialize (ESEM _ eq_refl).
      all: destruct (fs ai) eqn: F; ins; desf; ins.
      all: unfold idom_loop, idom_seq in *; ins; desf; ins; desf.
      all: clear_gpos; rewrite_gpos; event_ofs; desf.
      all: set_rewrite gEQ in Gai. 
      all: try (generalize dependent ai; clear; ins; contra).
      all: constructor; exists eq_refl; ins;  eauto using RC11.has_loc_unique; desf.
    * specialize (ESEM _ Gbi); generalize dependent bi; clear; ins.
      unfold env_compose, env_of_impl in *; ins; desf; try specialize (ESEM _ eq_refl). 
      all: destruct (fs bi) eqn: F; ins; desf; ins. 
      all: unfold idom_loop, idom_seq in *; ins; desf; ins; desf.
      all: clear_gpos; rewrite_gpos; event_ofs; desf.
      all: set_rewrite gEQ in Gbi. 
      all: exists eq_refl; ins; eauto using RC11.has_loc_unique; desf.
      specialize (NLAST _ Gbi); desf; ins; desf.
      clear_gpos; rewrite_gpos; event_ofs; desf. 
      apply gEQ1 in Gbi0; desf.
      clear - CASFAIL RMWbi; contra.
  }

  assert (RF'codom : codom_rel rf' ≡₁ LockLib.is_lock ∩₁ set_collect fs g).
  { split; red; ins.
    + unfolder in H; desc.
      split; [eby eapply RF'base|desf; vauto].       
    + clear RF'base RFinvfunc RFfunc RF'dom.
      destruct H as [LOCKx GSx]; red in LOCKx.
      assert (exists bi, << RMWbi: RC11.is_rmw bi >> /\ 
                         << ACQbi: RC11.is_acquire bi >> /\ 
                         << LOCbi: RC11.has_loc 
                            (fst (eq_rect (ev_args x) (f_equal _ LOCKx))) bi >> /\
                         << RVALbi: RC11.has_rval 0%Z bi >> /\
                         << Gbi: g bi >> /\ << Fbi: fs bi = x >>).
      {
        rewrite_gpos.
        destruct GSx as (xi & Gxi & ?); desf.
        specialize (ESEM _ Gxi) as XI.
        destruct (fs xi) eqn: FXI; ins.
        subst ev_op; unfold env_compose, env_of_impl in XI; desf.
        specialize_full XI; eauto.
        rewrite <- eq_rect_eq in *; simpls.
        unfold idom_loop, idom_seq, idom_loop_helper in XI; desf; ins; desf.
        clear_gpos; rewrite_gpos; event_ofs; desf.
        exists a; splits; vauto; apply gEQ; vauto.
      }
      desc.
      assert (exists ai, << RFaibi: rf ai bi >> /\ 
                         << LIBai: lib_events RC11.library ai >> /\
                         << Gai: g ai >> /\ 
                         << LOCai: RC11.has_loc 
                            (fst (eq_rect (ev_args x) (f_equal _ LOCKx))) ai >> /\
                         << WVALai: RC11.has_wval 0%Z ai >>).
      {
        clear ESEM RF'; red in LCONS; simpls; desf.
        forward apply (set_equiv_inl bi CODOM) as A. 
          unfolder; splits; vauto.
          by destruct bi; red in RMWbi; ins; desf; rewrite <- eq_rect_eq in *; red; ins; eauto.
        unfolder in A; desf.
        eexists; splits; ins; eauto.
        - by apply RF, proj1 in A.
        - eapply show_restr_rel, RFLOC in A; ins; unfolder in A; desf.
          by rewrite (unique_loc LOCbi A3). 
        - eapply show_restr_rel, RFVAL in A; ins; unfolder in A; desf.
          by rewrite (rmw_unique_rval RMWbi RVALbi A3). 
      }
      desc.
      exists (fs ai).        
      assert (RC11.is_store ai).
      { clear GSx Gbi RVALbi RMWbi ACQbi LOCbi RFaibi Fbi.
        forward eapply (GDOM ai) as LIBai'; ins.
        red in LIBai'; ins; desf; eauto.
        all: specialize (ESEM _ Gai).
        all: destruct (fs ai) eqn:X; ins; desf; ins.
        all: unfold env_compose, env_of_impl in ESEM; desf; rewrite <- eq_rect_eq in *.
        all: specialize (ESEM _ eq_refl); ins.
        all: unfold newlock_body, lock_body, unlock_body in ESEM; ins.
        all: unfold idom_seq, idom_loop,  idom_loop_helper in *; ins; desf; ins; desf. 
        all: clear_gpos; rewrite_gpos; event_ofs.
        all: forward apply (set_equiv_inr ai gEQ) as A; desf;
        unfolder in A; desf.
        * clear - LOCai ALLOC; contra.
        * specialize_full NLAST; eauto; desf; desf;
          clear_gpos; rewrite_gpos; event_ofs; desf.
          clear_gpos; rewrite_gpos; apply gEQ1 in A0; desf.
          clear - WVALai CASFAIL; contra.
        * clear - WVALai CASWVAL.
          exfalso; unfold RC11.has_wval in *; desf; destruct ai; ins; desf; 
          rewrite <- ?eq_rect_eq in *; ins; try omega.
      }
      clear LIBai LOCbi LOCai; desf.
      unfolder; splits; eauto.
      exists ai, bi; splits; eauto.
   }

  exists rf', rf'; splits; try rewrite RFt; clear RFt.
    1-2: by subst rf'; repeat split; desf; red; eauto.
    by unfolder; splits; ins; desf; eauto 8 using t_step.
  { repeat red; ins; desf.
    assert (exists loc, <<LOCai: RC11.has_loc loc ai >> /\ 
                        <<LOCbi: RC11.has_loc loc bi>>); desc.
    { red in LCONS; ins; desf.
      eapply show_restr_rel, RFLOC in RFi; ins; unfolder in RFi; desf; vauto.
        by red; ins; red in STOREai; ins; desf; eauto.
      by red; ins; red in RMWbi; ins; desf; eauto. }
    assert (RC11.is_release ai \/ libspec_allocs LockLib.library (fs ai) loc).
    {
      clear RF'base.
      specialize (ESEM _ Gai).
      assert (aOP := aLIB); red in aOP; ins; desf;
      unfold env_compose, env_of_impl in ESEM; desf;
      try solve [clear ESEM; rewrite <- aOP in *; desf];
      specialize (ESEM _ eq_refl).
      {
        right; exists (eq_sym aOP).
        destruct (fs ai) eqn: BI; ins; desf; ins.
        unfold idom_seq in *; desf.
        rewrite <- !eq_rect_eq in *.
        clear_gpos; rewrite_gpos; event_ofs.
        forward apply (set_equiv_inr ai gEQ) as B; ins; unfolder in B; desf; 
          eauto using RC11.has_loc_unique.
        clear - ALLOC STOREai; contra.
      }
      all: apply op_approx in ESEM; red in ESEM; forward apply ESEM as B; try red; eauto.
      all: destruct (fs ai) eqn: BI; ins; desf; ins; unfolder in B; ins; desf.
      all: red in STOREai; desf; try solve [destruct ai; ins; desf]. 
      left; red; destruct ai; ins; desf; ins. 
      eexists; split; [by repeat constructor; exists eq_refl|]; ins.
      rewrite <- eq_rect_eq; ins.
    }
    assert (aLIB' : lib_events RC11.library ai).
      by revert STOREai; clear; unfold RC11.is_store; red; ins; desf; eauto.
    assert (bLIB' : lib_events RC11.library bi).
      by revert RMWbi; clear; unfold RC11.is_rmw; red; ins; desf; ins; eauto.
    desf.
      enough (so ai bi) by (unfolder; eauto 10 using t_step).
      red in LCONS; desc; ins.
      apply SOEQ; unfolder; split; ins.
      eexists; splits; eauto.
      eexists; splits; eauto using rt_refl.
      eexists; splits; eauto using rt_refl.
    forward apply ENCAPS with (e:=fs bi); ins.
      by vauto.
    { instantiate (1 := loc).
      clear RF'base.
      specialize (ESEM _ Gbi).
      assert (bOP := bLIB); red in bOP; ins; desf;
      unfold env_compose, env_of_impl in ESEM; desf;
      try solve [clear ESEM; rewrite <- aOP in *; desf];
      specialize (ESEM _ eq_refl).
      all: apply op_approx in ESEM; red in ESEM; forward apply ESEM as B; try red; eauto.
      all: destruct (fs bi) eqn: BI; ins; desf; ins; unfolder in B; ins; desf.
      all: red in RMWbi; desf; try solve [destruct bi; ins; desf]. 
      repeat constructor; exists eq_refl; ins.
      red in LOCbi; desf; try solve [destruct bi; ins; desf]. 
      destruct bi; ins; desf; ins.
      rewrite <- !eq_rect_eq; ins. }
    desf.
    assert (a = fs ai); desf.
      by red in Ga; desf; eapply C_ALLOC; ins; eauto.
    by do 2 eexists; splits; unfolder; eauto using t_step. 
  }
  { by apply functional_inv_fsupp. }
  { red; splits; simpls. 
    * clear RF'.
      red; ins; red.
      red in H; desf.
      specialize (GDOM _ H); red in GDOM.
      specialize (ESEM _ H).
      all: destruct (fs y); ins; desf.
      all: unfold env_compose, env_of_impl in ESEM; desf.
      all: specialize_full ESEM; eauto.
      all: by eexists; split; ins; repeat constructor; exists eq_refl.
    * unfold set_inter, set_collect; ins; desf; eapply C_ALLOC; ins; eauto.
  }
Qed. 


Program Definition lock_impl_verified : libimpl_verified LockLib.library := 
  {| li_base := lock_impl |}.
Next Obligation.
  unfold env_compose, env_of_impl; ins; desf; split; ins; desf; eauto.
Qed.
Next Obligation.
  unfold env_compose, env_of_impl in *; solve_non_empty.
Qed.
Next Obligation.
 apply op_approx in SEM; rewrite SEM; clear SEM.
 unfold env_compose, env_of_impl in *; ins; desf.
 all: ins; unfolder; ins; desf; eexists; split; ins.
 all: destruct x; red; ins; desf; eauto.
Qed.
Next Obligation.
  do 2 eexists; split; ins; exists eq_refl; ins.
Qed.
Next Obligation.
  red; ins; desf; ins; rewrite <- !eq_rect_eq in *; congruence.
Qed.
Next Obligation.
  unfold env_compose, env_of_impl in *; ins; desf.
  * (* Allocation *)
    destruct a; ins; desf; rewrite <- eq_rect_eq in SEM; ins;
    unfold idom_seq in *; desf; ins; desf.
    unfolder; ins; desf; rewrite_gpos; event_ofs.
    apply gEQ in Ge; unfolder in Ge; desf.
      by exfalso; clear - H ALLOC; contra.
    right; exists a0; splits; ins; eauto.
      by eapply gEQ; vauto.
      eexists; split; ins; apply poEQ; vauto.
      assert (x = (res0, 0)) by eauto using RC11.has_loc_unique; desf.
 
  * (* Lock *)
    apply op_approx in SEM; apply SEM in Ge; clear SEM.
    destruct a; ins; desf; rewrite <- eq_rect_eq in Ge; ins; unfolder in Ge; desf.
    unfolder; ins; desf.
    left; eexists; split; repeat constructor; exists eq_refl; ins.
    by red in H; destruct e; ins; desf; ins; rewrite <- eq_rect_eq; ins.

  * (* Unlock *)
    apply op_approx in SEM; apply SEM in Ge; clear SEM.
    destruct a; ins; desf; rewrite <- eq_rect_eq in Ge; ins; unfolder in Ge; desf.
    unfolder; ins; desf.
    left; eexists; split; repeat constructor; exists eq_refl; ins.
    by red in H; destruct e; ins; desf; ins; rewrite <- eq_rect_eq; ins.
Qed.
Next Obligation.
  red in ALLOC; desc.
  destruct a; ins; unfolder; unfold env_compose, env_of_impl in *; ins; desf.
  rewrite <- eq_rect_eq in *; desf.
  unfold initlock_body, newlock_body, lock_body, unlock_body in *; ins;
    unfold idom_seq, idom_loop in *; desf; ins; desf.
  clear_gpos; rewrite_gpos; event_ofs.
  autorewrite with hahn in gEQ.
  exists a0; splits; ins.
  apply gEQ; vauto.
Qed. 
Next Obligation. 
  exact lock_impl_is_locally_sound.
Qed.


Corollary lock_impl_ok libs (COMPAT: libs_compatible (libs ∪₁ eq LockLib.library))
          (L: eq RC11.library ⊆₁ libs) A As (e: expr A As) : 
  behaviours libs (expand_fcall (li_code lock_impl) e) 
  ⊆₁ behaviours (libs ∪₁ eq LockLib.library) e.
Proof.
  apply impl_ok with (impl := lock_impl_verified); ins.
Qed.
