Require Import Hahn Omega Operations LibBase LibLock LibC11.
Require Import Language LangSemantics ExpandFcall ImplBase. 
Require Import LibQueue impl_resources.
Require Import ClassicalDescription IndefiniteDescription.

Set Implicit Arguments.

Definition locked_newqueue_body: op_impl_type op_newqueue := 
  fun argvector =>
    LET queue_block <- CALL alloc_block ( ) in 
    LET lock_addr <- CALL op_newlock ( ) in
    CALL C11_store ( (queue_block, 1) , 0%Z , WMna ) ;;;
    CALL C11_store ( (queue_block, 2) , 0%Z , WMna ) ;;;
    Evalue (queue_block, lock_addr).

Definition locked_enqueue_body: op_impl_type op_enqueue :=   
  fun argvector =>
    let q := fst (fst argvector) in
    let l := snd (fst argvector) in
    let v := fst (snd argvector) in
    CALL op_lock ( l ) ;;; 
    LET last_index <- CALL C11_load ( (q,1) , RMna ) in
    CALL C11_store ( (q,1) , (last_index + 1)%Z , WMna ) ;;;
    CALL C11_store ( (q,(Z.to_nat last_index) + 3) , v , WMna ) ;;;
    CALL op_unlock ( l ).

Definition locked_dequeue_body : op_impl_type op_dequeue := 
  fun argvector =>
    let q := fst (fst argvector) in
    let l := snd (fst argvector) in
    CALL op_lock ( l ) ;;; 
    LET last_index <- CALL C11_load ( (q,1) , RMna ) in
    LET first_index <- CALL C11_load ( (q,2) , RMna ) in
    if (first_index =? last_index)%Z then 
      CALL op_unlock ( l ) ;;; 
      Evalue None
    else
      CALL C11_store ( (q,2) , (first_index + 1)%Z , WMna ) ;;;
      LET res <- CALL C11_load ( (q,(Z.to_nat first_index) + 3) , RMna ) in
      CALL op_unlock ( l ) ;;;
      Evalue (Some res).



Lemma list_destruct_end: forall A (l: list A), l = nil \/ exists l' a, l = l' ++ a :: nil.
  induction l; desf; eauto.
  by right; eexists nil, _; rewrite app_nil_l.
  eby right; eexists _, _; rewrite app_comm_cons.
Qed.


Fixpoint run_queue es :=
  match es with
    | nil => (0%Z, 0%Z)
    | e :: es' => let (last_index,first_index) := run_queue es' in
                  if excluded_middle_informative (QueueLib.is_enqueue e)
                  then ((last_index + 1)%Z, first_index)
                  else if excluded_middle_informative (QueueLib.is_dequeue e)
                  then if (first_index =? last_index)%Z 
                       then (last_index, first_index)
                       else (last_index, (first_index + 1)%Z)
                  else (0%Z, 0%Z)
  end.


Ltac find_in_list := (by repeat progress (rewrite ?in_app_iff; ins); eauto 10).

   
  
Program Definition LockedQueue_impl : libimpl_base QueueLib.library := 
  {| li_assumes := fun lib => lib = RC11.library \/ lib = LockLib.library;
     li_code := IMPL locked_newqueue_body , locked_enqueue_body , locked_dequeue_body ; 
     li_locf addr lib' addr' := 
       (exists (LIB: lib' = RC11.library), fst (eq_rect addr' LIB) = fst addr) \/
       (exists (LIB: lib' = LockLib.library), eq_rect addr' LIB = snd addr) |}.


Lemma LockedQueue_impl_acyclic_rf : 
  forall g po rf so hb 
    (RFDOM : rf ⊆ (⋃l ∈ fun lib => lib = RC11.library \/ lib = LockLib.library, lib_events l × lib_events l))
    (LCONS : forall lib' : libspec,
          lib' = RC11.library \/ lib' = LockLib.library ->
          libspec_consistent lib'
            {|
            tup_ev := g ∩₁ lib_events lib';
            tup_po := restr_rel (lib_events lib') po;
            tup_rf := restr_rel (lib_events lib') rf;
            tup_so := restr_rel (lib_events lib') so;
            tup_hb := restr_rel (lib_events lib') hb |}),
  acyclic rf.
Proof.
  ins; arewrite (rf ⊆ restr_rel (lib_events LockLib.library) rf ∪ 
                      restr_rel (lib_events RC11.library) rf).
    by unfolder; unfolder in RFDOM; ins; desf; specialize_full RFDOM; eauto; desf; eauto.
  apply acyclic_absorb; splits; ins.
  - left; unfold lib_events; unfolder; ins; desc.
    exfalso; clear - H1 H4; destruct z; ins; desf.
  - specialize (LCONS LockLib.library); apply LockLib.acyclic_rf in LCONS; auto.
  - specialize (LCONS RC11.library); apply RC11.acyclic_rf in LCONS; auto. 
Qed.

Lemma acyclic_unionE A (r r' : relation A) :
  acyclic (r ∪ r') <-> acyclic (r ⨾ r'^*) /\ acyclic r'.
Proof.
  split; ins; desf.
    split; eapply irreflexive_inclusion; eauto; eauto with hahn. 
    eapply inclusion_t_t2; rewrite ct_begin; eauto with hahn.
  unfold acyclic; rewrite <- ct_of_union_ct_l, unionC. 
  apply acyclic_ut; splits; ins; clear H0;
  eapply irreflexive_inclusion; eauto.  
    by rewrite <- inclusion_id_rt; rels. 
  apply inclusion_t_t2; rewrite inclusion_t_rt with (r := r').
  rewrite !ct_end.
  do 2 hahn_frame_r; rewrite <- inclusion_id_rt with (r' := r'); rels. 
Qed.

Lemma LockedQueue_impl_acyclic_porf : 
  forall g po rf so hb 
    (RFDOM : rf ⊆ (⋃l ∈ fun lib => lib = RC11.library \/ lib = LockLib.library, lib_events l × lib_events l))
    (IRR : acyclic (po ⨾ rf＊))
    (LCONS : forall lib' : libspec,
          lib' = RC11.library \/ lib' = LockLib.library ->
          libspec_consistent lib'
            {|
            tup_ev := g ∩₁ lib_events lib';
            tup_po := restr_rel (lib_events lib') po;
            tup_rf := restr_rel (lib_events lib') rf;
            tup_so := restr_rel (lib_events lib') so;
            tup_hb := restr_rel (lib_events lib') hb |}),
  acyclic (po ∪ rf).
Proof.
  ins; eapply acyclic_unionE; eauto using LockedQueue_impl_acyclic_rf.
Qed.


Lemma LockedQueue_impl_expand_so : 
  forall deps g po rf so hb 
    (SODOM : so ⊆ (⋃l ∈ deps, lib_events l × lib_events l))
    (LCONS : forall lib',
          deps lib' ->
          libspec_consistent lib'
            {|
            tup_ev := g ∩₁ lib_events lib';
            tup_po := restr_rel (lib_events lib') po;
            tup_rf := restr_rel (lib_events lib') rf;
            tup_so := restr_rel (lib_events lib') so;
            tup_hb := restr_rel (lib_events lib') hb |}),
  so ⊆ (po ∪ rf)⁺.
Proof.
  ins; arewrite (so ⊆ (⋃l ∈ deps, restr_rel (lib_events l) so)).
    by unfolder; unfolder in SODOM; ins; desf; specialize_full SODOM; eauto; desf; eauto.
  unfolder; ins; desf. 
  specialize (LCONS _ H); eapply libspec_cons_sodom in LCONS; ins.
  eapply inclusion_trans; eauto; ins. 
  by rewrite !inclusion_restr.
Qed.


Lemma LockedQueue_impl_acyclic_posohb : 
  forall (g : event -> Prop) po rf so hb (fs : event -> event)
    (SODOM : so ⊆ (⋃l ∈ fun lib => lib = RC11.library \/ lib = LockLib.library, lib_events l × lib_events l))
    (RFDOM : rf ⊆ (⋃l ∈ fun lib => lib = RC11.library \/ lib = LockLib.library, lib_events l × lib_events l))
    (HBACYC: acyclic ((po ∪ restr_rel g (map_rel fs hb)) ⨾ rf＊))
    (LCONS : forall lib' : libspec,
          lib' = RC11.library \/ lib' = LockLib.library ->
          libspec_consistent lib'
            {|
            tup_ev := g ∩₁ lib_events lib';
            tup_po := restr_rel (lib_events lib') po;
            tup_rf := restr_rel (lib_events lib') rf;
            tup_so := restr_rel (lib_events lib') so;
            tup_hb := restr_rel (lib_events lib') (po ∪ so ∪ restr_rel g (map_rel fs hb))^+ |}),
  acyclic (po ∪ so ∪ restr_rel g (map_rel fs hb)).
Proof.
  ins.
  rewrite (LockedQueue_impl_expand_so _ _ _ _ SODOM LCONS).
  rewrite union_absorb_l with (r := po); vauto.
  red; rewrite ct_of_union_ct_l. 
  rewrite unionC, <- unionA, (unionC _ po). 
  ins; eapply acyclic_unionE; eauto using LockedQueue_impl_acyclic_rf.
Qed.

Section BasicProperties.

Variables (g : event -> Prop) (po : relation event) (fs : event -> event).
Hypothesis GDOM : forall a, g a -> lib_events QueueLib.library (fs a).
Hypothesis ESEM : forall a,
    g a ->
    forall body : op_impl_type (ev_op (fs a)),
      env_compose (env_of_impl locked_newqueue_body)
                  (env_compose (env_of_impl locked_enqueue_body)
                               (env_compose (env_of_impl locked_dequeue_body) (fun op : operation => None))) 
                  (ev_op (fs a)) = Some body ->
      expr_sem (body (ev_args (fs a))) (Some (RTbase (ev_res (fs a)))) 
               (g ∩₁ (fun x => fs x = fs a))
               (restr_rel (fun x => fs x = fs a) po).

Lemma po_from_lock : 
  forall ai lcki (Gai: g ai) (GL: g lcki) (EQ: fs ai = fs lcki) 
         (LCK: LockLib.is_lock lcki),
    po^* lcki ai.
Proof.
  ins. assert (E:=ESEM GL); assert (G:=GDOM GL).
  revert E; unfold env_compose, env_of_impl.
  red in G; simpls; des; clarify.
  all: destruct (fs lcki) eqn: FSL; simpl; subst; desf.
  all: ins; specialize_full E; auto.
  all: apply op_approx2 with (po' := po) in E; auto with hahn.
  all: ins; desf; ins; desf; ins.
  all: rewrite_gpos.
  all: forward apply (set_equiv_inr lcki E2) as L; try done.
  all: unfolder in L; desf; try by destruct lcki; ins; desf.
  all: forward apply (set_equiv_inr ai E2) as A; try done.
  all: unfolder in A; desf; auto using rt_refl; apply rt_step.
  all: apply E3; unfolder; tauto. 
Qed.

Lemma po_to_unlock : 
  forall ai ulcki (Gai: g ai) (GL: g ulcki) (EQ: fs ai = fs ulcki) 
         (LCK: LockLib.is_unlock ulcki), 
    po^* ai ulcki.
Proof.
  ins; assert (E:=ESEM GL); assert (G:=GDOM GL).
  revert E; unfold env_compose, env_of_impl.
  red in G; simpls; des; clarify.
  all: destruct (fs ulcki) eqn: FSL; simpl; subst; desf.
  all: ins; specialize_full E; auto.
  all: apply op_approx2 with (po' := po) in E; auto with hahn.
  all: ins; desf; ins; desf; ins.
  all: rewrite_gpos.
  all: forward apply (set_equiv_inr ulcki E2) as L; try done.
  all: unfolder in L; desf; try by destruct ulcki; ins; desf.
  all: forward apply (set_equiv_inr ai E2) as A; try done.
  all: clear E2; unfolder in A; desf; auto using rt_refl; apply rt_step.
  all: try solve [apply E3; clear; unfolder; tauto]. 
  all: try solve [apply E7; clear; unfolder; tauto]. 
  all: try solve [apply E10; clear; unfolder; tauto]. 
  all: try solve [apply E13; clear; unfolder; tauto].
  all: try solve [apply E16; clear; unfolder; tauto].
Qed.

Lemma has_lock_event : forall l a (GSa: set_collect fs g a) (HL: QueueLib.has_loc l a),
    exists ai, << Gai: g ai >> /\ << FSai: fs ai = a >> /\ 
               << LOCKai: LockLib.is_lock_ok (snd l) ai >>.
Proof.
  ins.
  unfolder in GSa; desf.
  assert (E := ESEM GSa); clear ESEM.
  destruct (fs y) eqn:FSy; ins; desf.
  red in HL; unfolder in HL; ins.
  desf; red in HL; ins; desf.
  all: unfold env_compose, env_of_impl in E; ins; desf.
  all: rewrite <- eq_rect_eq in *; ins.
  all: specialize_full E; eauto; ins.
  all: unfold idom_seq in E; desf; ins; unfold idom_seq in *; desf.
  all: clear_gpos; rewrite_gpos; event_ofs. 
  all: match goal with H : LockLib.is_lock ?e |- _ => exists e end; splits.
  all: try by apply gEQ; unfolder; eauto.
  all: clear - LOCK0 LOCKLOC0.
  all: red in LOCK0, LOCKLOC0; unfolder in LOCKLOC0; desf.
  all: by red in LOCKLOC0; desf; clear LOCKLOC0; rewrite LOCK0 in *.
Qed.

Lemma queue_from_lock : 
  forall lckp (LCK: lckp = LockLib.is_lock_ok \/ lckp = LockLib.is_unlock_ok)
         ai (Gai: g ai) l (Lai: lckp l ai),
  exists q, QueueLib.has_loc (q,l) (fs ai).
Proof.
  ins; assert (E := ESEM Gai); assert (G := GDOM Gai).
  desf; red in Lai; desf.
  all: destruct (fs ai) eqn:FSai; ins; red in G; ins; desf.
  all: unfold env_compose, env_of_impl in E; ins; desf.
  all: specialize_full E; eauto.
  all: rewrite <- eq_rect_eq in *; ins.
  all: unfold idom_seq in E; desf; ins; unfold idom_seq in *; desf.
  all: clear_gpos; rewrite_gpos.
  all: forward apply (set_equiv_inr ai gEQ) as A; try done.
  all: unfolder in A; desf.
  all: try by match goal with O: ev_op ?x = _ |- _ 
              => match goal with C: event_of_call _ _ _ x |- _ 
                 => clear - C O; red in C; desf; destruct x; ins; desf 
                 end 
              end.
  all: exists (fst (fst ev_args)); constructor; exists eq_refl.
  all: rewrite <- !eq_rect_eq in *; ins.
  all: unfold event_of_call in *; desf; 
       try destruct a; try destruct a0; try destruct a1; try destruct a2; try destruct ai; ins; desf. 
  all: rewrite <- !eq_rect_eq.
  all: clear; unfold queue_address in *;
       repeat match goal with H : _ * _ |- _ => destruct H | H : unit |- _ => destruct H end; ins.
Qed.

Lemma queue_from_access :
  forall ai (Gai: g ai) l (Lai: RC11.has_loc l ai),
  exists lck, QueueLib.is_newqueue_loc (fst l,lck) (fs ai) \/ 
              QueueLib.has_loc (fst l,lck) (fs ai).
Proof.
  ins; assert (E := ESEM Gai); assert (G := GDOM Gai).
  desf; red in Lai; desf.
  all: destruct (fs ai) eqn:FSai; ins; red in G; ins; desf.
  all: unfold env_compose, env_of_impl in E; ins; desf.
  all: specialize_full E; eauto.
  all: rewrite <- eq_rect_eq in *; ins.
  all: unfold idom_seq in E; desf; ins; unfold idom_seq in *; desf.
  all: clear_gpos; rewrite_gpos.
  all: forward apply (set_equiv_inr ai gEQ) as A; try done.
  all: unfolder in A; desf.
  all: try by match goal with O: ev_op ?x = _ |- _ 
              => match goal with C: event_of_call _ _ _ x |- _ 
                 => clear - C O; red in C; desf; destruct x; ins; desf 
                 end 
              end.
  all: unfold event_of_call in *; desf; destruct ai; try destruct a; try destruct a0; 
       try destruct a1; try destruct a2; try destruct a3; ins; desf.
  all: first [exists (snd (fst ev_args)) | eexists]; 
    first [constructor; exists eq_refl | do 2 constructor; exists eq_refl].
  all: rewrite <- !eq_rect_eq in *; ins.
  all: clear; unfold queue_address in *;
       repeat match goal with H : _ * _ |- _ => destruct H | H : unit |- _ => destruct H end; ins.
Qed.

Lemma queue_uniq :
  forall (POACYC: acyclic po)
         lck (LCK: lck = LockLib.is_lock \/ lck = LockLib.is_unlock)
         ai ai' (FS: fs ai = fs ai') (G:g ai) (G': g ai') (L: lck ai) (L': lck ai')
         (LIB: lib_events QueueLib.library (fs ai)) (LIB': lib_events QueueLib.library (fs ai')),
    ai = ai'.
Proof.
  ins; apply NNPP; intro NEQ; desf.
  1: eapply po_from_lock with (ai := ai'), rtE in L; eauto; unfolder in L; desf.
  1: eapply po_from_lock with (ai := ai), rtE in L'; eauto; unfolder in L'; desf.
  2: eapply po_to_unlock with (ai := ai'), rtE in L; eauto; unfolder in L; desf.
  2: eapply po_to_unlock with (ai := ai), rtE in L'; eauto; unfolder in L'; desf.
  all: eapply POACYC; eauto using rt_refl, t_trans.
Qed.

End BasicProperties.
