Require Import Hahn Omega Operations LibBase LibLock LibC11.
Require Import Language LangSemantics ExpandFcall ImplBase. 
Require Import LibQueue impl_resources ImplLockedQueue.
Require Import ClassicalDescription IndefiniteDescription.

Set Implicit Arguments.

Lemma LockedQueue_impl_is_locally_sound :
  locally_sound_impl LockedQueue_impl.
Proof.
  red; ins.
  assert (IRR': acyclic (po ∪ rf)) by (eapply LockedQueue_impl_acyclic_porf; eauto).
  assert (HBACYC': acyclic (po ∪ so ∪ restr_rel g (map_rel fs hb)))
    by (eapply LockedQueue_impl_acyclic_posohb with (rf:=rf); eauto).
  clear HBACYC IRR.

  remember (fun a b => << aLIB: lib_events QueueLib.library a >> /\
                       << bLIB: lib_events QueueLib.library b >> /\
                       exists ai bi, 
                         << Gai: g ai >> /\ << FA: fs ai = a >> /\ << aiUNLOCK: LockLib.is_unlock ai >> /\
                         << Gbi: g bi >> /\ << FB: fs bi = b >> /\ << biLOCK: LockLib.is_lock bi >> /\
                         << RFi: rf ai bi >>) as toimm eqn:TOIMM.
  (* basic properties of toimm *)
  assert (TObase: forall a b (TOab: toimm^+ a b), 
                  << aLIB: (set_collect fs g ∩₁ (QueueLib.is_enqueue ∪₁ QueueLib.is_dequeue)) a >> /\
                  << bLIB: (set_collect fs g ∩₁ (QueueLib.is_enqueue ∪₁ QueueLib.is_dequeue)) b >> /\
                  (exists qloc, << aQL : QueueLib.has_loc qloc a >> /\ << bQL : QueueLib.has_loc qloc b >>) /\
                  exists ai bi, <<Gai: g ai>> /\ <<FSai: fs ai = a>> /\ <<Uai: LockLib.is_unlock ai>> /\
                                <<Gbi: g bi>> /\ <<FSbi: fs bi = b>> /\ <<Lbi: LockLib.is_lock bi>> /\
                                <<PORF: (po ∪ rf)⁺ ai bi>> /\ 
                                <<POSO: (po ∪ so)⁺ ai bi>>).
  { intros; splits.
    * unfold QueueLib.is_enqueue, QueueLib.is_dequeue; unfolder.
      apply ct_begin in TOab; unfolder in TOab; desf.
      specialize (ESEM _ Gai).
      destruct (fs ai) eqn:FSai; ins; desf.
      red in aLIB; ins.
      unfold env_compose, env_of_impl in ESEM; ins; desf.
      all: specialize_full ESEM; eauto.
      all: rewrite <- eq_rect_eq in *; ins.
      all: unfold idom_seq in ESEM; desf; ins; unfold idom_seq in *; desf.
      all: clear_gpos; rewrite_gpos.
      all: forward apply (set_equiv_inr ai gEQ) as X; try done.
      all: unfolder in X; desf. 
      all: by match goal with O: LockLib.is_unlock ?x |- _ 
              => match goal with C: event_of_call _ _ _ x |- _ 
                 => clear - C O; red in C; desf; destruct x; ins; desf 
                 end 
              end.
    * unfold QueueLib.is_enqueue, QueueLib.is_dequeue; unfolder.
      apply ct_end in TOab; unfolder in TOab; desf.
      specialize (ESEM _ Gbi).
      destruct (fs bi) eqn:FSbi; ins; desf.
      red in bLIB; ins.
      unfold env_compose, env_of_impl in ESEM; ins; desf.
      all: specialize_full ESEM; eauto.
      all: rewrite <- eq_rect_eq in *; ins.
      all: unfold idom_seq in ESEM; desf; ins; unfold idom_seq in *; desf.
      all: clear_gpos; rewrite_gpos.
      all: forward apply (set_equiv_inr bi gEQ) as X; try done.
      all: unfolder in X; desf.
      all: by match goal with O: LockLib.is_lock ?x |- _ 
              => match goal with C: event_of_call _ _ _ x |- _ 
                 => clear - C O; red in C; desf; destruct x; ins; desf 
                 end 
              end.
   * induction TOab.
     + desf. 
       assert (MM: (⋃l, (LockLib.is_newlock_ok l ∪₁ LockLib.is_unlock_ok l) × LockLib.is_lock_ok l) ai bi).
         by apply (LCONS LockLib.library); ins; repeat red; splits; eauto; try red; ins; eauto.
       red in aiUNLOCK, biLOCK; desc. 
       unfold LockLib.is_newlock_ok, LockLib.is_unlock_ok, LockLib.is_lock_ok in MM;
       unfolder in MM; desf; [by destruct ai; ins; desf|].
       red in aLIB; ins; red in bLIB; ins.
       generalize (ESEM _ Gai), (ESEM _ Gbi); intros Sa Sb.
       unfold env_compose, env_of_impl in Sa, Sb. 
       destruct (fs ai) eqn:FSai; ins;  destruct (fs bi) eqn:FSbi; ins. 
       des; subst; ins; desf.
       all: specialize_full Sa; auto; apply op_approx in Sa. 
       all: specialize_full Sb; auto; apply op_approx in Sb. 
       all: forward apply (Sa ai) as Sa'; clear Sa; vauto.
       all: forward apply (Sb bi) as Sb'; clear Sb; vauto.
       all: rewrite <- eq_rect_eq in *; ins.
       all: unfolder in Sa'; unfolder in Sb'. 
       all: destruct ai, bi; ins; desf.
       all: rewrite <- !eq_rect_eq in *; ins; desf.
       all: repeat match goal with x : unit |- _ => destruct x | x : _ * _ |- _ => destruct x end; ins.
       all: tertium_non_datur (q0 = q) as [|NEQ]; subst.
       all: try by exists q; split; constructor; exists eq_refl; ins.
       all: forward eapply C_LOC with 
           (a := {| ev_id := ev_id1; ev_op := _; ev_args := _; ev_res := _ |})
           (b := {| ev_id := ev_id2; ev_op := _; ev_args := _; ev_res := _ |}); eauto;
         unfolder; rewrite ?FSai, ?FSbi; try solve [repeat constructor; exists eq_refl; ins].
       all: intro X; exfalso; eapply X; right; exists eq_refl; ins.
       all: try unfolder in Sa'2; desf; ins; desf.
       all: try unfolder in Sb'2; desf; ins; desf.
       all: rewrite <- ?eq_rect_eq in *; ins; desf.
     + desc; assert (qloc0 = qloc) by (eapply QueueLib.has_loc_unique; eauto); desf; eauto.
   * induction TOab.
     + desf. eexists ai, bi; splits; try done. 
         by apply t_step; vauto.
       specialize (LCONS _ (or_intror eq_refl)); ins; red in LCONS; ins; desf.
       eapply show_restr_rel, SOEQ in RFi.
       by unfolder in RFi; desf; apply t_step; vauto.
       all: by red in aiUNLOCK, biLOCK; red; ins; eauto.
     + destruct IHTOab1 as (ai & bi & Gai & Fai & Uai & Gbi & Fbi & Lbi & PORFaibi & POSOaibi).
       destruct IHTOab2 as (ci & di & Gci & Fci & Uci & Gdi & Fdi & Ldi & PORFcidi & POSOcidi).
       unnw; desf.
       exists ai, di; splits; try done.
       all: eapply t_trans, rt_t_trans; eauto.
       all: eapply clos_refl_trans_mon; [eapply po_from_lock|]; eauto; vauto. 
  }

  assert (TOM: forall x (Gx: g x) y (Gy: g y) (P: toimm⁺ (fs x) (fs y)), (po ∪ so)⁺ x y).
  {
    clear TOIMM.
    ins; specialize_full TObase; eauto; desc.
    assert (po^* x ai) by (eapply po_to_unlock; eauto).
    assert (po^* bi y) by (eapply po_from_lock; eauto).
    eapply rt_t_trans, t_rt_trans; eauto;
    eapply clos_refl_trans_mon; eauto; vauto.
  }
  
  assert (IRRto: acyclic toimm).
  { clear TOIMM; red; apply NNPP; intro REFL; clarify_not.
    apply TObase in REFL; desf.
    forward eapply po_from_lock with (ai:=ai) as POlck; eauto. 
    eapply IRR', t_rt_trans, clos_refl_trans_mon; eauto; vauto.
  }
  assert (FUNto: functional toimm).
  {
    red; ins; desf.
    clear IRRto C_ALLOC.
    forward eapply queue_uniq as U; try eassumption; eauto using acyclic_mon with hahn; desf.
    specialize (LCONS _ (or_intror eq_refl)); ins.
    red in LCONS; ins; desc.
    f_equal; eapply RFFUN.
    all: apply show_restr_rel; eauto.
    all: red; ins; eauto.
  }
  assert (FSUPPto: fsupp toimm^+).
  { clear TOIMM LCONS C_ALLOC.
    red; ins.
    tertium_non_datur ((set_collect fs g ∩₁ (QueueLib.is_enqueue ∪₁ QueueLib.is_dequeue)) y) as [Q|?].
    2-3: by exists nil; ins; apply TObase in REL; desf; unfolder in bLIB; desf; destruct H; vauto. 
    assert (exists yi, << LOCKyi: LockLib.is_lock yi >> /\ << FSyi: fs yi = y >> /\ <<Gyi: g yi>>).
    { 
      destruct Q as [GSy Q].
      destruct GSy as (yi & Gyi & FSyi); desf.
      specialize (ESEM _ Gyi).
      destruct (fs yi) eqn:FSyi; ins; desf.
      unfolder in Q; ins; desf; red in Q; ins; desf.
      all: unfold env_compose, env_of_impl in ESEM; ins; desf.
      all: specialize_full ESEM; eauto.
      all: rewrite <- eq_rect_eq in *; ins.
      all: unfold idom_seq in ESEM; desf; ins; unfold idom_seq in *; desf.
      all: clear_gpos; rewrite_gpos.
      * exists a3; splits; vauto.
        by red in aCALL3; desf.
        all: apply gEQ; unfolder; eauto.
      * exists a1; splits; vauto.
        by red in aCALL1; desf.
        all: apply gEQ; unfolder; eauto.
      * exists a1; splits; vauto.
        by red in aCALL1; desf.
        all: apply gEQ; unfolder; eauto.
    }
    desf.
    assert (MAP: forall x (TOxy: toimm⁺ x (fs yi)), 
                 exists xi, <<FSxi: fs xi = x>> /\ <<UNLOCKxi: LockLib.is_unlock xi>> /\
                            <<PORFxiyi: (po ∪ rf)⁺ xi yi>>).
    { ins; apply TObase in TOxy; desf.
      forward eapply queue_uniq with (lck :=  LockLib.is_lock); eauto using acyclic_mon with hahn.
      ins; desf; eexists; splits; eauto.
    }
    destruct (classic ((fun x => toimm⁺ x (fs yi)) ≡₁ ∅)) as [EMP|NEMP].
      by exists nil; ins; apply EMP in REL.
    apply conditional_functional_choice in MAP; desf.
    clear NEMP.
    specialize (FS yi); desf.
    eapply finite_codomain_implies_finite_domain with (f := f); ins.
    by apply MAP in Px; apply MAP in Py; desf; congruence.
    by eapply FS, MAP.
  }
  assert (SAMEloc: forall a b l, toimm⁺ a b -> QueueLib.has_loc l a -> QueueLib.has_loc l b).
  { clear - TObase; ins; specialize_full TObase; eauto; desf.
    assert (l = qloc) by eauto using QueueLib.has_loc_unique; desf. }
  assert (SAMEloc': forall a b l, toimm⁺ a b -> QueueLib.has_loc l b -> QueueLib.has_loc l a).
  { clear - TObase; ins; specialize_full TObase; eauto; desf.
    assert (l = qloc) by eauto using QueueLib.has_loc_unique; desf. }

  assert (TOmin: forall m l (M: (set_collect fs g ∩₁ QueueLib.has_loc l) m)
                   (MIN: min_elt_restr (set_collect fs g ∩₁ QueueLib.has_loc l) toimm⁺ m),
                 exists mi ni, << Gmi: g mi >> /\ << Gni: g ni >> /\
                               << FSmi: fs mi = m >> /\ << Lmi: LockLib.is_lock_ok (snd l) mi >> /\
                               << RFmn: rf ni mi >> /\ << NL: LockLib.is_newlock_ok (snd l) ni >>).
  { ins; clear C_ALLOC TObase IRRto FUNto FSUPPto SAMEloc.
    unfolder in M; desc.
    assert (M' := M0); eapply has_lock_event in M0; eauto; try solve [vauto]; ins; desc.
    exists ai.
    specialize (LCONS _ (or_intror eq_refl)); ins; red in LCONS; ins; desc.
    forward apply (set_equiv_inl ai CODOM) as A.
      by unfolder; splits; vauto; red in LOCKai; desf; red; ins; eauto.
    unfolder in A; desc.
    (* subst z z0. *)
    assert (Gx: g x) by (apply RF in A; eapply A).
    exists x; splits; eauto. 
    assert (RF' := A); eapply show_restr_rel, RFDOM0 in A; try done.
    unfolder in A; desc.
    destruct A2 as [NL | U]; try done.
    all: forward apply (@LockLib.unique_loc ai (snd l) a); try (by vauto); ins; subst a; try done.
    enough (toimm (fs x) m).
    { clear TOIMM.
      exfalso; eapply MIN; eauto using t_step.
      split.
        by red; eauto.
      eapply queue_from_lock in U; eauto; desf.
      tertium_non_datur (fst l = q).
        by (destruct l; ins; congruence).
      exfalso.
      rewrite <- FSai in *.
      forward eapply C_LOC with (a := ai) (b := x) (libi := LockLib.library); eauto.
        1-2: by right; eauto.
        by destruct l; ins; congruence.
      clear.
      unfolder; intro H; eapply H. 
      all: by constructor; exists eq_refl.
    }
    desf; clear MIN; splits; eauto.
    exists x, ai; splits; eauto.
    by red in U; desf.
    by red in A3; desf.
  }
  assert (TOT: forall l, strict_total_order (set_collect fs g ∩₁ QueueLib.has_loc l) toimm^+).
  { clear TOIMM; ins; repeat split; eauto with hahn.
    eapply partial_order_is_total_sufficient; 
    eauto using irreflexive_restr, transitive_restr, fsupp_restr with hahn.
    * rewrite clos_trans_restr, immediate_clos_trans.
        1,3: by apply functional_restr.
        by apply acyclic_restr.
      red; ins; unfolder in H; desf.
      split.
      by eapply TObase; vauto.
      by eapply SAMEloc; vauto.
    * clear TObase ESEM.
      forward eapply well_founded_nonempty_has_min 
              with (R:=toimm⁺) (cond:= set_collect fs g ∩₁ QueueLib.has_loc l) as MIN.
        by apply wf_restr, fsupp_well_founded.
      desf; eauto.
      right; eexists; splits; eauto.
      split; red; ins; desf; eauto.
      red in MIN; desf.
      apply TOmin in MIN; eauto; desf.
      assert (Cx := min_elt_restr_in H).
      unfolder in Cx; desf.
      apply TOmin in H; eauto; desf.
        2: by unfolder; eauto.
      rewrite <- FSmi in *.
      enough (ni = ni0).
      { subst; f_equal.
        specialize (LCONS _ (or_intror eq_refl)); ins; red in LCONS; ins; desf.
        eapply RFFUN; eapply show_restr_rel; eauto.
        all: red; red in Lmi, Lmi0, NL; destruct mi, mi0, ni0; ins; desf; eauto.
      }
      specialize (LCONS _ (or_intror eq_refl)); ins; red in LCONS; ins; desc.
      eapply ALLOC; unfolder; splits; eauto.
      all: red; ins; red in NL, NL0; desf; eauto.
  }
      
  clear TOIMM. (* by now we have all needed facts about toimm *)    
  
  assert (TOporf: forall ai bi (Gai: g ai) (Gbi: g bi) (TOab: toimm⁺ (fs ai) (fs bi)), (po ∪ rf)⁺ ai bi).
  { clear - GDOM ESEM TObase; ins. 
    apply TObase in TOab; clear TObase; desf.
    rename ai0 into a_unlock; rename bi0 into b_lock.
    forward eapply po_to_unlock with (ai := ai) (ulcki := a_unlock) as A; eauto.
    forward eapply po_from_lock with (ai := bi) (lcki := b_lock) as B; eauto.
    eapply rt_t_trans, t_rt_trans; eauto; eapply clos_refl_trans_mon; eauto; vauto.
  } 
  
  assert (TOposo: forall ai bi (Gai: g ai) (Gbi: g bi) (TOab: toimm⁺ (fs ai) (fs bi)), (po ∪ so)⁺ ai bi).
  { clear - GDOM ESEM TObase; ins. 
    apply TObase in TOab; clear TObase; desf.
    rename ai0 into a_unlock; rename bi0 into b_lock.
    forward eapply po_to_unlock with (ai := ai) (ulcki := a_unlock) as A; eauto.
    forward eapply po_from_lock with (ai := bi) (lcki := b_lock) as B; eauto.
    eapply rt_t_trans, t_rt_trans; eauto; eapply clos_refl_trans_mon; eauto; vauto.
  } 


  assert (RC11rfTO: forall w r (Gw: g w) (Gr: g r) (wLIB: (QueueLib.is_enqueue ∪₁ QueueLib.is_dequeue) (fs w))
                      (rLIB: (QueueLib.is_enqueue ∪₁ QueueLib.is_dequeue) (fs r)) 
                      (C11w: lib_events RC11.library w) (C11r: lib_events RC11.library r) (RF: rf w r),
                    toimm⁺ (fs w) (fs r)).
  { clear - TOporf IRRto TOT C_LOC GDOM ESEM LCONS IRR'; ins.
    assert (exists l, QueueLib.has_loc l (fs w) /\ QueueLib.has_loc l (fs r)).
    { specialize (LCONS _ (or_introl eq_refl)); ins; red in LCONS; ins; desc.
      eapply show_restr_rel, RFLOC in RF; try done.
      unfolder in RF; desf.
      eapply queue_from_access in RF0; eauto; desc.
      eapply queue_from_access in RF1; eauto; desf.
      all: try by destruct (fs w), (fs r); red in RF0, RF1; do 2 (red in wLIB, rLIB; desf; ins); desf.
      enough (lck = lck0) by (desf; eexists; eauto).
      apply NNPP; intro NEQ.
      forward eapply C_LOC with (a := w) (b := r) (libi := RC11.library); eauto.
        1-2: by right; eauto. 
        congruence.
      unfolder; clear; simpls; intro D.
      by apply (D a); constructor; exists eq_refl.
    }
    desf.
    specialize (TOT l).
    red in TOT; desc.
    forward eapply TOT0 with (a := fs w) (b := fs r).
    1-2: by unfolder; splits; eauto; apply gS; red; ins; eauto.
    * intro EQ.
      eapply IRR', t_trans; vauto.
      apply t_step; left.
      specialize (LCONS _ (or_introl eq_refl)); ins; red in LCONS; ins; desc.
      assert (WR := RF); eapply show_restr_rel, RFOK in WR; eauto.
      destruct WR as [W R].
      eapply show_restr_rel, RFLOC in RF; eauto.
      clear - W R RF EQ EQ Gw Gr IRR' ESEM wLIB.
      unfolder in RF; desf.
      rename a into l.
      specialize (ESEM _ Gw).
      destruct (fs w) eqn: FSw; unfolder in wLIB; desf; red in wLIB; ins; desf.
      all: unfold env_compose, env_of_impl in ESEM; ins; desf.
      all: specialize_full ESEM; eauto; rewrite <- eq_rect_eq in *; ins.
      all: unfold idom_seq in ESEM; desf; ins; unfold idom_seq in *; desf.
      all: rewrite_gpos.
      all: forward apply (set_equiv_inr w gEQ) as INw; try done.
      all: forward apply (set_equiv_inr r gEQ) as INr; try done.
      all: unfolder in INw; desf.
      all: try by match goal with O: RC11.is_write ?x |- _ 
                  => match goal with C: event_of_call _ _ _ x |- _ 
                     => clear - C O; red in C; desf; destruct x; do 2 (red in O; desf; ins); ins; desf 
                     end 
                  end.
      all: unfolder in INr; desf.
      all: try by match goal with O: RC11.is_read ?x |- _ 
                  => match goal with C: event_of_call _ _ _ x |- _ 
                     => clear - C O; red in C; desf; destruct x; do 2 (red in O; desf; ins); ins; desf 
                     end 
                  end.
      all: try by forward eapply (same_relation_inl r w poEQ) as POrw; 
                  [unfolder; eauto 10 | unfolder in POrw; desf].
      clear_gpos.
      unfold event_of_call in *; desf; ins.
      destruct r, w, a, a0, a1, a2, l; red in RF0, RF1; ins; desf; rewrite <- !eq_rect_eq in *; ins.
      inv RF0; omega.
    * ins; desf. 
      exfalso; apply TOporf in H1; eauto. 
      eapply IRR', t_trans; eauto. 
      eapply inclusion_step_t; eauto with hahn.
  }

  assert (WF: well_founded (fun xi yi => g xi /\ g yi /\ toimm⁺ (fs xi) (fs yi))).
  { 
    assert (WF: well_founded toimm⁺) by (apply fsupp_well_founded; eauto with hahn). 
    clear -WF.
    red; ins; remember (fs a) as b; generalize dependent a.
    induction b using (well_founded_ind WF).
    constructor; ins; desf; eapply H; eauto.
  }
  
  assert (LSTIDXw: forall q ai (Gai: g ai) (LOADai: RC11.is_read ai) (LOC: RC11.has_loc (q,1) ai), 
            exists wi, << Gwi: g wi >> /\ << RF: rf wi ai >> /\ << LOC: RC11.has_loc (q,1) wi >> /\ 
            << STOREwi: RC11.is_store wi >> /\
            ((<< MIN: forall e (DEQ: QueueLib.is_enqueue e), ~ toimm⁺ e (fs ai) >> /\
              << OPwi: ev_op (fs wi) = op_newqueue >>)
             \/
            (<< TOw: toimm⁺ (fs wi) (fs ai) >> /\ << ENQw: QueueLib.is_enqueue (fs wi) >> /\
             << IMM: forall e (DEQ: QueueLib.is_enqueue e), ~ (toimm⁺ (fs wi) e /\ toimm⁺ e (fs ai)) >> /\
             << OPwi: ev_op (fs wi) = op_enqueue>>))).
 { ins; clear TOmin.
    specialize (LCONS _ (or_introl eq_refl)); ins; red in LCONS; ins; desf.
    forward apply (set_equiv_inl ai CODOM) as W.
      by unfolder; splits; try done; red; destruct ai; do 2 (red in LOADai; ins; desf); eauto.
    destruct W as [wi W]; desf.
    specialize (RFOK _ _ W); unfolder in RFOK; desf.
    specialize (RFLOC _ _ W); unfolder in RFLOC; desf.
    rewrite (RC11.has_loc_unique RFLOC1 LOC) in *.
    unfolder in W; desf.
    assert (Gwi: g wi) by (by apply RF in W; unfolder in W; desf).
    assert (<< STOREwi: RC11.is_store wi >> /\ << ND: ~ ev_op (fs wi) = op_dequeue >>).
    { specialize (ESEM _ Gwi).
      specialize (GDOM _ Gwi).
      destruct (fs wi) eqn:FSwi; ins; red in GDOM; ins; desf.
      all: unfold env_compose, env_of_impl in ESEM; ins; desf.
      all: specialize_full ESEM; eauto.
      all: rewrite <- eq_rect_eq in *; ins.
      all: unfold idom_seq in ESEM; desf; ins; unfold idom_seq in *; desf.
      all: clear_gpos; rewrite_gpos.
      all: forward apply (set_equiv_inr wi gEQ) as WW; try done.
      all: unfolder in WW; desf.
      all: unfold event_of_call in *; desf.
      all: destruct wi; do 2 (red in RFOK; ins; desf).
      exfalso.
      destruct a0, a1, a2, a3, a4; ins; desf; rewrite <- !eq_rect_eq in *; ins.
      clear - RFLOC0; red in RFLOC0; desf; ins.
      by rewrite <- eq_rect_eq in *.
    }
    exists wi; desf; splits; try done. 

    assert (PE: forall e (ENQe: QueueLib.is_enqueue e) (PREV: toimm⁺ e (fs ai)), 
                 exists ei, <<Gei: g ei>> /\ <<FSei: fs ei = e>> /\ 
                   <<STOREei: RC11.is_store ei>> /\ <<LOCei: RC11.has_loc (q,1) ei>>).
    { ins.
      forward eapply queue_from_access with (ai:=ai) (l := (q, 1)) as QfromACCESS; eauto; ins; desf.
        by red in QfromACCESS; apply TObase in PREV; desf; destruct (fs ai); 
           unfolder in bLIB; ins; desf.
      assert (EE: exists l, QueueLib.is_enqueue_loc l e).
         by red in ENQe; destruct e; ins; desf; vauto.
      desf; specialize_full SAMEloc; vauto.
      rewrite (QueueLib.has_loc_unique SAMEloc QfromACCESS) in *.
      assert (Ei: exists ei, g ei /\ e = fs ei) by (apply TObase in PREV; desf; eauto).
      desf; specialize (ESEM _ Ei).
      destruct (fs ei) eqn:FSei; ins; red in EE; ins; desf.
      unfold env_compose, env_of_impl in ESEM; ins; desf.
      specialize_full ESEM; eauto.
      rewrite <- eq_rect_eq in *; ins.
      unfold idom_seq in ESEM; desf; ins; unfold idom_seq in *; desf.
      clear_gpos; rewrite_gpos.
      exists a2; splits.
      1-2: by apply gEQ; unfolder; eauto.
      all: unfold event_of_call in *.
      all: destruct ev_args, a0, a1, a2, a3, a4; ins; desf; rewrite <- !eq_rect_eq in *.
      by repeat constructor; exists eq_refl.
    }
    assert (aLIB := GDOM _ Gai).
    assert (wLIB := GDOM _ Gwi); red in wLIB; ins; desf; rewrite <- wLIB in *; try done.
    * left; splits; ins.
      intro TOea; specialize_full PE; eauto; desf.
      red in ENCAPS.
      forward eapply queue_from_access with (ai := ei) as LOCe; eauto; desf.
        by red in LOCe; apply TObase in TOea; desf; destruct (fs ei); unfolder in aLIB; ins; desf.
      forward eapply (ENCAPS (fs ei)) as ALL; try by vauto.
      desf; ins.
      forward eapply queue_from_access with (ai := wi) as NEWQ; eauto; ins; desf.
        2: by destruct (fs wi); do 3 (red in NEWQ; desf; ins).
      unfolder in Ga; desf.
      assert (lck = lck0).
      { apply NNPP; intro NEQ.
        forward eapply C_LOC with (a := wi) (b := y) (libi := RC11.library); eauto.
          1-2: by left; eauto.
          congruence.
        clear; intro D; apply (D (q,0)).
        all: constructor; exists eq_refl; ins.
      }
      subst lck0.
      forward eapply C_ALLOC with (a := y) (b := wi) as EQ; ins; eauto.
      rewrite EQ in *.
      specialize_full TOporf; try exact TOea; eauto.
      specialize_full TOposo; try exact TOea; eauto.
      clear TOea.
        eapply (ACYC ei), t_trans with ai; apply t_step.
        - repeat left; red; splits; ins.
            by eapply clos_trans_mori, TOposo; eauto with hahn.
          by red; ins; eauto.
        - repeat right; unfolder.
          split; [ | eby intro; desf; eapply IRR' ].
          eexists; splits; eauto.
          forward eapply MOT with (a := wi) (b := ei).
          1-2: by unfolder; splits; vauto; red; ins; eauto. 
          -- intro; desf. eapply ACYC, t_step.
             repeat left; red; splits.
             by apply t_step; repeat right; vauto.
             all: by red; ins; eauto.
          -- intro M; desf.
             exfalso; eapply ACYC, t_trans; apply t_step.
               eby left; left; right.
             repeat left; red; splits.
             by apply t_step; repeat right; vauto.
             all: by red; ins; eauto.

    * right; splits; ins.
      + eapply RC11rfTO; vauto.
        clear - LOADai aLIB Gai ESEM.
        red in aLIB; ins; desf; vauto.
        specialize (ESEM _ Gai).
        destruct (fs ai) eqn:FSai; ins; desf.
        unfold env_compose, env_of_impl in ESEM; ins; desf.
        specialize_full ESEM; eauto; ins; rewrite <- eq_rect_eq in *; ins.
        unfold idom_seq in ESEM; ins; desf.
        clear_gpos; rewrite_gpos.
        forward eapply (set_equiv_inr ai gEQ) as A; try done.
        unfolder in A; desf.
        all: unfold event_of_call in *.
        all: destruct ai; do 2 (red in LOADai; ins; desf); ins; desf.
      + intro TOe; desf; specialize_full PE; eauto; desf.
        specialize_full TOporf; try exact TOe0; eauto.
        apply TOposo in TOe; try done.
        apply TOposo in TOe0; try done.
        eapply (ACYC ei), t_trans; apply t_step.
        - repeat left; red; repeat split.
          by eapply inclusion_t_t; try exact TOe0; eauto with hahn.
          all: red; ins; eauto.
        - repeat right; unfolder.
          split; [ | eby intro; desf; eapply IRR' ].
          eexists; splits; eauto.
          forward eapply MOT with (a := wi) (b := ei).
          1-2: by unfolder; splits; vauto; red; ins; eauto. 
          -- intro; desf. eapply ACYC, t_step.
             repeat left; red; splits.
             by eapply inclusion_t_t; try eassumption; eauto with hahn.
             all: by red; ins; eauto.
          -- intro M; desf.
             exfalso; eapply ACYC, t_trans; apply t_step.
               eby left; left; right.
             repeat left; red; splits.
             by eapply inclusion_t_t; try eassumption; eauto with hahn.
             all: by red; ins; eauto.
  }
  
  assert (FSTIDXw: forall q ai (Gai: g ai) (LOADai: RC11.is_read ai) (LOC: RC11.has_loc (q,2) ai), 
            exists wi, << Gwi: g wi >> /\ << RF: rf wi ai >> /\ << LOC: RC11.has_loc (q,2) wi >> /\ 
            << STOREwi: RC11.is_store wi >> /\
            ((<< MIN: forall e v (DEQ: QueueLib.is_dequeueV' v e), ~ toimm⁺ e (fs ai) >> /\
              << OPwi: ev_op (fs wi) = op_newqueue >>)
             \/
            (<< TOw: toimm⁺ (fs wi) (fs ai) >> /\ << DEQw: QueueLib.is_dequeue (fs wi) >> /\
             << IMM: forall e v (DEQ: QueueLib.is_dequeueV' v e), ~ (toimm⁺ (fs wi) e /\ toimm⁺ e (fs ai)) >> 
             /\ << OPwi: ev_op (fs wi) = op_dequeue>>))).
  { ins; clear TOmin.
    specialize (LCONS _ (or_introl eq_refl)); ins; red in LCONS; ins; desf.
    forward apply (set_equiv_inl ai CODOM) as W.
      by unfolder; splits; try done; red; destruct ai; do 2 (red in LOADai; ins; desf); eauto.
    destruct W as [wi W]; desf.
    specialize (RFOK _ _ W); unfolder in RFOK; desf.
    specialize (RFLOC _ _ W); unfolder in RFLOC; desf.
    rewrite (RC11.has_loc_unique RFLOC1 LOC) in *.
    unfolder in W; desf.
    assert (Gwi: g wi) by (by apply RF in W; unfolder in W; desf).
    assert (<< STOREwi: RC11.is_store wi >> /\ << ND: ~ ev_op (fs wi) = op_enqueue >>).
    { specialize (ESEM _ Gwi).
      specialize (GDOM _ Gwi).
      destruct (fs wi) eqn:FSwi; ins; red in GDOM; ins; desf.
      all: unfold env_compose, env_of_impl in ESEM; ins; desf.
      all: specialize_full ESEM; eauto.
      all: rewrite <- eq_rect_eq in *; ins.
      all: unfold idom_seq in ESEM; desf; ins; unfold idom_seq in *; desf.
      all: clear_gpos; rewrite_gpos.
      all: forward apply (set_equiv_inr wi gEQ) as WW; try done.
      all: unfolder in WW; desf.
      all: unfold event_of_call in *.
      all: destruct wi; do 2 (red in RFOK; ins; desf).
      all: destruct a0, a3, a4; try destruct a1; try destruct a2; ins; desf; rewrite <- !eq_rect_eq in *; ins.
      clear - RFLOC0; red in RFLOC0; desf; ins; rewrite <- eq_rect_eq in *; ins.
      all: clear - RFLOC0; red in RFLOC0; desf; ins.
      rewrite <- eq_rect_eq in *; ins; inv RFLOC0; omega.
    }
    exists wi; desf; splits; try done. 
    assert (PD: forall e v (DEQe: QueueLib.is_dequeueV' v e) (PREV: toimm⁺ e (fs ai)), 
                 exists ei, <<Gei: g ei>> /\ <<FSei: fs ei = e>> /\ 
                   <<STOREei: RC11.is_store ei>> /\ <<LOCei: RC11.has_loc (q,2) ei>>).
    { ins.
      forward eapply queue_from_access with (ai:=ai) as Q; eauto; desf.
        by red in Q; apply TObase in PREV; desf; destruct (fs ai); unfolder in bLIB; ins; desf.
      assert (EE: exists l, QueueLib.is_dequeueV l v e).
        by red in DEQe; destruct e; ins; desf; eexists; exists eq_refl; ins; vauto.
      simpls; desf.
      desf; specialize_full SAMEloc; try eassumption. 
        by clear Q; apply QueueLib.is_dequeue_impl in EE; red; ins; unfolder; eauto.
      rewrite (QueueLib.has_loc_unique SAMEloc Q) in *.
      assert (Ei: exists ei, g ei /\ e = fs ei) by (apply TObase in PREV; desf; eauto).
      desf; specialize (ESEM _ Ei).
      destruct (fs ei) eqn:FSei; ins; red in EE; ins; desf.
      unfold env_compose, env_of_impl in ESEM; ins; desf.
      specialize_full ESEM; eauto.
      rewrite <- eq_rect_eq in *; ins.
      unfold idom_seq in ESEM; desf; ins; unfold idom_seq in *; desf.
      clear_gpos; rewrite_gpos.
      exists a5; splits; try done.
      1-2: by apply gEQ; unfolder; eauto.
      all: unfold event_of_call in *.
      all: destruct ev_args, a0, a1, a2, a3, a4, a5; ins; desf; rewrite <- !eq_rect_eq in *.
      by repeat constructor; exists eq_refl.
    }
    assert (aLIB := GDOM _ Gai).
    assert (wLIB := GDOM _ Gwi); red in wLIB; ins; desf; rewrite <- wLIB in *; try done.
    * left; splits; ins.
      intro TOea; specialize_full PD; eauto; desf.
      red in ENCAPS.
      forward eapply queue_from_access with (ai := ei) as LOCe; eauto; desf. 
        by red in LOCe; apply TObase in TOea; desf; destruct (fs ei); unfolder in aLIB0; ins; desf.
      forward eapply (ENCAPS (fs ei)) as ALL; try by vauto.
        by red in DEQ; desf; red; ins; eauto.
      desf; ins.
      forward eapply queue_from_access with (ai := wi) as NEWQ; eauto; ins; desf.
        2: by destruct (fs wi); do 3 (red in NEWQ; desf; ins).
      unfolder in Ga; desf.
      assert (lck = lck0).
      { apply NNPP; intro NEQ.
        forward eapply C_LOC with (a := wi) (b := y) (libi := RC11.library); eauto.
          1-2: by left; eauto.
          congruence.
        clear; intro D; apply (D (q,0)).
        all: constructor; exists eq_refl; ins.
      }
      subst lck0.
      forward eapply C_ALLOC with (a := y) (b := wi) as EQ; ins; eauto.
      rewrite EQ in *.
      specialize_full TOporf; try exact TOea; eauto.
        apply TOposo in TOea; try done.
        eapply (ACYC ei), t_trans; apply t_step.
        - repeat left; red; repeat split.
          by eapply inclusion_t_t; try exact TOea; eauto with hahn.
          all: red; ins; eauto.
        - repeat right; unfolder.
          split; [ | eby intro; desf; eapply IRR' ].
          eexists; splits; eauto.
          forward eapply MOT with (a := wi) (b := ei).
          1-2: by unfolder; splits; vauto; red; ins; eauto. 
          -- intro; desf. eapply ACYC, t_step.
             repeat left; red; splits.
             by apply t_step; repeat right; vauto.
             all: by red; ins; eauto.
          -- intro M; desf.
             exfalso; eapply ACYC, t_trans; apply t_step.
               eby left; left; right.
             repeat left; red; splits.
             by apply t_step; repeat right; vauto.
             all: by red; ins; eauto.
    * right; splits; ins.
      + eapply RC11rfTO; vauto.
        clear - LOADai aLIB Gai ESEM.
        red in aLIB; ins; desf; vauto.
        specialize (ESEM _ Gai).
        destruct (fs ai) eqn:FSai; ins; desf.
        unfold env_compose, env_of_impl in ESEM; ins; desf.
        specialize_full ESEM; eauto; ins; rewrite <- eq_rect_eq in *; ins.
        unfold idom_seq in ESEM; ins; desf.
        clear_gpos; rewrite_gpos.
        forward eapply (set_equiv_inr ai gEQ) as A; try done.
        unfolder in A; desf.
        all: unfold event_of_call in *.
        all: destruct ai; do 2 (red in LOADai; ins; desf); ins; desf.
      + intro TOe; desf; specialize_full PD; eauto; desf.
        specialize_full TOporf; try exact TOe0; eauto.
        apply TOposo in TOe; try done.
        apply TOposo in TOe0; try done.
        eapply (ACYC ei), t_trans; apply t_step.
        - repeat left; red; repeat split.
          by eapply inclusion_t_t; try exact TOe0; eauto with hahn.
          all: red; ins; eauto.
        - repeat right; unfolder.
          split; [ | eby intro; desf; eapply IRR' ].
          eexists; splits; eauto.
          forward eapply MOT with (a := wi) (b := ei).
          1-2: by unfolder; splits; vauto; red; ins; eauto. 
          -- intro; desf. eapply ACYC, t_step.
             repeat left; red; splits.
             by eapply inclusion_t_t; try eassumption; eauto with hahn.
             all: by red; ins; eauto.
          -- intro M; desf.
             exfalso; eapply ACYC, t_trans; apply t_step.
               eby left; left; right.
             repeat left; red; splits.
             by eapply inclusion_t_t; try eassumption; eauto with hahn.
             all: by red; ins; eauto.
  }
  
  remember (fun l e => match excluded_middle_informative ((set_collect fs g ∩₁ QueueLib.has_loc l) e) with
                         | left L => length (filterP QueueLib.is_enqueue 
                                               (starting_segment (TOT l) FSUPPto e L))
                         | _      => 0
                       end) as which_enq eqn:WE.

  assert (WEval: forall l ai (Gai: g ai) (LOCa: (QueueLib.is_newqueue_loc l ∪₁ QueueLib.has_loc l) (fs ai)) 
                       (STOREai: RC11.is_write ai) (LOC: RC11.has_loc ((fst l), 1) ai), 
                   RC11.has_wval (Z.of_nat (which_enq l (fs ai))) ai).
  { ins; clear C_ALLOC TOmin TObase.
    specialize (GDOM _ Gai); red in GDOM; ins.
    destruct GDOM as [opEQ | opEQ].
    { (* If (fs ai) is newqueue. Unimportant case. *)
      clear LCONS; desf. 
        by unfolder in s; desf; destruct (fs ai); ins; desf; repeat (red in s0; desf).
      clear n; specialize (ESEM _ Gai).
      destruct (fs ai) eqn:FSai; ins; desf.
      unfold env_compose, env_of_impl in ESEM; ins; desf; eauto.
      rewrite <- eq_rect_eq in *.
      specialize_full ESEM; eauto; ins.
      unfold idom_seq in ESEM; desf.
      clear_gpos; rewrite_gpos.
      forward eapply (set_equiv_inr ai gEQ) as A; try done.
      unfold event_of_call in *.
      unfolder in A; destruct ai; red in STOREai, LOC; ins; desf; ins; rewrite <- eq_rect_eq in *; ins; desf.
      destruct a, a1, a2; ins; desf; rewrite <- eq_rect_eq; ins.
      all: by constructor; exists eq_refl.
    }
    induction ai using (well_founded_induction WF); rename H into IH.
    assert (<< ENQa: QueueLib.is_enqueue (fs ai) >> /\
            exists ai', <<Gai': g ai'>> /\ <<FSai': fs ai' = fs ai >> /\ <<LOADai': ev_op ai' = C11_load>> /\ 
                        <<LOCai': RC11.has_loc (fst l, 1) ai'>> /\
                        <<VAL: exists v, RC11.has_wval (v + 1) ai /\ RC11.has_rval v ai'>>).
    { specialize (ESEM _ Gai).
      destruct (fs ai) eqn:FSai; ins; desf.
      all: unfold env_compose, env_of_impl in ESEM; ins; desf.
      all: specialize_full ESEM; eauto.
      all: rewrite <- eq_rect_eq in *; ins.
      all: unfold idom_seq in ESEM; desf; ins; unfold idom_seq in *; desf.
      all: clear_gpos; rewrite_gpos.
      all: forward apply (set_equiv_inr ai gEQ) as A; try done.
      all: unfold event_of_call in *. 
      all: unfolder in A; desf; try by destruct ai; repeat (red in STOREai; desf); ins; desf.
      * splits; try done.           
        exists a2; splits; vauto.
        1-2: by apply gEQ; unfolder; eauto.
        all: red in LOC; destruct ev_args, a0, ai, a2, a3; ins; desf; rewrite <- !eq_rect_eq in *; ins.
        by repeat constructor; exists eq_refl.
        by eexists; splits; repeat constructor; exists eq_refl.
      * red in LOC; destruct ev_args, a, a1, ai, a2, a3; ins; desf; rewrite <- !eq_rect_eq in *; ins.
        inv LOC; omega.
      * red in LOC; destruct ev_args, a, a1, ai, a2, a3; ins; desf; rewrite <- !eq_rect_eq in *; ins.
    }
    clear opEQ; desc.
    specialize_full LSTIDXw; try exact LOADai'; eauto; desc.
    assert (VALw: RC11.has_wval v wi).
    { clear - RF0 STOREwi LOADai' Gwi Gai' VAL0 LCONS.
      specialize (LCONS _ (or_introl eq_refl)); ins; red in LCONS; ins; desf.
      eapply show_restr_rel, RFVAL in RF0.
      2-3: by red; ins; eauto.
      unfolder in RF0; desf.
      by destruct ai'; red in VAL0, RF2; ins; desf; rewrite <- eq_rect_eq in *.
    }
    destruct LOCa as [LOCa|LOCa]. (* to note that we are not dealing with a newqueue *)
      by red in LOCa, ENQa; destruct (fs ai); ins; desf. 
    destruct LSTIDXw1; desc.
    * (* base case *)
      desf.
        2: by destruct n; split; try done; red; eauto.
      match goal with [ |- context[length ?l] ] => assert (EQ: l = (fs ai) :: nil) end.
      { apply nodup_eq_one; ins.
          by apply nodup_filterP, NoDup_starting_segment.
          by apply in_filterP_iff; split; eauto using starting_segment_contains_endpoint. 
        apply in_filterP_iff in IN; desf.
        apply starting_segment_property in IN; desf.
        apply clos_refl_transE in IN1; desf.
        hahn_rewrite ct_of_ct in IN1.
        eby exfalso; rewrite FSai' in MIN; eapply MIN.
      }
      rewrite EQ; ins.
      specialize (ESEM _ Gwi).
      destruct (fs wi) eqn:FSwi; ins; desf.
      unfold env_compose, env_of_impl in ESEM; ins; desf.
      specialize_full ESEM; eauto; ins.
      rewrite <- eq_rect_eq in *; ins.
      unfold idom_seq in ESEM; ins; desf.
      clear_gpos; rewrite_gpos.
      forward apply (set_equiv_inr wi gEQ) as W; try done.
      unfold event_of_call in *.
      unfolder in W; desf; destruct wi; 
        try destruct a; try destruct a0; try destruct a1; try destruct a2; ins; desf.
      all: rewrite <- eq_rect_eq in *.
        2: by red in LOC0; desf; ins; rewrite <- eq_rect_eq in *.
      by red in VALw; desf; ins; rewrite <- eq_rect_eq in *.
    * (* inductive case *)
      desf.
        2: by destruct n; split; try done; red; eauto.
      rewrite FSai' in TOw.
      assert (W: (set_collect fs g ∩₁ QueueLib.has_loc l) (fs wi)).
        eby split; vauto; eapply SAMEloc'.
      specialize_full IH; try exact TOw; eauto; desf.
        by right; eauto.
      erewrite starting_segment_decomp; eauto using rt_step.
      rewrite filterP_app, app_length.
      rewrite Nat2Z.inj_add, (RC11.functional_has_wval IH VALw ).
      match goal with [ |- context[length ?l]] => assert (EQ1: l = fs ai :: nil) end.
      { rewrite FSai' in *;  clear - IMM ENQa TOw.
        assert (T := TOT l); do 2 (red in T; desc).
        eapply StronglySorted_eq; vauto.
          by apply StronglySorted_filterP, segment_property.
        split; ins; desf.
        * apply in_filterP_iff in H; desc.
          apply segment_property in H; desc.
          apply clos_refl_transE in H2; desf; eauto.
          hahn_rewrite ct_of_ct in H2; firstorder.
        * apply in_filterP_iff; split; eauto.
          apply segment_property; splits; vauto.
      }
      by rewrite EQ1. 
  }






  remember (fun l e => match excluded_middle_informative ((set_collect fs g ∩₁ QueueLib.has_loc l) e) with
                         | left L => length (filterP (fun e => exists v, QueueLib.is_dequeueV l v e)
                                               (starting_segment (TOT l) FSUPPto e L))
                         | _      => 0
                       end) as which_deq eqn:WD.

  assert (WDval: forall l ai (Gai: g ai) (LOCa: (QueueLib.is_newqueue_loc l ∪₁ QueueLib.has_loc l) (fs ai)) 
                       (STOREai: RC11.is_write ai) (LOC: RC11.has_loc ((fst l), 2) ai), 
                   RC11.has_wval (Z.of_nat (which_deq l (fs ai))) ai).
  { ins; clear C_ALLOC TOmin TObase.
    specialize (GDOM _ Gai); red in GDOM; ins.
    destruct GDOM as [opEQ | opEQ].
    { (* If (fs ai) is newqueue. Unimportant case. *)
      clear LCONS; desf. 
        by unfolder in s; desf; destruct (fs ai); ins; desf; repeat (red in s0; desf).
      clear n; specialize (ESEM _ Gai).
      destruct (fs ai) eqn:FSai; ins; desf.
      unfold env_compose, env_of_impl in ESEM; ins; desf; eauto.
      rewrite <- eq_rect_eq in *.
      specialize_full ESEM; eauto; ins.
      unfold idom_seq in ESEM; desf.
      clear_gpos; rewrite_gpos.
      forward eapply (set_equiv_inr ai gEQ) as A; try done.
      unfold event_of_call in *.
      unfolder in A; destruct ai; red in STOREai, LOC; ins; desf; ins; rewrite <- eq_rect_eq in *; ins; desf.
      destruct a0, a1, a2; ins; desf; rewrite <- eq_rect_eq; ins.
      all: by constructor; exists eq_refl.
    }
    induction ai using (well_founded_induction WF); rename H into IH.
    assert (<< DEQa: exists v', QueueLib.is_dequeueV' v' (fs ai) >> /\
            exists ai', <<Gai': g ai'>> /\ <<FSai': fs ai' = fs ai >> /\ <<LOADai': ev_op ai' = C11_load>> /\ 
                        <<LOCai': RC11.has_loc (fst l, 2) ai'>> /\
                        <<VAL: exists v, RC11.has_wval (v + 1) ai /\ RC11.has_rval v ai'>>).
    { specialize (ESEM _ Gai).
      destruct (fs ai) eqn:FSai; ins; desf.
      all: unfold env_compose, env_of_impl in ESEM; ins; desf.
      all: specialize_full ESEM; eauto.
      all: rewrite <- eq_rect_eq in *; ins.
      all: unfold idom_seq in ESEM; desf; ins; unfold idom_seq in *; desf.
      all: clear_gpos; rewrite_gpos.
      all: forward apply (set_equiv_inr ai gEQ) as A; try done.
      all: unfold event_of_call in *.
      all: unfolder in A; desf; try by destruct ai; repeat (red in STOREai; desf); ins; desf.
      * red in LOC; destruct ev_args, a, a0, ai, a2, a3; ins; desf; rewrite <- !eq_rect_eq in *; ins.
      * red in LOC; destruct ev_args, a, a1, ai, a2, a3; ins; desf; rewrite <- !eq_rect_eq in *; ins.
        inv LOC; omega.
      * splits; try done.
        + eexists _, eq_refl; desf.
        + exists a; splits; vauto.
          1-2: by apply gEQ; unfolder; eauto.
          all: red in LOC; destruct ev_args, a, a1, a0, ai, a2, a3; ins; desf; rewrite <- !eq_rect_eq in *; ins.
          by repeat constructor; exists eq_refl.
          by eexists; splits; repeat constructor; exists eq_refl.
    }
    clear opEQ; desc.
    specialize_full FSTIDXw; try exact LOADai'; eauto; desc.
    assert (VALw: RC11.has_wval v wi).
    { clear - RF0 STOREwi LOADai' Gwi Gai' VAL0 LCONS.
      specialize (LCONS _ (or_introl eq_refl)); ins; red in LCONS; ins; desf.
      eapply show_restr_rel, RFVAL in RF0.
      2-3: by red; ins; eauto.
      unfolder in RF0; desf.
      by destruct ai'; red in VAL0, RF2; ins; desf; rewrite <- eq_rect_eq in *.
    }
    destruct LOCa as [LOCa|LOCa]. (* to note that we are not dealing with a newqueue *)
      by red in LOCa, DEQa; destruct (fs ai); ins; desf.
    destruct FSTIDXw1; desc.
    * (* base case *)
      desf.
        2: by destruct n; split; try done; red; eauto.
      match goal with [ |- context[length ?l] ] => assert (EQ: l = (fs ai) :: nil) end.
      { apply nodup_eq_one; ins.
          by apply nodup_filterP, NoDup_starting_segment. 
          { apply in_filterP_iff; split; eauto using starting_segment_contains_endpoint.
            clear - DEQa LOCa.
            destruct (fs ai); repeat (red in LOCa; desf; ins); red in DEQa; ins; desf.
            rewrite <- eq_rect_eq in *; eexists _, eq_refl; ins; desf. }
        apply in_filterP_iff in IN; desf.
        apply starting_segment_property in IN; desf.
        apply clos_refl_transE in IN1; desf.
        hahn_rewrite ct_of_ct in IN1.
        exfalso; rewrite FSai' in MIN; eapply MIN; try exact IN1. 
        eby eapply QueueLib.is_dequeueV_weaken.
      }
      rewrite EQ; ins.
      specialize (ESEM _ Gwi).
      destruct (fs wi) eqn:FSwi; ins; desf.
      unfold env_compose, env_of_impl in ESEM; ins; desf.
      specialize_full ESEM; eauto; ins.
      rewrite <- eq_rect_eq in *; ins.
      unfold idom_seq in ESEM; ins; desf.
      clear_gpos; rewrite_gpos.
      forward apply (set_equiv_inr wi gEQ) as W; try done.
      all: unfold event_of_call in *.
      unfolder in W; desf; destruct wi; 
        try destruct a; try destruct a0; try destruct a1; try destruct a2; ins; desf.
      all: rewrite <- eq_rect_eq in *.
      all: by red in VALw; desf; ins; rewrite <- eq_rect_eq in *.
    * (* inductive case *)
      desf.
        2: by destruct n; split; try done; red; eauto.
      rewrite FSai' in TOw.
      assert (W: (set_collect fs g ∩₁ QueueLib.has_loc l) (fs wi)).
        eby split; vauto; eapply SAMEloc'.
      specialize_full IH; try exact TOw; eauto; desf.
        by right; eauto.
      erewrite starting_segment_decomp; eauto using rt_step.
      rewrite filterP_app, app_length.
      rewrite Nat2Z.inj_add, (RC11.functional_has_wval IH VALw ).
      match goal with [ |- context[length ?l]] => assert (EQ1: l = fs ai :: nil) end.
      { rewrite FSai' in *;  clear - IMM DEQa TOw.
        assert (T := TOT l); do 2 (red in T; desc).
        eapply StronglySorted_eq; vauto.
          by apply StronglySorted_filterP, segment_property.
        split; ins; desf.
        * apply in_filterP_iff in H; desc.
          apply segment_property in H; desc.
          apply clos_refl_transE in H2; desf; eauto.
          hahn_rewrite ct_of_ct in H2. 
          apply QueueLib.is_dequeueV_weaken, IMM in H0; clarify_not.
        * apply in_filterP_iff; split; eauto.
          apply segment_property; splits; vauto.
          destruct (fs ai); red in DEQa; repeat (red in s; ins; desf).
          repeat (red in s1; ins; desf); ins; desf; eexists _, eq_refl; ins; desf.
          rewrite <- eq_rect_eq in *; eauto.
      }
      by rewrite EQ1. 
  }
  
  clear FSTIDXw LSTIDXw.
  
  assert (WEval': forall l ai (Gai: g ai) (LOCa: QueueLib.is_dequeueN l (fs ai)) 
                       (LOADai: ev_op ai = C11_load) (LOC: RC11.has_loc ((fst l), 1) ai), 
                   RC11.has_rval (Z.of_nat (which_enq l (fs ai))) ai).
  { ins; clear TOmin.
    specialize (LCONS _ (or_introl eq_refl)); ins; red in LCONS; ins; desc.
    forward apply (set_equiv_inl ai CODOM) as W.
      by unfolder; splits; ins; red; ins; vauto.
    red in W; desc.
    specialize (RFVAL _ _ W); specialize (RFLOC _ _ W).
    unfolder in RFVAL; unfolder in RFLOC; desc.
    specialize (RC11.has_loc_unique RFLOC1 LOC); ins; desc.
    subst a.
    rename a0 into v.
    unfolder in W; desc.
    assert (Gx: g x) by (apply RF in W; unfolder in W; desf).
    assert (X: (QueueLib.is_newqueue_loc l ∪₁ QueueLib.has_loc l) (fs x)).
    { clear WD WE.
      forward eapply queue_from_access with (ai:=x) as Q; eauto; ins; desc.
      assert (lck = snd l).
      { apply NNPP; intro NEQ.
        forward eapply C_LOC with (a := x) (b := ai) (libi := RC11.library); try done.
          exact Q. 
          { instantiate (1 := l); right; clear - LOCa.
            red in LOCa; destruct (fs ai); ins; desf.
            by repeat constructor; exists eq_refl. }
          destruct l; ins; congruence.
          by eauto.
        by clear; intro D; apply (D (fst l, 0)); constructor; exists eq_refl.
      }
      by subst lck; rewrite <- surjective_pairing in *.
    }
    forward eapply WEval as VAL; try exact RFLOC0; ins.
      by clear - RFVAL0; destruct x; red in RFVAL0; desf; ins; desf; vauto; constructor; exists eq_refl.
    specialize (RC11.functional_has_wval RFVAL0 VAL); ins; subst v.
    enough (which_enq l (fs x) = which_enq l (fs ai)) by congruence.
    assert (PE: forall e (ENQe: QueueLib.is_enqueue e) (PREV: toimm⁺ e (fs ai)), 
                 exists ei, <<Gei: g ei>> /\ <<FSei: fs ei = e>> /\ 
                   <<STOREei: RC11.is_store ei>> /\ <<LOCei: RC11.has_loc (fst l,1) ei>>).
    { ins; clear WE WD.
      forward eapply queue_from_access with (ai:=ai) as Q; eauto; ins; desf.
        by red in Q; apply TObase in PREV; desf; destruct (fs ai); unfolder in bLIB; ins; desf.
      assert (EE: exists l, QueueLib.is_enqueue_loc l e).
         by red in ENQe; destruct e; ins; desf; vauto.
      desf; specialize_full SAMEloc; vauto.
      rewrite (QueueLib.has_loc_unique SAMEloc Q) in *.
      assert (Ei: exists ei, g ei /\ e = fs ei) by (apply TObase in PREV; desf; eauto).
      desf; specialize (ESEM _ Ei).
      destruct (fs ei) eqn:FSei; ins; red in EE; ins; desf.
      unfold env_compose, env_of_impl in ESEM; ins; desf.
      specialize_full ESEM; eauto.
      rewrite <- eq_rect_eq in *; ins.
      unfold idom_seq in ESEM; desf; ins; unfold idom_seq in *; desf.
      clear_gpos; rewrite_gpos.
      exists a1; splits.
      1-2: by apply gEQ; unfolder; eauto.
      all: unfold event_of_call in *.
      all: destruct ev_args, a, a0, a1, a2, a3; ins; desf; rewrite <- !eq_rect_eq in *.
      by repeat constructor; exists eq_refl.
    }
    clear WEval WDval.
    destruct X as [X|X].
    * desf. 
      { exfalso; clear - s X. apply proj2 in s.
        red in X; repeat (red in s; desf); destruct (fs x); ins; desf. }
      { destruct n; clear - Gai LOCa. split; vauto.
        red in LOCa; destruct (fs ai); ins; desf.
        by repeat constructor; exists eq_refl. }
      symmetry; apply length_zero_iff_nil.
      apply filterP_eq_nil.
      intros e IN ENQ.
      apply starting_segment_property in IN; desc.
      rewrite clos_refl_transE in IN0; desf.
        by clear - ENQ LOCa; destruct (fs ai); red in LOCa, ENQ; ins; desf.
      hahn_rewrite ct_of_ct in IN0.
      specialize_full PE; try exact ENQ; desf.
      forward eapply queue_from_access with (ai := ei) as LOCe; eauto; desf.     
        by red in LOCe; apply TObase in IN0; desf; destruct (fs ei); unfolder in aLIB; ins; desf.
      forward eapply (ENCAPS (fs ei)) as ALL; try by vauto.
      desf; ins.
      assert (a = fs x).
      { unfolder in Ga; desf.
        assert (l = (fst l, lck)).
        { apply NNPP; intro NEQ.
          forward eapply C_LOC with (a := x) (b := y) (libi := RC11.library); eauto.
             1-2: by left. 
          clear; intro D; apply (D (fst l,0)).
          all: constructor; exists eq_refl; ins.
        }
        rewrite H in *; ins.
        eapply C_ALLOC; ins; eauto.
      }
      subst a.
      specialize_full TOporf; try exact IN0; try done.
      apply TOposo in IN0; try done.
      eapply (ACYC ei), t_trans; apply t_step.
      + repeat left; red; repeat split.
        by eapply inclusion_t_t; try exact IN0; eauto with hahn.
        all: red; ins; eauto.
      + repeat right; unfolder.
        split; [ | eby intro; desf; eapply IRR' ].
        eexists; splits; eauto.
        forward eapply MOT with (a := x) (b := ei).
          by unfolder; splits; eauto.
          by unfolder; splits; vauto; red; ins; eauto.
          by intro; desf. 
        intro M; desf.
        exfalso; eapply ACYC, t_trans; apply t_step.
          eby left; left; right.
        repeat left; red; splits.
        by apply t_step; repeat right; vauto.
        all: by red; ins; eauto.
    * desf. 
      { specialize_full RC11rfTO; try exact W; eauto.
        { clear - X.  repeat (red in X; desf); vauto. }
        { clear - LOCa; right. red in LOCa; desf; vauto. }
        erewrite starting_segment_decomp with (y := fs ai); eauto using rt_step.
        rewrite plus_n_O at 1.
        rewrite filterP_app, app_length.
        f_equal. 
        symmetry; apply length_zero_iff_nil, filterP_eq_nil.
        intros e IN ENQ.
        apply segment_property in IN; desc.
        rewrite clos_refl_transE in IN1; desf.
          by clear - ENQ LOCa; destruct (fs ai); red in LOCa, ENQ; ins; desf.
        hahn_rewrite ct_of_ct in IN1.
        specialize_full PE; try exact ENQ; eauto; desf.
        specialize_full TOporf; try exact IN1; eauto.
        apply TOposo in IN0; try done.
        apply TOposo in IN1; try done.
        eapply (ACYC ei), t_trans; apply t_step.
        + repeat left; red; repeat split.
          by eapply inclusion_t_t; try exact IN1; eauto with hahn.
          all: red; ins; eauto.
        + repeat right; unfolder.
          split; [ | eby intro; desf; eapply IRR' ].
          eexists; splits; eauto.
          forward eapply MOT with (a := x) (b := ei).
            by unfolder; splits; eauto.
            by unfolder; splits; vauto; red; ins; eauto. 
          - intro; desf. eapply ACYC, t_step.
            repeat left; red; splits.
            by eapply inclusion_t_t; try eassumption; eauto with hahn.
            all: by red; ins; eauto.
          - intro M; desf.
            exfalso; eapply ACYC, t_trans; apply t_step.
              eby left; left; right.
            repeat left; red; splits.
            by eapply inclusion_t_t; try eassumption; eauto with hahn.
            all: by red; ins; eauto.
      }
      { destruct n; clear - Gai LOCa. split; vauto.
        red in LOCa; destruct (fs ai); ins; desf.
        by repeat constructor; exists eq_refl. }
      { destruct n; split; vauto. }
  }
  
  assert (WDval': forall l ai (Gai: g ai) (LOCa: QueueLib.is_dequeueN l (fs ai)) 
                       (LOADai: ev_op ai = C11_load) (LOC: RC11.has_loc ((fst l), 2) ai), 
                   RC11.has_rval (Z.of_nat (which_deq l (fs ai))) ai).
  { ins; clear TOmin.
    specialize (LCONS _ (or_introl eq_refl)); ins; red in LCONS; ins; desc.
    forward apply (set_equiv_inl ai CODOM) as W.
      by unfolder; splits; ins; red; ins; vauto.
    red in W; desc.
    specialize (RFVAL _ _ W); specialize (RFLOC _ _ W).
    unfolder in RFVAL; unfolder in RFLOC; desc.
    specialize (RC11.has_loc_unique RFLOC1 LOC); ins; desc.
    subst a.
    rename a0 into v.
    unfolder in W; desc.
    assert (Gx: g x) by (apply RF in W; unfolder in W; desf).
    assert (X: (QueueLib.is_newqueue_loc l ∪₁ QueueLib.has_loc l) (fs x)).
    { clear WD WE.
      forward eapply queue_from_access with (ai:=x) as Q; eauto; ins; desc.
      assert (lck = snd l).
      { apply NNPP; intro NEQ.
        forward eapply C_LOC with (a := x) (b := ai) (libi := RC11.library); try done.
          exact Q.
          { instantiate (1 := l); right; clear - LOCa.
            red in LOCa; destruct (fs ai); ins; desf.
            by repeat constructor; exists eq_refl. }
          destruct l; ins; congruence.
          by eauto.
        by clear; intro D; apply (D (fst l, 0)); constructor; exists eq_refl.
      }
      by subst lck; rewrite <- surjective_pairing in *.
    }
    forward eapply WDval as VAL; try exact RFLOC0; ins.
      by clear - RFVAL0; destruct x; red in RFVAL0; desf; ins; desf; vauto; constructor; exists eq_refl.
    specialize (RC11.functional_has_wval RFVAL0 VAL); ins; subst v.
    enough (which_deq l (fs x) = which_deq l (fs ai)) by congruence.
    assert (PD: forall e v (DEQe: QueueLib.is_dequeueV' v e) (PREV: toimm⁺ e (fs ai)), 
                 exists ei, <<Gei: g ei>> /\ <<FSei: fs ei = e>> /\ 
                   <<STOREei: RC11.is_store ei>> /\ <<LOCei: RC11.has_loc (fst l,2) ei>>).
    { ins; clear WE WD.
      forward eapply queue_from_access with (ai:=ai) as Q; eauto; ins; desf.
        by red in Q; apply TObase in PREV; desf; destruct (fs ai); unfolder in bLIB; ins; desf.
      assert (EE: exists l, QueueLib.is_dequeueV l v e).
        by red in DEQe; destruct e; ins; desf; eexists; exists eq_refl; ins; vauto.
      simpls; desf.
      desf; specialize_full SAMEloc; try eassumption. 
        by clear Q; apply QueueLib.is_dequeue_impl in EE; red; ins; unfolder; eauto.
      rewrite (QueueLib.has_loc_unique SAMEloc Q) in *.
      assert (Ei: exists ei, g ei /\ e = fs ei) by (apply TObase in PREV; desf; eauto).
      desf; specialize (ESEM _ Ei).
      destruct (fs ei) eqn:FSei; ins; red in EE; ins; desf.
      unfold env_compose, env_of_impl in ESEM; ins; desf.
      specialize_full ESEM; eauto.
      rewrite <- eq_rect_eq in *; ins.
      unfold idom_seq in ESEM; desf; ins; unfold idom_seq in *; desf.
      clear_gpos; rewrite_gpos.
      exists a4; splits; try done.
      1-2: by apply gEQ; unfolder; eauto.
      all: unfold event_of_call in *.
      all: destruct ev_args, a, a0, a1, a2, a3, a4; ins; desf; rewrite <- !eq_rect_eq in *.
      by repeat constructor; exists eq_refl.
    }
    clear WEval WDval.
    destruct X as [X|X].
    * desf. 
      { exfalso; clear - s X. apply proj2 in s.
        red in X; repeat (red in s; desf); destruct (fs x); ins; desf. }
      { destruct n; clear - Gai LOCa. split; vauto.
        red in LOCa; destruct (fs ai); ins; desf.
        by repeat constructor; exists eq_refl. }
      symmetry; apply length_zero_iff_nil.
      apply filterP_eq_nil.
      intros e IN DEQ.
      apply starting_segment_property in IN; desc.
      rewrite clos_refl_transE in IN0; desf. 
        by clear - DEQ LOCa; destruct (fs ai); red in LOCa, DEQ; ins; desf; rewrite <- !eq_rect_eq in *; desf.
      hahn_rewrite ct_of_ct in IN0.
      specialize_full PD; try exact IN0; desf.
        eby eapply QueueLib.is_dequeueV_weaken.
      forward eapply queue_from_access with (ai:=ei) as Q; eauto; ins; desf.
        by red in Q; apply TObase in IN0; desf; destruct (fs ei); unfolder in aLIB; ins; desf.
      forward eapply (ENCAPS (fs ei)) as ALL; try by vauto.
      { clear - IN; apply proj2 in IN. repeat (red in IN; desf); red; ins; eauto. }
      desf; ins.
      assert (a = fs x).
      { unfolder in Ga; desf.
        assert (l = (fst l, lck)).
        { apply NNPP; intro NEQ.
          forward eapply C_LOC with (a := x) (b := y) (libi := RC11.library); eauto.
             1-2: by left. 
          clear; intro D; apply (D (fst l,0)).
          all: constructor; exists eq_refl; ins.
        }
        rewrite H in *; ins.
        eapply C_ALLOC; ins; eauto.
      }
      subst a.
      specialize_full TOporf; try exact IN0; try done.
      apply TOposo in IN0; try done.
      eapply (ACYC ei), t_trans; apply t_step.
      + repeat left; red; repeat split.
        by eapply inclusion_t_t; try exact IN0; eauto with hahn.
        all: red; ins; eauto.
      + repeat right; unfolder.
        split; [ | eby intro; desf; eapply IRR' ].
        eexists; splits; eauto.
        forward eapply MOT with (a := x) (b := ei).
          by unfolder; splits; eauto.
          by unfolder; splits; vauto; red; ins; eauto.
          by intro; desf. 
        intro M; desf.
        exfalso; eapply ACYC, t_trans; apply t_step.
          eby left; left; right.
        repeat left; red; splits.
        by apply t_step; repeat right; vauto.
        all: by red; ins; eauto.
    * desf. 
      { specialize_full RC11rfTO; try exact W; eauto.
        { clear - X.  repeat (red in X; desf); vauto. }
        { clear - LOCa; right. red in LOCa; desf; vauto. }
        erewrite starting_segment_decomp with (y := fs ai); eauto using rt_step.
        rewrite plus_n_O at 1.
        rewrite filterP_app, app_length.
        f_equal. 
        symmetry; apply length_zero_iff_nil, filterP_eq_nil.
        intros e IN DEQ.
        apply segment_property in IN; desc.
        rewrite clos_refl_transE in IN1; desf.
          by clear - DEQ LOCa; destruct (fs ai); red in LOCa, DEQ; ins; desf; rewrite <- !eq_rect_eq in *; desf.
        hahn_rewrite ct_of_ct in IN1.
        specialize_full PD; try exact IN1; eauto; desf.
          eby eapply QueueLib.is_dequeueV_weaken.
        specialize_full TOporf; try exact IN1; eauto.
        apply TOposo in IN0; try done.
        apply TOposo in IN1; try done.
        eapply (ACYC ei), t_trans; apply t_step.
        + repeat left; red; repeat split.
          by eapply inclusion_t_t; try exact IN1; eauto with hahn.
          all: red; ins; eauto.
        + repeat right; unfolder.
          split; [ | eby intro; desf; eapply IRR' ].
          eexists; splits; eauto.
          forward eapply MOT with (a := x) (b := ei).
            by unfolder; splits; eauto.
            by unfolder; splits; vauto; red; ins; eauto. 
          - intro; desf. eapply ACYC, t_step.
            repeat left; red; splits.
            by eapply inclusion_t_t; try eassumption; eauto with hahn.
            all: by red; ins; eauto.
          - intro M; desf.
            exfalso; eapply ACYC, t_trans; apply t_step.
              eby left; left; right.
            repeat left; red; splits.
            by eapply inclusion_t_t; try eassumption; eauto with hahn.
            all: by red; ins; eauto.
      }
      { destruct n; clear - Gai LOCa. split; vauto.
        red in LOCa; destruct (fs ai); ins; desf.
        by repeat constructor; exists eq_refl. }
      { destruct n; split; vauto. } 
  }
  
  assert (DEQN: forall l d (Gd: set_collect fs g d) (DEQ: QueueLib.is_dequeueN l d), 
                which_enq l d = which_deq l d).
  { ins.
    assert (exists r1 r2 v, <<Gr1: g r1>> /\ <<Gr2: g r2>> /\ <<FSr1: fs r1 = d>> /\ <<FSr2: fs r2 = d >> /\
                            <<OPr1: ev_op r1 = C11_load>> /\ <<OPr2: ev_op r2 = C11_load>> /\ 
                            <<LOCr1: RC11.has_loc (fst l, 1) r1>> /\ <<LOCr2: RC11.has_loc (fst l, 2) r2>> /\
                            <<VALr1: RC11.has_rval v r1>> /\ <<VALr2: RC11.has_rval v r2>>).
    { clear - Gd DEQ ESEM.
      red in Gd, DEQ; ins; desf.
      specialize (ESEM _ Gd).
      destruct (fs y); ins; desf; rewrite <- eq_rect_eq.
      unfold env_compose, env_of_impl in ESEM; ins; desf.
      specialize_full ESEM; ins.
      rewrite <- eq_rect_eq in *; ins.
      unfold idom_seq in ESEM; desf; ins; unfold idom_seq in *; desf.
      clear_gpos; rewrite_gpos.
      unfold event_of_call in *; desf; ins.
      exists a0, a.
      destruct a, a0, a1, a2; ins; desf; rewrite <- !eq_rect_eq in *.
      eexists; splits; ins.
      1-4: by apply gEQ; unfolder; eauto.
      all: unfold event_of_call in *.
      all: repeat constructor; exists eq_refl; ins.
      by apply Z.eqb_eq.
    }
    clear - H DEQ WEval' WDval'; desf.
    specialize_full WEval'; try exact LOCr1; eauto.
    specialize_full WDval'; try exact LOCr2; eauto; try congruence.
    rewrite FSr2 in *.
    red in VALr1, VALr2, WEval', WDval'; destruct r1, r2; ins; desf; rewrite <- !eq_rect_eq in *; ins; desf.
    clear - VALr2 WEval'.
    rewrite WEval' in VALr2 at 2; clear WEval'.
    by apply Nat2Z.inj.
  }
  
  clear WEval' WDval'.
  
  remember(fun a b => exists l v, << ENQa: (set_collect fs g ∩₁ QueueLib.is_enqueue_loc l) a >> /\
                                  << DEQb: (set_collect fs g ∩₁ QueueLib.is_dequeueV l v) b  >> /\
                                  << W: which_enq l a = which_deq l b >>) as rf' eqn:RF'.
  
  assert (RF'ct: rf'⁺ ≡ rf').
  { clear WE WD; desf; clear; apply ct_no_step.
    unfolder; split; ins; desf.
    clear - DEQb3 ENQa0.
    red in DEQb3, ENQa0; destruct (fs y2); ins; desf.
  }  
  
  assert (ISREAD: forall e v (Ge: g e) (READ: RC11.has_rval v e), ev_op e = C11_load).
  { clear - GDOM ESEM; ins.
    specialize (GDOM _ Ge).
    specialize (ESEM _ Ge).
    red in GDOM; ins.
    destruct (fs e) eqn:FSe; ins; desf.
    all: unfold env_compose, env_of_impl in ESEM; ins; desf.
    all: specialize_full ESEM; eauto; rewrite <- eq_rect_eq in *; ins.
    all: unfold idom_seq in ESEM; desf; ins; unfold idom_seq in *; desf.
    all: clear_gpos; rewrite_gpos.
    all: forward eapply (set_equiv_inr e gEQ) as E; try done.
    all: unfold event_of_call in *.
    all: unfolder in E; desf.
    all: red in READ; destruct e.
    all: try destruct a; try destruct a0; try destruct a1; 
         try destruct a2; try destruct a3; try destruct a4; ins; desf; rewrite <- !eq_rect_eq in *.
  }
  
  assert (OFFPOS: forall r (Gr: g r) q i (I: i = 1 \/ i = 2) (LOC: RC11.has_loc (q,i) r)
                         v (VAL: RC11.has_rval v r), (v >= 0)%Z).
  { clear - GDOM ESEM LCONS WEval WDval RF ISREAD; ins.
    specialize_full ISREAD; eauto.
    specialize (LCONS _ (or_introl eq_refl)); red in LCONS; ins; red in LCONS; ins; desc.
    forward apply (set_equiv_inl r CODOM) as RFwr.
      by unfolder; splits; eauto; clear I; red in VAL; desf; red; ins; eauto.
    red in RFwr; desc.
    specialize (RFVAL _ _ RFwr); specialize (RFLOC _ _ RFwr).
    unfolder in RFVAL; unfolder in RFLOC; desc.
    specialize (RC11.has_loc_unique RFLOC1 LOC); ins; subst.
    assert (a0 = v).
    { clear - VAL RFVAL1 ISREAD.
      destruct r; repeat (red in RFVAL1; desf); repeat (red in VAL; desf); ins; desf.
      rewrite <- !eq_rect_eq in *; desf.
    }
    subst a0.
    assert (Gx: g x) by (unfolder in RFwr; desc; apply RF in RFwr; unfolder in RFwr; desf).
    forward eapply queue_from_access with (ai:=x) as Q; eauto; ins; desc.
    destruct I; subst.
    1: eapply WEval in Q; eauto.
    2: eapply WDval in Q; eauto.
    all: specialize (RC11.functional_has_wval Q RFVAL0); ins; desf; omega.
  }
  
  assert(DEQUEUE: forall d v (DEQ: QueueLib.is_dequeueV' v d) (GSe: set_collect fs g d), 
                  exists dir dir' dir'' diw q offset, 
                  <<Gdir: g dir>> /\ <<Gdir': g dir'>> /\ <<Gdir': g dir''>> /\ <<Gdiw: g diw>> /\
                  <<FSdir: fs dir = d>> /\ <<FSdir': fs dir' = d >> /\ 
                  <<FSdir'': fs dir'' = d >> /\<<FSdiw: fs diw = d >> /\
                  <<VALdir: RC11.has_rval v dir>> /\ <<LOCdir: RC11.has_loc (q,offset+3) dir>> /\
                  <<VALdir': RC11.has_rval (Z.of_nat offset) dir'>> /\ <<LOCdir': RC11.has_loc (q,2) dir'>> /\
                  <<VALdiw: RC11.has_wval (Z.of_nat (offset + 1)) diw>> /\
                  <<LOCdiw: RC11.has_loc (q,2) diw>> /\ 
                  exists v'', <<VALdir'': RC11.has_rval (Z.of_nat v'') dir''>> /\
                              <<LOCdir'': RC11.has_loc (q,1) dir''>> /\ <<VAL'': v'' <> offset>>).
  { ins; clear WD WE RF'.
    red in GSe, DEQ; desf.
    specialize (ESEM _ GSe).
    destruct (fs y) eqn:FSy; ins; desf; rewrite <- eq_rect_eq in *. 
    unfold env_compose, env_of_impl in ESEM; ins; desf.
    specialize_full ESEM; eauto; rewrite <- eq_rect_eq in *; ins.
    all: unfold idom_seq in ESEM; desf; ins; unfold idom_seq in *; desf.
    all: clear_gpos; rewrite_gpos.
    all: unfold event_of_call in *; desf; ins.
    eexists a3, a, a0, a4, (fst (fst ev_args)), _; splits; ins. 
    1-8: by eapply gEQ; unfolder; eauto 10.
    1-2: by destruct a3; ins; desf; constructor; exists eq_refl.
    all: destruct a, a0, a1, a2, a3, a4; ins; desf; rewrite <- !eq_rect_eq; ins.
    1-4: repeat constructor; exists eq_refl; ins.
    * forward eapply (OFFPOS {|ev_id := ev_id0; ev_op := C11_load;
                               ev_args := (fst (fst ev_args), 2, (RMna, tt));
                               ev_res := ev_res|}) with (i := 2) as POS; ins; eauto.
      by apply gEQ; unfolder; eauto 10.
      1-2: by constructor; exists eq_refl.
      clear - POS; ins.
      rewrite Z2Nat.id; omega.
    * forward eapply (OFFPOS {|ev_id := ev_id0; ev_op := C11_load;
                               ev_args := (fst (fst ev_args), 2, (RMna, tt));
                               ev_res := ev_res|}) with (i := 2) as POS; ins; eauto.
      by apply gEQ; unfolder; eauto 10.
      1-2: by constructor; exists eq_refl.
      clear - POS; ins.
      replace  1 with (Z.to_nat 1) by done.
      by rewrite <- Z2Nat.inj_add, Z2Nat.id; omega.
    * forward eapply (OFFPOS {|
                 ev_id := ev_id1;
                 ev_op := C11_load;
                 ev_args := (fst (fst ev_args), 1, (RMna, tt));
                 ev_res := ev_res0 |}) with (i := 1) as POS; ins; eauto.
      by apply gEQ; unfolder; eauto 10.
      1-2: by constructor; exists eq_refl.
      simpl in POS.
      assert (EQ := POS); apply Z.ge_le, Z2Nat.id in EQ.
      exists (Z.to_nat ev_res0); splits.
      1-2: repeat constructor; exists eq_refl; ins.
      forward eapply (OFFPOS {|ev_id := ev_id0; ev_op := C11_load;
                               ev_args := (fst (fst ev_args), 2, (RMna, tt));
                               ev_res := ev_res|}) with (i := 2) as POS'; ins; eauto.
      by apply gEQ; unfolder; eauto 10.
      1-2: by constructor; exists eq_refl.
      clear - Heq POS POS'; ins.
      apply Z.eqb_neq in Heq.
      intro EQ; apply Z2Nat.inj in EQ; omega.
  }
  
  assert(ENQUEUE: forall e (ENQ: QueueLib.is_enqueue e) (GSe: set_collect fs g e), 
                  exists eiw eir eiw' q offset v, 
                  <<Geiw: g eiw>> /\ <<Geir: g eir>> /\ <<Geiw': g eiw'>> /\
                  <<FSeiw: fs eiw = e >> /\ <<FSeir: fs eir = e>> /\ <<FSeiw': fs eiw' = e>> /\
                  <<VALeiw: RC11.has_wval v eiw>> /\ <<LOCeiw: RC11.has_loc (q,offset+3) eiw>> /\
                  <<VALeir: RC11.has_rval (Z.of_nat offset) eir>> /\ <<LOCeir: RC11.has_loc (q,1) eir>> /\
                  <<VALeiw': RC11.has_wval (Z.of_nat (offset + 1)) eiw'>> /\ 
                  <<LOCeiw': RC11.has_loc (q,1) eiw'>> /\ <<ENQUEUEval: QueueLib.is_enqueueV' v e>>).
  { ins; clear WD WE RF'.
    red in GSe, ENQ; desf.
    specialize (ESEM _ GSe).
    destruct (fs y) eqn:FSy; ins; desf.
    unfold env_compose, env_of_impl in ESEM; ins; desf.
    specialize_full ESEM; eauto; rewrite <- eq_rect_eq in *; ins.
    all: unfold idom_seq in ESEM; desf; ins; unfold idom_seq in *; desf.
    all: clear_gpos; rewrite_gpos.
    all: unfold event_of_call in *; desf; ins.
    eexists a0, a2, a1, (fst (fst ev_args)). 
    destruct a, a0, a1, a2, a3; ins; desf; rewrite <- !eq_rect_eq in *.
    eexists _, _; splits.
    1-6: by eapply gEQ; unfolder; eauto 10.
    all: repeat constructor; exists eq_refl; ins.
    * forward eapply (OFFPOS {|ev_id := ev_id3; ev_op := C11_load;
                             ev_args := (fst (fst ev_args), 1, (RMna, tt));
                             ev_res := ev_res2 |}) with (i := 1) as POS; ins; eauto.
      by apply gEQ; unfolder; eauto 10.
      1-2: by constructor; exists eq_refl.
      clear - POS; ins.
      rewrite Z2Nat.id; omega.
    * forward eapply (OFFPOS {|ev_id := ev_id3; ev_op := C11_load;
                             ev_args := (fst (fst ev_args), 1, (RMna, tt));
                             ev_res := ev_res2 |}) with (i := 1) as POS; ins; eauto.
      by apply gEQ; unfolder; eauto 10.
      1-2: by constructor; exists eq_refl.
      clear - POS; ins.
      replace 1 with (Z.to_nat 1) by done.
      by rewrite <- Z2Nat.inj_add, Z2Nat.id; omega.
  }
  
  assert (WEinj: forall l a b (A: (set_collect fs g ∩₁ QueueLib.is_enqueue_loc l) a)
                   (B: (set_collect fs g ∩₁ QueueLib.is_enqueue_loc l) b) 
                   (EQ: which_enq l a = which_enq l b), a = b).
  { desf; clear; ins; desf.
    * apply NNPP; intro NEQ; eapply TOT in NEQ; eauto; desf.
      + erewrite starting_segment_decomp with (y := b) in EQ; eauto using rt_step.
        rewrite filterP_app, app_length in EQ.
        rewrite plus_n_O in EQ at 1.   
        apply Nat.add_cancel_l in EQ.
        symmetry in EQ; apply length_zero_iff_nil in EQ.
        enough (In b nil) by done.
        rewrite <- EQ; clear EQ.
        apply in_filterP_iff; splits.
        by apply segment_property; splits; vauto.
        by destruct b; unfold QueueLib.is_enqueue_loc in B; unfolder in B; red; ins; desf; eauto.
      + erewrite starting_segment_decomp with (y := a) in EQ; eauto using rt_step.
        rewrite filterP_app, app_length in EQ.
        rewrite plus_n_O in EQ.   
        apply Nat.add_cancel_l in EQ.
        apply length_zero_iff_nil in EQ.
        enough (In a nil) by done.
        rewrite <- EQ; clear EQ.
        apply in_filterP_iff; splits.
        by apply segment_property; splits; vauto.
        by destruct a; unfold QueueLib.is_enqueue_loc in A; unfolder in A; red; ins; desf; eauto.
    * destruct n; clear - B.
      split; [by red in B; desf | ].
      destruct b; unfold QueueLib.is_enqueue_loc in B; unfolder in B; ins; desf.
      by constructor; exists eq_refl.
    * destruct n; clear - A.
      split; [by red in A; desf | ].
      destruct a; unfold QueueLib.is_enqueue_loc in A; unfolder in A; ins; desf.
      by constructor; exists eq_refl.
    * destruct n; clear - A.
      split; [by red in A; desf | ].
      destruct a; unfold QueueLib.is_enqueue_loc in A; unfolder in A; ins; desf.
      by constructor; exists eq_refl.
  }

     
  assert (WDinj: forall l a b v v' (A: (set_collect fs g ∩₁ QueueLib.is_dequeueV l v) a)
                   (B: (set_collect fs g ∩₁ QueueLib.is_dequeueV l v') b) 
                   (EQ: which_deq l a = which_deq l b), a = b).
  { desf; clear; ins; desf.
    * apply NNPP; intro NEQ; eapply TOT in NEQ; eauto; desf.
      + erewrite starting_segment_decomp with (y := b) in EQ; eauto using rt_step.
        rewrite filterP_app, app_length in EQ.
        rewrite plus_n_O in EQ at 1.   
        apply Nat.add_cancel_l in EQ.
        symmetry in EQ; apply length_zero_iff_nil in EQ.
        enough (In b nil) by done.
        rewrite <- EQ; clear EQ.
        apply in_filterP_iff; splits.
        by apply segment_property; splits; vauto.
        eexists; exact (proj2 B).
      + erewrite starting_segment_decomp with (y := a) in EQ; eauto using rt_step.
        rewrite filterP_app, app_length in EQ.
        rewrite plus_n_O in EQ.   
        apply Nat.add_cancel_l in EQ.
        apply length_zero_iff_nil in EQ.
        enough (In a nil) by done.
        rewrite <- EQ; clear EQ.
        apply in_filterP_iff; splits.
        by apply segment_property; splits; vauto.
        eexists; exact (proj2 A).
    * destruct n; clear - B.
      split; [by red in B; desf | ].
      destruct b; unfold QueueLib.is_dequeueV in B; unfolder in B; ins; desf.
      by constructor; exists eq_refl.
    * destruct n; clear - A.
      split; [by red in A; desf | ].
      destruct a; unfold QueueLib.is_dequeueV in A; unfolder in A; ins; desf.
      by constructor; exists eq_refl.
    * destruct n; clear - A.
      split; [by red in A; desf | ].
      destruct a; unfold QueueLib.is_dequeueV in A; unfolder in A; ins; desf.
      by constructor; exists eq_refl.
  }   

 
  assert (WEval'':forall ai q i (Gai: g ai) (STOREai: RC11.is_write ai) (LOCai: RC11.has_loc (q,i+3) ai), 
          exists lck, <<ENQ: QueueLib.is_enqueue_loc (q,lck) (fs ai)>> /\ 
                      << WE: which_enq (q,lck) (fs ai) = i + 1 >>).
  { ins.
    assert (exists ai' lck, <<ENQ: QueueLib.is_enqueue_loc (q,lck) (fs ai)>> /\ <<Gai': g ai'>> /\
                                    <<FSai': fs ai' = fs ai >> /\
                                    <<LOCai': RC11.has_loc (q, 1) ai'>> /\
                                    <<VALai': RC11.has_wval (Z.of_nat (i + 1)) ai'>>).
    { specialize (ESEM _ Gai).
      specialize (GDOM _ Gai).
      destruct (fs ai) eqn:FSai; red in GDOM; ins; desf.
      all: unfold env_compose, env_of_impl in ESEM; ins; desf.
      all: specialize_full ESEM; eauto; rewrite <- eq_rect_eq in *; ins.
      all: unfold idom_seq in ESEM; desf; ins; unfold idom_seq in *; desf.
      all: clear_gpos; rewrite_gpos.
      all: forward apply (set_equiv_inr ai gEQ) as A; try done.
      all: unfold event_of_call in *; desf; ins.
      all: unfolder in A; repeat (red in STOREai; desf; ins); destruct ai; ins; desf.
      all: repeat (red in LOCai; desf; ins);  rewrite <- !eq_rect_eq in *; ins.
      all: try by inv LOCai; omega.
      exists a1; destruct a, a1, a2, a3; ins; desf; rewrite <- !eq_rect_eq in *; ins.
      eexists; splits; ins. 
      2-3: by apply gEQ; unfolder; eauto 10.
      all: repeat constructor; exists eq_refl; ins.
      * instantiate (1 := snd (fst ev_args)). 
        destruct ev_args; ins.
        by destruct q.
      * forward eapply (OFFPOS {|
                 ev_id := ev_id3;
                 ev_op := C11_load;
                 ev_args := (fst (fst ev_args), 1, (RMna, tt));
                 ev_res := ev_res2 |}) with (i := 1) as POS; ins; eauto.
        by apply gEQ; unfolder; eauto 10.
        1-2: by constructor; exists eq_refl.
        clear - POS H0; ins.
        apply Nat.add_cancel_r in H0; desf.
        replace  1 with (Z.to_nat 1) by done.
        by rewrite <- Z2Nat.inj_add, Z2Nat.id; omega.
    }
    desf; eexists; splits; eauto.
    rewrite <- FSai' in *.
    forward eapply (WEval (q,lck) _ Gai'); ins.
      by unfolder; vauto.
      by destruct ai'; red; red in VALai'; ins; desf; ins; repeat constructor; exists eq_refl.
    specialize (RC11.functional_has_wval H VALai'); clear; ins.
    by apply Nat2Z.inj.
  }
    

  assert (RF'property: forall e d, rf' e d -> <<TOrf': toimm⁺ e d >> /\ 
            << SAMEVAL: exists v l, QueueLib.is_enqueueV l v e /\ QueueLib.is_dequeueV l v d>>).
  { clear WE WD TObase TOmin OFFPOS.
    ins; desf.
    forward eapply (DEQUEUE d v); ins; desf.
      by eapply QueueLib.is_dequeueV_weaken, (proj2 DEQb).
      by exact (proj1 DEQb).
    assert (q = fst l).
    { forward eapply queue_from_access with (ai:=dir) as Q; eauto; desf.
      { clear - Q DEQb; apply proj2 in DEQb.
        red in Q, DEQb; destruct (fs dir); ins; desf. }
      clear - Q DEQb; ins.
      apply proj2 in DEQb; destruct (fs dir); red in DEQb; ins; desf.
      red in Q; unfolder in Q; ins; desf.
      all: red in Q; ins; desf.
      rewrite <- eq_rect_eq in *.
      by rewrite Q.
    }
    subst q.
    rewrite <- FSdiw in DEQb, W.
    specialize_full WDval; try exact LOCdiw; ins.
      { clear - DEQb; apply proj2 in DEQb; right.
        red in DEQb; destruct (fs diw); ins; desf; ins.
        by repeat constructor; exists eq_refl. }
      { clear - VALdiw. red in VALdiw; destruct diw; ins; desf; ins; vauto.
        by repeat constructor; exists eq_refl. } 
    assert (WD := Nat2Z.inj _ _ (RC11.functional_has_wval WDval VALdiw)).
    forward eapply (ENQUEUE e); ins; desf.
      { clear - ENQa; apply proj2 in ENQa. red in ENQa; destruct e; ins; desf. }
      { exact (proj1 ENQa). } 
    assert (q = fst l).
    { forward eapply queue_from_access with (ai:=eiw) as Q; eauto; desf.
      { clear - Q ENQUEUEval.
        red in Q, ENQUEUEval; destruct (fs eiw); ins; desf. }
      clear - Q ENQa; ins.
      apply proj2 in ENQa; destruct (fs eiw); red in ENQa; ins; desf.
      red in Q; unfolder in Q; ins; desf.
      all: red in Q; ins; desf.
      rewrite <- eq_rect_eq in *.
      by rewrite Q.
    }
    subst q.
    rewrite <- FSeiw' in ENQa, W.
    specialize_full WEval; try exact LOCeiw'; ins.
      by apply proj2 in ENQa; right; vauto.
      { clear - VALeiw'. red in VALeiw'; destruct eiw'; ins; desf; ins; vauto.
        by repeat constructor; exists eq_refl. }
    assert (WE := Nat2Z.inj _ _ (RC11.functional_has_wval WEval VALeiw')).
    assert (offset0 = offset) by omega.
    subst offset0.
    assert (exists w, <<Gw: g w >> /\ <<RFw: rf w dir>> /\ <<LOCw: RC11.has_loc (fst l, offset + 3) w>> /\
                      <<VALw: RC11.has_wval v w >>).
    { specialize (LCONS _ (or_introl eq_refl)); red in LCONS; ins; red in LCONS; ins; desc.
      forward eapply (set_equiv_inl dir CODOM) as R.
      { clear - VALdir Gdir. unfold lib_events; ins.
        red in VALdir; destruct dir; desf; unfolder; splits; ins; desf; eauto; vauto.
        tertium_non_datur (ev_res = fst (snd ev_args)).
        all: by repeat constructor; exists eq_refl. }
      red in R; desc.
      exists x; splits.
      * unfolder in R; desf. apply RF in R; unfolder in R; desf.
      * unfolder in R; desf.
      * apply RFLOC in R; unfolder in R; desf.
        clear - R0 R1 LOCdir. specialize (RC11.has_loc_unique R1 LOCdir); ins; desf.
      * apply RFVAL in R; unfolder in R; desf.
        specialize (ISREAD _ _ Gdir VALdir); clear - ISREAD VALdir R0 R1.
        red in VALdir, R1; destruct dir; ins; desf.
        by rewrite <- eq_rect_eq in *.
    }
    desf.
    forward eapply WEval''; try exact LOCw; eauto; ins; desf.
    assert (lck = snd l).
    { apply NNPP; intro NEQ.
      forward eapply C_LOC with (a := w) (b := eiw') (libi := RC11.library); try done.
        by right; left; exact ENQ.
        by right; left; exact (proj2 ENQa).
        by destruct l; ins; congruence.
        by eauto.
      by clear ; intro D; apply (D (fst l, 0)); constructor; exists eq_refl.
    }
    subst lck. 
    rewrite <- surjective_pairing in *.
    assert (w = eiw).
    { assert (fs w = fs eiw).
      { specialize_full WEval''; try exact LOCeiw; eauto; desf.
        assert (lck = snd l).
        { apply NNPP; intro NEQ.
          forward eapply C_LOC with (a := w) (b := eiw) (libi := RC11.library); try done.
            by right; left; exact ENQ.
            by right; left; exact ENQ0.
            by destruct l; ins; congruence.
            by eauto.
          by clear ; intro D; apply (D (fst l, 0)); constructor; exists eq_refl.
        }
        subst lck; rewrite <- surjective_pairing in *.
        rewrite FSeiw', <- WE0 in WE.
        eapply WEinj.
        3: symmetry; exact WE.
        by split; vauto.
        by rewrite <- FSeiw'.
      }
      clear - H ESEM Geiw Gw LOCw LOCeiw ENQ.
      specialize (ESEM _ Gw).
      destruct (fs w) eqn:FSw; red in ENQ; ins; desf.
      unfold env_of_impl, env_compose in ESEM; ins; desf.
      specialize_full ESEM; eauto; rewrite <- eq_rect_eq in *; ins.
      unfold idom_seq in ESEM; desf.
      clear_gpos; rewrite_gpos.
      enough (w = a0 /\ eiw = a0) by desf.
      split.
      * forward eapply (set_equiv_inr w gEQ) as W; try done.
        unfold event_of_call in *.
        unfolder in W; red in LOCw; destruct w; ins; desf; ins; desf.
        all: try destruct a; try destruct a0; try destruct a1; try destruct a2; try destruct a3; 
             ins; rewrite <- !eq_rect_eq in *; ins.
        all: inv LOCw; omega.
      * forward eapply (set_equiv_inr eiw gEQ) as EIW; try done.
        all: unfold event_of_call in *.
        unfolder in EIW; red in LOCeiw; destruct eiw; ins; desf; ins; desf.
        all: try destruct a; try destruct a0; try destruct a1; try destruct a2; try destruct a3; 
             ins; rewrite <- !eq_rect_eq in *; ins.
        all: inv LOCeiw; omega.
    }
    subst w.    
    rewrite <- WE0 in WE; eapply WEinj in WE.
    2-3: congruence.
    splits.
    * rewrite <- FSeiw', WE.
      eapply RC11rfTO; eauto.
      { clear - ENQ; left.
        red in ENQ; destruct (fs eiw); red; desf; eauto. }
      { rewrite <- FSdiw; clear - DEQb.
        right; apply proj2 in DEQb; red in DEQb; destruct (fs diw); red; desf; eauto. }
      { clear - VALw; red in VALw; red; ins; desf; eauto. }
      { clear - VALdir; red in VALdir; red; ins; desf; eauto. }
    * specialize (RC11.functional_has_wval VALeiw VALw); ins; desf.
      exists v, l; splits; eauto.
      { clear - ENQ ENQUEUEval.
        red in ENQ, ENQUEUEval; destruct (fs eiw); ins; desf.
        exists eq_refl; ins; rewrite <- !eq_rect_eq in *; splits; ins. }
      by rewrite <- FSdiw; exact (proj2 DEQb).
  }
    
    
                                
   assert (CODOMrf': forall d v (Gd: set_collect fs g d) (DEQ: QueueLib.is_dequeueV' v d), exists e, rf' e d).
   { clear WE WD RF'ct TOmin RC11rfTO; ins; desf.
     assert (DEQl: exists l, QueueLib.has_loc l d).
     { clear - DEQ; red in DEQ; destruct d; ins; desf.
       by eexists; repeat constructor; exists eq_refl.
     }  
     specialize_full DEQUEUE; eauto; desf.
     assert (q = fst l).
     { eapply queue_from_access in LOCdir; eauto; ins; desf.
       by clear - DEQ LOCdir; red in DEQ, LOCdir; destruct (fs dir); ins; desf.
       by clear - DEQl LOCdir; specialize (QueueLib.has_loc_unique DEQl LOCdir); ins; desf.
     }
     subst q.
     specialize_full WDval; try exact VALdiw; eauto.
       by right; congruence.
     assert (QEQ := RC11.functional_has_wval WDval VALdiw).
     apply Nat2Z.inj in QEQ.
     specialize (LCONS _ (or_introl eq_refl)); ins; red in LCONS; ins; desc.
     forward apply (set_equiv_inl dir CODOM) as R.
     { clear - Gdir VALdir. red in VALdir; destruct dir; unfolder; splits; ins; desf.
       1,3: by repeat constructor; exists eq_refl.
       by tertium_non_datur (ev_res = fst (snd ev_args)); repeat constructor; exists eq_refl.
       by red; ins; eauto. }
     red in R; desc.
     specialize (RFVAL _ _ R); specialize (RFLOC _ _ R).
     destruct RFVAL as [val RFVAL]; unfolder in RFVAL; desc.
     destruct RFLOC as [loc RFLOC]; unfolder in RFLOC; desc.
     unfolder in R; desc.
     assert (Gx: g x) by (apply RF in R; unfolder in R; desf).
     specialize (RC11.has_loc_unique RFLOC1 LOCdir); ins; desf.
     specialize_full WEval''; try exact RFLOC0; eauto; desf.
     assert (lck = snd l).
     { apply NNPP; intro NEQ.
       forward eapply C_LOC with (a := dir) (b := x) (libi := RC11.library); try done.
         by right; exact DEQl.
         by right; left; exact ENQ.
         by destruct l; ins; congruence.
         by eauto.
       by clear; intro D; apply (D (fst l, 0)); constructor; exists eq_refl.
     }
     subst lck; rewrite <- surjective_pairing in *.
     eexists _, _, _; splits.
     * split; try exact ENQ; vauto.
     * split; vauto.
       clear - DEQ DEQl.
       destruct (fs dir); red in DEQ; repeat (red in DEQl; desf; ins).
       exists eq_refl; ins; rewrite <- eq_rect_eq; split; eauto.
     * congruence.
   }
       
   assert (RF'1FUNC: functional rf'⁻¹).
   {
    clear WE WD  ESEM LCONS C_LOC TObase TOmin RC11rfTO WEval
          WDval DEQN ISREAD OFFPOS DEQUEUE ENQUEUE WEval''.
    unfolder; ins; desc.
    assert (Y := RF'property _ _ H); desc.
    specialize (RF'property _ _ H0); ins; desc.
    assert (l0 = l).
      by destruct x; red in SAMEVAL0, SAMEVAL2; ins; desf; ins; rewrite <- eq_rect_eq.
    desf.
    assert (l0 = l).
    { clear - DEQb SAMEVAL0; apply proj2 in DEQb.
      by destruct x; red in SAMEVAL0, DEQb; ins; desf; ins; rewrite <- eq_rect_eq.
    }
    assert (l1 = l).
    { clear - DEQb0 SAMEVAL0; apply proj2 in DEQb0.
      by destruct x; red in SAMEVAL0, DEQb0; ins; desf; ins; rewrite <- eq_rect_eq.
    }
    desf.
    eapply WEinj; eauto.
    congruence.
  }
    
         
       
  exists rf', rf'; splits; try done.
  
  1-2: by desf; clear; unfolder; ins; desf; desf; eauto 10.
  
  { (* rf contained in (po ∪ rf)⁺ *)
    rewrite RF'ct.
    clear - TObase TOporf RF'property.
    red; ins.
    apply RF'property in H; clear RF'property; desc.
    specialize (TObase _ _ TOrf'); desf.
    apply TOporf in TOrf'; eauto.
    eexists _, _; splits; eauto.
  }
  
  { (* rf contained in (po ∪ so)⁺ *)
    rewrite RF'ct.
    clear - TObase TOposo RF'property.
    red; ins.
    apply RF'property in H; clear RF'property; desc.
    specialize (TObase _ _ TOrf'); desf.
    apply TOposo in TOrf'; eauto.
    eexists _, _; splits; eauto.
    eapply inclusion_t_t; eauto.
    eauto with hahn.
  } 
      
  { by rewrite RF'ct; apply functional_inv_fsupp. }
  
  red; ins; splits; try done.
  
  { (* rf' connects enqueues and dequeues with same value *)
    clear - RF'property; unfolder; ins.
    apply RF'property in H; desc.
    eby eexists; split; ins; eexists; splits.
  }
  
  { (* each queue is allocated only once *)
    clear - C_ALLOC; red in C_ALLOC; unfold set_collect; ins; desf; eauto.
  }
  
  { (* functionality of rf' *)
    clear WE WD ESEM LCONS C_LOC TObase TOmin RC11rfTO WEval
          WDval DEQN ISREAD OFFPOS DEQUEUE ENQUEUE WEval''.
    unfolder; ins; desc.
    assert (Y := RF'property _ _ H); desc.
    specialize (RF'property _ _ H0); ins; desc.
    assert (l0 = l).
      by destruct x; red in SAMEVAL, SAMEVAL1; ins; desf; ins; rewrite <- eq_rect_eq.
    assert (v0 = v).
      by destruct x; red in SAMEVAL, SAMEVAL1; ins; desf; ins; rewrite <- eq_rect_eq in *; desf.
    desf.
    assert (l0 = l).
    { clear - ENQa SAMEVAL; apply proj2 in ENQa.
      by destruct x; red in SAMEVAL, ENQa; ins; desf; ins; rewrite <- eq_rect_eq.
    }
    assert (l1 = l).
    { clear - ENQa0 SAMEVAL; apply proj2 in ENQa0.
      by destruct x; red in SAMEVAL, ENQa0; ins; desf; ins; rewrite <- eq_rect_eq.
    }
    assert (v0 = v).
    { clear - DEQb SAMEVAL0; apply proj2 in DEQb.
      destruct y; red in SAMEVAL0, DEQb; ins; desf; ins; rewrite <- eq_rect_eq in *; desf.
    }
    assert (v1 = v).
    { clear - DEQb0 SAMEVAL2; apply proj2 in DEQb0.
      destruct z; red in SAMEVAL2, DEQb0; ins; desf; ins; rewrite <- eq_rect_eq in *; desf.
    }
    desf.
    eapply WDinj; eauto.
    congruence.
  }
  
  { (* All "real" dequeues ar in rf' *)
    clear WE WD RF'.
    unfolder; ins; desf.
    apply NNPP; intro N; destruct H1.
    clear - CODOMrf' H H0 N.
    assert (exists v, QueueLib.is_dequeueV' v (fs y)).
    { red in H0; unfold QueueLib.is_dequeueN in N; destruct (fs y); ins; desf.
      specialize (not_ex_all_not _ _ N); clear N; intro N; ins.
      specialize (N eq_refl); clarify_not; desf.
      eexists _, eq_refl; ins; desf.
    }
    desc; eapply CODOMrf'; vauto.
  }

  { (* the total order *)
    ins; exists (restr_rel (set_collect fs g ∩₁ QueueLib.has_loc l) toimm⁺); splits; ins.
    { clear - TOT.
      red; splits.
        by apply strict_partial_order_restr, TOT.
      red; ins.
      eapply TOT in NEQ; eauto.
      desf; unfolder; eauto.
    }
    {
      clear RF'.
 
      assert (FDOM: collect_rel fs po \ ⦗fun _ : event => True⦘ ∪ rf' ∪ hb <<= 
                    set_collect fs g × set_collect fs g).
        rewrite PO, HBDOM, collect_rel_cross; unionL; eauto with hahn.
        unfolder; ins; specialize_full RF'property; eauto; desc.
        by specialize_full TObase; eauto; unfolder in TObase; desc; clarify; split; eauto.

      assert (B: forall x (Gx: g x) y (Gy: g y)
                   (P: (collect_rel fs po \ ⦗fun _ : event => True⦘ ∪ rf' ∪ hb)⁺ (fs x) (fs y)),
                 (po ∪ so ∪ restr_rel g (map_rel fs hb))⁺ x y).
      {
        assert (R: rf' <<= toimm⁺) by (red; ins; edestruct RF'property; eauto).
        ins.
        remember (fs x) as x'.
        remember (fs y) as y'.
        revert x y Gx Gy Heqx' Heqy'; induction P; ins; desf.
          unfolder in H; desf.
          clarify_not; apply t_step; repeat left; eapply poWF; eauto; congruence.
          by eapply R, TOM in H; eauto; eapply clos_trans_mon; eauto; unfolder; ins; desf; eauto.
         apply t_step; unfolder; eauto.
        hahn_rewrite FDOM in P1; hahn_rewrite ct_of_cross in P1; apply proj2 in P1; red in P1; desf.
        eauto using t_trans.
      }

      unfold restr_rel, set_inter; red; ins; desf.  
      assert (set_collect fs g x /\ set_collect fs g y).
        by hahn_rewrite FDOM in H; hahn_rewrite ct_of_cross in H; red in H; desf.
      tertium_non_datur (x = y) as [|NEQ]; desf.
        by exfalso; red in H2; desf; apply B in H; ins; eauto.
      eapply TOT with (l:=l) in NEQ; desf; eauto.
      exfalso; apply TObase in NEQ; desf.
      eapply HBACYC', t_trans; eauto.
      eapply clos_trans_mon; eauto; unfolder; ins; desf; eauto.
    }
    assert (IND: forall S s, 
                 (S = nil \/ 
                 exists x Cx p1 p2, starting_segment (TOT l) FSUPPto x Cx = p1 ++ p2 ++ S /\
                 length (filterP QueueLib.is_enqueue p1) 
                 = 
                 length (filterP (fun e => exists v, QueueLib.is_dequeueV l v e) (p1 ++ p2))
                 /\
                 s = filterP (QueueLib.is_enqueue_loc l) p2)
                 -> QueueLib.is_fifo rf' l s S).
    { clear - RF' RF'property CODOMrf' WE WD DEQN.
      induction S; ins.
      destruct H; ins; desc.
      assert (G: forall x (IN: In x (p1 ++ p2 ++ a :: S)), (set_collect fs g ∩₁ QueueLib.has_loc l) x).
      { clear - H; ins.
        specialize (starting_segment_property (TOT l) FSUPPto x Cx); ins; desf.
        specialize (INIT x0); desf.
        rewrite H in INIT0.
        apply INIT0 in IN; desf.
      }
      assert (AL: (set_collect fs g ∩₁ QueueLib.has_loc l) a) by (apply G; find_in_list).
      destruct (proj2 AL) as [A|A].
      * left; split; try done.
        apply IHS.
        right; exists x, Cx, p1, (p2 ++ a :: nil); splits.
        + by rewrite <- app_assoc, <- app_comm_cons.
        + rewrite plus_n_O at 1.
          rewrite app_assoc, filterP_app, app_length, H0.
          apply Nat.add_cancel_l.
          symmetry; apply length_zero_iff_nil.
          simpl; clear - A; desf; desf.
          red in A, e; destruct a; ins; desf.
        + subst s; rewrite filterP_app; f_equal; ins; desf.
      * right; destruct s as [ | e s'].
        + assert (QueueLib.is_dequeueN l a).
          { clear IHS DEQN.
            apply QueueLib.is_dequeue_alt_def in A; destruct A as [A|A]; desc; try done.
            exfalso.
            specialize_full CODOMrf'; desc.
              by exact (proj1 AL).
              eby eapply QueueLib.is_dequeueV_weaken.
            specialize_full RF'property; eauto; desf.
            - rename s into Ce; rename s0 into Ca.
              assert (l0 = l).
              { clear - Ca AL; apply proj2 in Ca; apply proj2 in AL.
                eapply QueueLib.has_loc_unique; eauto.
              }
              subst l0.
              assert (AS: starting_segment (TOT l) FSUPPto a Ca = (p1 ++ p2) ++ a :: nil).
              { eapply start_of_starting_segment; rewrite H.
                eby rewrite <- app_assoc. 
              }
              rewrite AS, filterP_app, app_length, <- H0 in W; ins; desf; ins.
                2: by destruct n; eauto.
              assert (IN: In e (p1 ++ p2)).
              { 
                specialize (starting_segment_property (TOT l) FSUPPto a Ca); ins; desf.
                forward eapply (proj1 (INIT e)) as IN; vauto.
                rewrite AS in IN; apply in_app_iff in IN; ins; desf.
                eby exfalso; eapply TOT.
              }
              apply in_app_iff in IN; desf.
              * apply in_split in IN; desf.
                rewrite <- !app_assoc, <- app_comm_cons in AS.
                apply start_of_starting_segment with (Cy := Ce) in AS.
                rewrite AS in W; clear - W.
                rewrite !filterP_app, !app_length in *; ins; desf; ins; omega.
              * clear - IN H1 ENQa.
                symmetry in H1; rewrite filterP_eq_nil in H1.
                eapply H1; eauto.
                by exact (proj2 ENQa).
            - destruct n; clear - DEQb.
              split; [exact (proj1 DEQb) | apply proj2 in DEQb].
              by red in DEQb; destruct a; ins; desf; repeat constructor; exists eq_refl.
            - destruct n; clear - ENQa.
              split; [exact (proj1 ENQa) | apply proj2 in ENQa].
              by red in ENQa; destruct e; ins; desf; repeat constructor; exists eq_refl.
            - destruct n; clear - ENQa.
              split; [exact (proj1 ENQa) | apply proj2 in ENQa].
              by red in ENQa; destruct e; ins; desf; repeat constructor; exists eq_refl.
          }
          splits; try done.
          apply IHS; right; exists x, Cx, p1, (p2 ++ a :: nil); splits.
          - by rewrite <- app_assoc, <- app_comm_cons.
          - rewrite plus_n_O at 1.
            rewrite app_assoc, filterP_app, app_length, H0.
            apply Nat.add_cancel_l.
            symmetry; apply length_zero_iff_nil.
            simpl; clear - H2; desf; desf.
            red in e, H2; destruct a; ins; desf; ins.
            rewrite <- eq_rect_eq in *; desf.
          - rewrite filterP_app, <- H1; clear - A; ins; desf.
            red in A, i; destruct a; ins; desf.
        + assert (V: exists v, QueueLib.is_dequeueV l v a).
          { enough (N: ~ QueueLib.is_dequeueN l a).
            { clear - A N.
              red in A; destruct a; ins; desf; ins.
              unfold QueueLib.is_dequeueN in N; ins.
              specialize (not_ex_all_not _ _ N); clear N; intro N; ins.
              specialize (N eq_refl); clarify_not.
              rewrite <- eq_rect_eq in N; desf.
              by eexists _ , eq_refl.
            }   
            intro N; specialize_full DEQN; eauto; desf.
              by eapply AL.
            rename s into Ca.
            assert (SA: starting_segment (TOT l) FSUPPto a Ca = p1 ++ p2 ++ a :: nil).
            { rewrite app_assoc; eapply start_of_starting_segment.
              eby rewrite <- app_assoc.
            }
            rewrite SA in DEQN; clear - N H0 DEQN H1.
            rewrite app_assoc in DEQN at 2.
            rewrite filterP_app in DEQN at 1.
            rewrite filterP_app in DEQN.
            rewrite filterP_app in DEQN.
            rewrite !app_length in DEQN.
            rewrite H0 in DEQN.
            apply Nat.add_cancel_l in DEQN.
            ins; desf; ins; desf; try omega.
            { clear - N e0. red in N, e0; destruct a; ins; desf; ins. 
              rewrite <- eq_rect_eq in *; desf. }
            { clear - N e0. red in N, e0; destruct a; ins; desf; ins. 
              rewrite <- eq_rect_eq in *; desf. }
            rewrite <- plus_n_O, length_zero_iff_nil, filterP_eq_nil in DEQN.
            assert (IN := in_eq e s'); rewrite H1 in IN.
            rewrite in_filterP_iff in IN; desf.
            eapply (DEQN e); eauto.
            clear - IN0.
            red in IN0; destruct e; desf.
          }
          symmetry in H1; apply filterP_eq_cons in H1; desc.
          subst p2.
          assert (P: filterP QueueLib.is_enqueue p = nil).
          { apply filterP_eq_nil.
            intros y N Q.
            forward apply (G y) as Y; [find_in_list | ].
            clear - N Q Y H3.
            eapply H3; eauto.
            apply proj2 in Y; red in Q; repeat (red in Y; desf; ins); destruct y; ins; desf.
          }
          splits.
          - desf; exists l, v; splits.
            { clear - H1 G. split; vauto. eapply G; find_in_list. }
            { clear - V AL. split; vauto. by exact (proj1 AL). }
            clear IHS DEQN; desf. 
            2: destruct n; apply G; find_in_list.
            rename s into Ce; rename s0 into Ca.
            assert (ES: starting_segment (TOT l) FSUPPto e Ce = p1 ++ p ++ e :: nil).
            { rewrite app_assoc; eapply start_of_starting_segment.
              rewrite <- app_assoc.
              eby rewrite <- app_assoc, <- app_comm_cons in H.
            }
            assert (AS: starting_segment (TOT l) FSUPPto a Ca = (p1 ++ p ++ e :: p') ++ a :: nil).
            { eapply start_of_starting_segment.
              rewrite <- !app_assoc.
              eby rewrite <- app_assoc in H.
            }
            rewrite ES, AS.
            rewrite filterP_app at 1.
            rewrite filterP_app with (l' := a :: nil).
            rewrite !app_length, H0.
            apply Nat.add_cancel_l.
            rewrite filterP_app, app_length; ins; desf; ins; desf.
            by rewrite P.
            by destruct n; eauto.
            { destruct n; clear - H1.
              red; red in H1; destruct e; ins; desf. }
            by destruct n0; eauto.
          - clear WD WE RF'; desf.
            rewrite <- !app_assoc, <- app_comm_cons in H. 
            apply IHS; right; exists x, Cx, (p1 ++ p ++ e :: nil), (p' ++ a :: nil); splits.
            -- by rewrite <- !app_assoc, <- app_comm_cons.
            -- rewrite filterP_app at 1.
               rewrite app_assoc, (filterP_app _ (_ ++ _)), <- !app_assoc, !app_length; ins.
               rewrite H0; f_equal.
               rewrite filterP_app, app_length, P; ins; desf.
                 by destruct n; eauto.
               destruct n; clear - H1.
               red; red in H1; destruct e; ins; desf.
            -- rewrite <- app_nil_r at 1. 
               rewrite filterP_app; f_equal.
               symmetry; apply filterP_eq_nil.
               clear - V; ins; desf; red in V; red in COND; destruct x; ins; desf.
    }
    apply IND; clear - SDOM SPREFIX; ins; desf.
    red in SPREFIX; desf; eauto.
    specialize (SDOM _ FIN).
    right; exists el, SDOM, nil, nil; splits; ins.
    assert (T := TOT l); do 2 (red in T; desc).
    eapply StronglySorted_eq with (r := toimm⁺); ins.
    * apply starting_segment_property. 
    * eby eapply StronglySorted_restr.
    * rewrite <- FIN0.
      etransitivity; [by symmetry; apply starting_segment_property|].
      rewrite <- FIN0 in FIN.
      unfold restr_rel; split; ins; desf; eauto using rt_step. 
      hahn_rewrite rt_of_ct in H0; apply rtE in H0; unfolder in H0; desf.
  }
  Grab Existential Variables. all: ins. 
Qed.

Program Definition LockedQueue_impl_verified : libimpl_verified QueueLib.library := 
  {| li_base := LockedQueue_impl |}.
Next Obligation.
  unfold env_compose, env_of_impl; ins; desf; split; ins; desf; eauto.
Qed.
Next Obligation.
  unfold env_compose, env_of_impl in *; ins; desf.
  all: unfold locked_newqueue_body, locked_enqueue_body, locked_dequeue_body in *; ins;
       unfold idom_seq in *; desf.
  all: repeat first [match goal with H : _ ≡₁ _ |- _ => rewrite H; clear H end
                    | rewrite set_le, set_bunion_union_l, set_bunion_eq].
  all: try solve [clear; intro X; apply proj1 in X; eapply X; unfolder; eauto].
Qed. 
Next Obligation.
 unfold env_compose, env_of_impl in *; ins; desf.
 all: unfold locked_newqueue_body, locked_enqueue_body, locked_dequeue_body in *; ins;
      unfold idom_seq in *; desf; ins; unfold idom_seq in *; desf. 
 all: clear_gpos; rewrite_gpos; unionL; unfold lib_events; unfolder; ins; desf.
 all: unfold event_of_call in *; desf; ins.
 all: destruct x; try destruct a; try destruct a0; 
      try destruct a2; try destruct a3; try destruct a4; ins; desf.
 all: try by exists RC11.library; simpls; eauto 8.
 all: try by exists LockLib.library; simpls; eauto 8.
Qed.
Next Obligation.
  exists LockLib.library, (snd loc); split; eauto.
  by constructor; exists eq_refl.
Qed.
Next Obligation.
  exfalso; eapply NEQ, QueueLib.newqueue_loc_unique; eauto.
Qed. 
Next Obligation. 
  unfolder; ins.
  all: revert LIBe; destruct a; simpls; desf.
  all: unfold env_compose, env_of_impl in ENV; simpls; desf; simpls.
  { right; unfold idom_seq in SEM; desf; ins; unfold idom_seq in *; desf.
    all: rewrite_gpos.
    all: unfold event_of_call in *; desf; ins.
    all: apply gEQ in Ge; unfolder in Ge; desf; destruct e; ins; desf; ins.
    all: red in Le, H; ins; desf.
    all: rewrite <- !eq_rect_eq in *.
    * exists a2; splits.
      by apply gEQ; unfolder; eauto.
      by red; rewrite OP2; vauto.
      eexists; splits; try edone.
      by apply poEQ; unfolder; eauto.
      destruct a2; ins; desf; rewrite <- eq_rect_eq.
      by exists eq_refl.
    * exists a2; splits.
      by apply gEQ; unfolder; eauto.
      by red; rewrite OP2; vauto.
      eexists; splits; try edone.
      by apply poEQ; unfolder; eauto.
      destruct a2; ins; desf; rewrite <- eq_rect_eq.
      by exists eq_refl. } 
  all: unfold idom_seq in SEM; desf; ins; unfold idom_seq in *; desf.
  all: clear_gpos; rewrite_gpos.
  all: apply gEQ in Ge; unfolder in Ge; desf; destruct e; ins; desf; ins.
  all: red in Le, H; ins; desf.
  all: unfold event_of_call in *; desf; ins.
  all: try destruct a; try destruct a0; try destruct a1; try destruct a2; try destruct a3; ins; desf. 
  all: rewrite <- !eq_rect_eq in *.
  all: left; eexists; splits. 
  all: try by constructor; exists eq_refl.
Qed.
Next Obligation.
  unfolder; ins; desf.
  all: red in ALLOC; desf.
  all: destruct a; ins; desf.
  all: unfold env_compose, env_of_impl in ENV; ins; desf; ins.
  all: rewrite <- eq_rect_eq in *; ins.
  all: unfold idom_seq in SEM; desf; ins; unfold idom_seq in *; desf.
  all: clear_gpos; rewrite_gpos.
  1: exists a2.
  2: exists a1.
  all: splits.
  all: try by apply gEQ; vauto.
  all: unfold event_of_call in *; desf; ins.
  all: by destruct a1, a2; ins; desf; vauto; exists eq_refl.
Qed.
Next Obligation.
  exact LockedQueue_impl_is_locally_sound.
Qed.

Corollary LockedQueue_impl_ok libs (COMPAT: libs_compatible (libs ∪₁ eq QueueLib.library))
          (L: eq RC11.library ∪₁ eq LockLib.library ⊆₁ libs) A As (e: expr A As) : 
  behaviours libs (expand_fcall (li_code LockedQueue_impl) e) 
  ⊆₁ behaviours (libs ∪₁ eq QueueLib.library) e.
Proof.
  apply impl_ok with (impl := LockedQueue_impl_verified); ins.
  unfolder in *; ins; desf; eauto.
Qed.
