Require Import Hahn ZArith Operations LibBase.

Set Implicit Arguments.

Module ExchangerLib.

  Definition is_newexchanger l e := 
    exists (OP: e.(ev_op) = op_newexchanger), 
      (eq_rect _ _ e.(ev_res) _ OP) = l.

  Definition is_exchangeL (l: nat) e := 
    exists (OP: e.(ev_op) = op_exchange), 
      fst (eq_rect _ _ e.(ev_args) _ (f_equal _ OP)) = l.

  Definition is_exchange (l: nat) (v: Z) (vo: option Z) e := 
    exists (OP: e.(ev_op) = op_exchange), 
      fst (eq_rect _ _ e.(ev_args) _ (f_equal _ OP)) = l /\
      fst (snd (eq_rect _ _ e.(ev_args) _ (f_equal _ OP))) = v /\
      eq_rect _ _ e.(ev_res) _ OP = vo.

  Definition is_ok l := 
    is_newexchanger l ∪₁ ⋃₁v, ⋃₁vo, is_exchange l v vo.
    
  Definition wf (t : tuple) := 
    ⟪WFev: tup_ev t ⊆₁ ⋃₁l, is_ok l⟫ /\
    ⟪WFloc: forall x, tup_ev t ∩₁ is_exchangeL x
                ⊆₁ codom_rel (⦗is_newexchanger x⦘ ⨾ (tup_po t ∪ tup_rf t)⁺)⟫ /\
    ⟪ALLOC: forall l a (EVa: tup_ev t a) (Aa: is_newexchanger l a)
                   b (EVb: tup_ev t b) (Ab: is_newexchanger l b), a = b⟫.

  Definition consistent (t : tuple) :=
    ⟪EOK: tup_ev t ⊆₁ ⋃₁l, is_ok l⟫ /\
    ⟪ALLOC: forall l a (EVa: tup_ev t a) (Aa: is_newexchanger l a)
                   b (EVb: tup_ev t b) (Ab: is_newexchanger l b), a = b⟫ /\
    ⟪RFDOM: tup_rf t ⊆ ⋃l, ⋃v, ⋃v', is_exchange l v (Some v') × is_exchange l v' (Some v)⟫ /\
    ⟪RFFUN: functional (tup_rf t)⟫ /\
    ⟪RFS: symmetric (tup_rf t)⟫ /\
    ⟪RFIRR: irreflexive (tup_rf t)⟫ /\
    ⟪RF: dom_rel (tup_rf t) ≡₁ tup_ev t ∩₁ ⋃₁l, ⋃₁v, ⋃₁v', is_exchange l v (Some v')⟫ /\
    ⟪SOEQ: tup_so t ≡ tup_rf t⟫.

  Local Ltac u := 
    unfold wf, consistent, is_ok, is_newexchanger, is_exchange, is_exchangeL in *; ins; desf.

  Lemma unique_loc: forall e l l' (OK: is_ok l e) (OK': is_ok l' e), l = l'.
  Proof.
    u; unfolder in *; ins; desf; destruct e; ins; desf; rewrite <- ?eq_rect_eq in *; ins.
  Qed.  

  Program Definition library := 
    {| libspec_ops := op_newexchanger :: op_exchange :: nil ;
       libspec_logaddr := nat ;
       libspec_allocs e x := is_newexchanger x e ; 
       libspec_requires e x := exists v vo, is_exchange x v vo e ;
       libspec_wf := wf ;
       libspec_consistent := consistent |}.
  Next Obligation.
    u; unfolder in *; specialize_full WFev; eauto; destruct e; ins; desf; ins; desf; ins. 
    forward eapply WFloc as X; [split; try edone; exists eq_refl; ins|desf; eauto]. 
  Qed.
  Next Obligation.
    unfold wf in *; ins; desf; splits; ins. 
    all: unfolder in *; ins; desc; splits; eauto 8 using t_step. 
    specialize_full WFloc; eauto; clear H H0; desc.
    eexists; split; eauto; clear WFloc1; desf.
    induction WFloc0; [eapply t_step|eapply t_trans]; desf; eauto 10 using t_step, t_trans.
  Qed.
  Next Obligation.
    unfold wf in *; desf; splits; ins.
    red; ins; destruct NEQ; eapply ALLOC with (l:=x); eauto.
  Qed.
  Next Obligation.
    unfold consistent in *; desf; splits; ins.
    red; ins; destruct NEQ; eapply ALLOC with (l:=x); eauto.
  Qed.
  Next Obligation.
    unfold consistent in *; ins; desf. 
    rewrite SOEQ; unfolder; eauto 10 using t_step.
  Qed.
  Next Obligation.
    unfold consistent in *; ins; desf. 
    splits; ins; eauto. 
    2: by unfolder in *; ins; desc; eauto.
    all: try by rewrite <- ?evEQ, <- ?poEQ, <- ?rfEQ, <- ?soEQ, <- ?hbEQ.
    clear - rfEQ RFS; unfolder in *; firstorder. 
  Qed.
  Next Obligation.
    eexists; splits; ins; eauto. 
    unfold consistent in *; desf; splits; ins; rewrite ?SOEQ; ins.
      by unfolder in *; ins; desc; eauto.
      by unfolder in *; ins; desc; eauto.
    all: try solve [rewrite inclusion_restr; ins].
      by clear - RFS; unfolder in *; ins; desf; eauto.
    transitivity (es ∩₁ dom_rel (tup_rf t)).
    2: rewrite RF; clear; unfolder; tauto.
    unfolder; split; ins; desf; eexists; splits; eauto.
    assert (K:=H0); apply SANE in K; unfolder in K; desc.
    apply RFDOM in H0; unfolder in H0; desc.
    rewrite REQ in *; ins; unfold not; unfolder; ins; desf; eauto.
    eapply H in IN'; ins; desf.
    unfold is_exchange in *; desf; destruct x, y; ins; desf;
      rewrite <- !eq_rect_eq in *; desf; ins.
    do 2 eexists; exists eq_refl; ins.
  Qed.

End ExchangerLib.
