Require Import Hahn Operations LibBase.

Set Implicit Arguments.

Module LibReg.

  Definition is_alloc e := ev_op e = reg_alloc.

  Definition is_alloc_loc x e := 
    exists (OP: ev_op e = reg_alloc), 
      eq_rect _ _ (ev_res e) _ OP = x.

  Definition is_read e := ev_op e = reg_read.

  Definition is_read_val x v e := 
    exists (OP: ev_op e = reg_read), 
      fst (eq_rect _ _ (ev_args e) _ (f_equal _ OP)) = x /\ 
      eq_rect _ _ (ev_res e) _ OP = v.

  Definition is_write e := ev_op e = reg_write.

  Definition is_write_val x v e := 
    exists (OP: ev_op e = reg_write), 
      eq_rect _ _ (ev_args e) _ (f_equal _ OP) = (x, (v, tt)).

  Definition wf (t : tuple) := 
      << EOK: tup_ev t ⊆₁ is_alloc ∪₁ is_read ∪₁ is_write >>
   /\ << WF_LOC: 
        forall x, tup_ev t ∩₁ (⋃₁v, is_read_val x v ∪₁ is_write_val x v) 
                ⊆₁ codom_rel (⦗is_alloc_loc x⦘ ⨾ (tup_po t ∪ tup_rf t)⁺) >> 
   /\ << ALLOC : forall x a (EVa: tup_ev t a) (Aa: is_alloc_loc x a)
                           b (EVb: tup_ev t b) (Ab: is_alloc_loc x b), a = b >> .


  Definition consistent (t : tuple) :=
    exists mo, 
(*      << FINSUPP: fsupp (tup_po t ∪ tup_rf t)⁺ >> 
      /\ << SUBHB: (tup_po t ∪ tup_so t)⁺ ⊆ tup_hb t >> /\ *)  
         << EOK: tup_ev t ⊆₁ is_alloc ∪₁ is_read ∪₁ is_write >>
      /\ << ALLOC : forall x a (EVa: tup_ev t a) (Aa: is_alloc_loc x a)
                           b (EVb: tup_ev t b) (Ab: is_alloc_loc x b), a = b >> 
      /\ << RFOK: tup_rf t ⊆ ⋃x, ⋃v, is_write_val x v × is_read_val x v >> 
      /\ << CODOM: codom_rel (tup_rf t) ≡₁ is_read ∩₁ tup_ev t >> 
      /\ << SOEQ: tup_so t ≡ tup_rf t >> 
(*      /\ << POTRANS: transitive (tup_po t) >> *)
      /\ << MO: strict_partial_order mo >>
      /\ << moT: is_total (tup_ev t ∩₁ is_write) mo >>
      /\ << ACYC: acyclic (tup_hb t ∪ mo ∪ tup_rf t ∪ (tup_rf t)⁻¹ ⨾ mo) >>.

  Local Ltac u := 
    unfold is_alloc, is_alloc_loc, is_read, is_read_val, is_write, is_write_val in *; ins; desf.

  Program Definition library := 
    {| libspec_ops := reg_alloc :: reg_read :: reg_write :: nil ;
       libspec_logaddr := address ;
       libspec_allocs e x := is_alloc_loc x e ; 
       libspec_requires e x := exists v, is_read_val x v e \/ is_write_val x v e ;
       libspec_wf := wf ; 
       libspec_consistent := consistent |}.
  Next Obligation.
    unfold wf in *; unfolder in *; desf; u;
    specialize_full EOK; eauto; destruct e; ins; desf; ins; desf.
    all: forward eapply WF_LOC as X; [split; try edone; eexists; split; ins;
                                      constructor; exists eq_refl; ins|desf].
    all: exists z; splits; eauto; destruct z; ins; desf; ins; desf; ins.
  Qed.
  Next Obligation.
    unfold wf in *; desc; splits; ins; eauto.
      by rewrite <- EOK; ins; unfolder; ins; desf. 
      2: by unfold set_inter in *; desf; eauto. 
    rewrite set_interC with (s' := es), set_interA, WF_LOC; clear WF_LOC.
    unfolder in *; ins; desf; eexists; split; eauto. clear H0.
    induction H1; [eapply t_step|eapply t_trans]; desf; eauto 10 using t_step, t_trans.
  Qed.
  Next Obligation.
    unfold wf in *; desc; splits; ins.
    red; ins; destruct NEQ; eapply ALLOC with (x:=x); eauto. 
  Qed.
  Next Obligation.
    unfold consistent in *; desc; splits; ins.
    red; ins; destruct NEQ; eapply ALLOC with (x:=x); eauto. 
  Qed.
(*  Next Obligation.
    unfold consistent, strict_partial_order in *; ins; desf. 
    exists mo; splits. eauto. 
    all: try rewrite SOEQ; ins.
    all: rewrite ?inclusion_seq_eqv_l, ?inclusion_seq_eqv_r; ins.
    by unfolder in *; ins; desf; eauto.
    by unfolder in *; ins; desf; eauto.
    { rewrite <- set_interA, <- CODOM.
      unfolder; splits; ins; desf; eauto.
      eexists x0; splits; eauto.
      eapply PREFIX; unfolder; eauto 10 using t_step.
    }
    by unfolder in *; ins; desf; eapply moT in NEQ; desf; eauto.
  Qed.*)
  Next Obligation.
    unfold consistent in *; ins; desf; rewrite SOEQ; eauto with hahn. 
  Qed.
  Next Obligation.
    unfold consistent, strict_partial_order in *; ins; desf. 
    exists mo; splits; eauto. 
    2: by unfolder in *; ins; desf; eauto.
    all: by rewrite <- ?evEQ, <- ?poEQ, <- ?rfEQ, <- ?soEQ, <- ?hbEQ.
  Qed.
  Next Obligation.
    unfold consistent in *; ins; desf; splits; eauto. 
    exists mo; splits; eauto; rewrite INC2; eauto.
  Qed.
  Next Obligation.
    unfold tuple_sane, consistent, strict_partial_order in *; ins; desf.
    assert (RFEQ: restr_rel es (tup_rf t) ≡ tup_rf t ⨾ ⦗es⦘).
    {
      clear -REQ RFOK RF_OK; split; unfolder; ins; desf; splits; ins.
      unfolder in RF_OK; specialize (RF_OK _ _ H); desc.
      specialize_full RFOK; eauto; desf; unfold is_write_val, is_read_val in *; desf.
      unfolder in RFOK; specialize (RFOK _ _ H); desf.
      rewrite REQ; ins; [rewrite REQ in H0; ins|]; clear REQ.
      all: destruct x, y; ins; desf; ins; desf; ins.
      2-3: by apply set_nonemptyE; repeat eexists; constructor; exists eq_refl; ins. 
      unfolder in *; ins; desf; ins; rewrite <- eq_rect_eq in *; desf.
      eapply H0 in IN'; eauto.
      by eexists; constructor; eexists eq_refl; ins. 
    }
    eexists; splits; ins.
    exists (restr_rel es mo); splits; rewrite ?SOEQ; ins; 
    rewrite ?inclusion_restr; eauto with hahn.
    - rewrite EOK; unfolder; tauto.
    - by unfolder in *; ins; desf; eauto.
    - by rewrite RFEQ, <- set_interA, <- CODOM; clear; unfolder; intuition; desf; eauto.
    - by unfolder; ins; desf; eauto. 
    - by unfolder; ins; desf; eauto; eapply moT in NEQ; desf; eauto.
  Qed.

End LibReg.

